#include "../std.h"
#include <iostream>

#include "../transfer/proxy.h"
#include "../misc/escaped.h"
#include <globus_gsi_credential.h>

#include "auth.h"


int AuthUser::match_all(const char* line) {
  default_voms_=NULL;
  default_vo_=NULL;
  default_role_=NULL;
  default_capability_=NULL;
  default_vgroup_=NULL;
  default_group_=NULL;
  return AAA_POSITIVE_MATCH;
}

int AuthUser::match_group(const char* line) {
  for(;;) {
    string s("");
    int n = input_escaped_string(line,s,' ','"');
    if(n == 0) break;
    line+=n;
    for(list<group_t>::iterator i = groups.begin();i!=groups.end();++i) {
      if(s == i->name) {
        default_voms_=i->voms;
        default_vo_=i->vo;
        default_role_=i->role;
        default_capability_=i->capability;
        default_vgroup_=i->vgroup;
        default_group_=i->name.c_str();
        return AAA_POSITIVE_MATCH;
      };
    };
  };
  return AAA_NO_MATCH;
}

AuthUser::source_t AuthUser::sources[] = {
  { "all", &AuthUser::match_all },
  { "group", &AuthUser::match_group },
  { "subject", &AuthUser::match_subject },
  { "file", &AuthUser::match_file },
  { "vo", &AuthUser::match_ldap },
  { "voms", &AuthUser::match_voms },
  { "plugin", &AuthUser::match_plugin },
  { NULL, NULL }
};


AuthUser::AuthUser(const char* s,const char* f):subject(""),filename("") {
  if(s) { subject=s; make_unescaped_string(subject); };
  if(f) filename=f;
  proxy_file_was_created=false;
#ifdef HAVE_VOMS
  voms_extracted=false;
#endif
  default_voms_=NULL;
  default_vo_=NULL;
  default_role_=NULL;
  default_capability_=NULL;
  default_vgroup_=NULL;
  default_group_=NULL;
}

AuthUser::AuthUser(const AuthUser& a) {
  subject=a.subject;
  filename=a.filename;
  proxy_file_was_created=false;
#ifdef HAVE_VOMS
  voms_extracted=false;
#endif
  default_voms_=NULL;
  default_vo_=NULL;
  default_role_=NULL;
  default_capability_=NULL;
  default_vgroup_=NULL;
  default_group_=NULL;
}

void AuthUser::operator=(const AuthUser& a) {
#ifdef HAVE_VOMS
  voms_data.clear();
  voms_extracted=false;
#endif
  subject=a.subject;
  filename=a.filename;
  proxy_file_was_created=false;
}

void AuthUser::set(const char* s,gss_cred_id_t cred) {
  if(s == NULL) { operator=(cred); return; };
#ifdef HAVE_VOMS
  voms_data.clear();
  voms_extracted=false;
#endif
  subject=s; make_unescaped_string(subject);
  filename="";
  char* p = write_proxy(cred);
  if(p) {
    filename=p;
    proxy_file_was_created=true;
  };
}

void AuthUser::operator=(gss_cred_id_t cred) {
#ifdef HAVE_VOMS
  voms_data.clear();
  voms_extracted=false;
#endif
  subject="";
  filename="";
  proxy_file_was_created=false;
  char* s = write_proxy(cred);
  char* name = NULL;
  if(s) {
    globus_gsi_cred_handle_t h;
    if(globus_gsi_cred_handle_init(&h,GLOBUS_NULL) == GLOBUS_SUCCESS) {
      if(globus_gsi_cred_read_proxy(h,s) == GLOBUS_SUCCESS) {
        if(globus_gsi_cred_get_subject_name(h,&name) != GLOBUS_SUCCESS) {
          name=NULL;
        };
      };
    };
  };
  if(!name) {
    if(s) unlink(s);
  } else {
    subject=name; make_unescaped_string(subject);
    filename=s;
    proxy_file_was_created=true;
  };
} 

AuthUser::~AuthUser(void) {
  if(proxy_file_was_created && filename.length()) unlink(filename.c_str());
}

int AuthUser::evaluate(const char* line) {
  bool invert = false;
  bool no_match = false;
  const char* command = "subject";
  int command_len = 7;
  if(subject.length()==0) return AAA_NO_MATCH;
  if(!line) return AAA_NO_MATCH;
  for(;*line;line++) if(!isspace(*line)) break;
  if(*line == 0) return AAA_NO_MATCH;
  if(*line == '#') return AAA_NO_MATCH;
  if(*line == '-') { line++; invert=true; }
  else if(*line == '+') { line++; };
  if(*line == '!') { no_match=true; line++; };
  if((*line != '/') && (*line != '"')) {
    command=line; 
    for(;*line;line++) if(isspace(*line)) break;
    command_len=line-command;
    for(;*line;line++) if(!isspace(*line)) break;
  };
  for(source_t* s = sources;s->cmd;s++) {
    if((strncmp(s->cmd,command,command_len) == 0) && 
       (strlen(s->cmd) == command_len)) {
      int res=(this->*(s->func))(line);
      if(res == AAA_FAILURE) return res;
      if(no_match) {
        if(res==AAA_NO_MATCH) { res=AAA_POSITIVE_MATCH; }
        else { res=AAA_NO_MATCH; };
      };
      if(invert) res=-res;
      return res;
    };
  };
  return AAA_FAILURE; 
}

#ifdef HAVE_VOMS
const vector<struct voms>& AuthUser::voms(void) {
  if(!voms_extracted) {
    const char* line = "* * * *";
    match_voms(line);
  };
  return voms_data;
};
#endif

AuthEvaluator::AuthEvaluator(void):name("") {

}

AuthEvaluator::AuthEvaluator(const char* s):name(s) {

}

AuthEvaluator::~AuthEvaluator(void) {

}

void AuthEvaluator::add(const char* line) {
  l.push_back(line);
}

int AuthEvaluator::evaluate(AuthUser &u) const {
  for(list<string>::const_iterator i = l.begin();i!=l.end();++i) {
    int r = u.evaluate(i->c_str());
    if(r != AAA_NO_MATCH) return r;
  };
  return AAA_NO_MATCH;
}

