#include "../std.h"
#include <iostream>

#include "../misc/url_options.h"
#include "../misc/stringtoint.h"
#include "../misc/globus_error_utils.h"
#include "../transfer/lister.h"
#include "../misc/log_time.h"
#include "../../userinterface/CertInfo.h"
#include "datapoint.h"
#include "databufferpar.h"
#include "datahandle.h"

#define RESULT_SUCCESS 0
#define RESULT_ERROR 1
#define RESULT_CREDENTIALS_EXPIRED 2

void DataHandle::ftp_complete_callback(void* arg,globus_ftp_client_handle_t* handle,globus_object_t* error) {
  DataHandle* it = (DataHandle*)arg;
  if(error == GLOBUS_SUCCESS) {
    odlog(2)<<"ftp_complete_callback: success"<<endl;
    it->ftp_completed.signal(RESULT_SUCCESS);
  }
  else {
    char* tmp = globus_object_printable_to_string(error);
    odlog(1)<<"ftp_complete_callback: error: "<<tmp<<endl; free(tmp);
    if(it->is_secure) {
      CertInfo ci;
      if(ci.TimeLeft() <= 0) { // proxy expired
        odlog(0)<<"ftp_complete_callback: proxy expired"<<endl;
        it->failure_code=credentials_expired_failure;
        it->ftp_completed.signal(RESULT_CREDENTIALS_EXPIRED);
      } else {
        it->ftp_completed.signal(RESULT_ERROR);
      };
    } else {
      it->ftp_completed.signal(RESULT_ERROR);
    };
  };
}

void DataHandle::ftp_check_callback(void* arg,globus_ftp_client_handle_t* handle,globus_object_t* error,globus_byte_t* buffer,globus_size_t length,globus_off_t offset,globus_bool_t eof) {
  odlog(2)<<"ftp_check_callback"<<endl;
  DataHandle* it = (DataHandle*)arg;
  if(error != GLOBUS_SUCCESS) { 
     odlog(2)<<"Globus error: "<<error<<endl;
     return;
  };
  if(eof) return;
  globus_result_t res=globus_ftp_client_register_read(&(it->ftp_handle),(globus_byte_t*)(it->ftp_buf),sizeof(it->ftp_buf),&ftp_check_callback,it);
  if(res != GLOBUS_SUCCESS) {
    odlog(1)<<"Registration of Globus FTP buffer failed - cancel check"<<endl;
    odlog(2)<<"Globus error: "<<GlobusResult(res)<<endl;
    globus_ftp_client_abort(&(it->ftp_handle));
    return;
  };
  return;
}

/* check if user is allowed to get specified file. */
bool DataHandle::check_ftp(void) {
  globus_result_t res;
  globus_off_t size = 0;
  globus_abstime_t gl_modify_time;
  time_t modify_time;
  int modify_utime;
  int c_res;
  res = globus_ftp_client_size(&ftp_handle,c_url.c_str(),&ftp_opattr,
                     &size,&ftp_complete_callback,this);
  if(res != GLOBUS_SUCCESS) {
    odlog(2)<<"check_ftp: globus_ftp_client_size failed"<<endl;
    odlog(1)<<"Globus error"<<GlobusResult(res)<<endl;
    return false;
  };
  if(!ftp_completed.wait(c_res,300000)) { /* 5 minutes timeout */
    odlog(1)<<"check_ftp: timeout waiting for size"<<endl;
    globus_ftp_client_abort(&ftp_handle);
    ftp_completed.wait(c_res,-1);
    return false;
  };
  if(c_res != 0) {
    odlog(1)<<"check_ftp: failed to get file's size"<<endl;
    return false;
  };
  url->meta_size(size);
  res = globus_ftp_client_modification_time(&ftp_handle,c_url.c_str(),
                     &ftp_opattr,&gl_modify_time,&ftp_complete_callback,this);
  if(res != GLOBUS_SUCCESS) {
    odlog(2)<<"check_ftp: globus_ftp_client_modification_time failed"<<endl;
    odlog(1)<<"Globus error"<<GlobusResult(res)<<endl;
    return false;
  };
  if(!ftp_completed.wait(c_res,300000)) { /* 5 minutes timeout */
    odlog(1)<<"check_ftp: timeout waiting for modification_time"<<endl;
    globus_ftp_client_abort(&ftp_handle);
    ftp_completed.wait(c_res,-1);
    return false;
  };
  if(c_res != 0) {
    odlog(1)<<"check_ftp: failed to get file's modification time"<<endl;
    return false;
  };
  GlobusTimeAbstimeGet(gl_modify_time,modify_time,modify_utime);
  url->meta_created(modify_time);
  // Do not use partial_get for ordinary ftp. Stupid globus tries to 
  // use non-standard commands anyway.
  if(is_secure) {
    res=globus_ftp_client_partial_get(&ftp_handle,c_url.c_str(),&ftp_opattr,
                         GLOBUS_NULL,0,1,&ftp_complete_callback,this);
    if(res != GLOBUS_SUCCESS) {
      odlog(2)<<"check_ftp: globus_ftp_client_get failed"<<endl;
      odlog(1)<<"Globus error"<<GlobusResult(res)<<endl;
      return false;
    };
    ftp_eof_flag=false; /* use eof_flag to pass result from callback */
    odlog(2)<<"check_ftp: globus_ftp_client_register_read"<<endl;
    res=globus_ftp_client_register_read(&ftp_handle,(globus_byte_t*)ftp_buf,sizeof(ftp_buf),&ftp_check_callback,this);
    if(res != GLOBUS_SUCCESS) { 
      globus_ftp_client_abort(&ftp_handle);
      ftp_completed.wait(c_res,-1);
      return false;
    };
    if(!ftp_completed.wait(c_res,300000)) { /* 5 minutes timeout */
      odlog(1)<<"check_ftp: timeout waiting for partial get"<<endl;
      globus_ftp_client_abort(&ftp_handle);
      ftp_completed.wait(c_res,-1);
      return false;
    };
    return (c_res==0);
  } else {
    // Do not use it at all. It does not give too much usefull information anyway.
    return true;
  };
}

bool DataHandle::remove_ftp(void) {
  char buf[32];
  globus_result_t res;
  res=globus_ftp_client_delete(&ftp_handle,c_url.c_str(),&ftp_opattr,
                         &ftp_complete_callback,this);
  if(res != GLOBUS_SUCCESS) {
    odlog(2)<<"delete_ftp: globus_ftp_client_delete failed"<<endl;
    odlog(1)<<"Globus error"<<GlobusResult(res)<<endl;
    return false;
  };
  int c_res;
  if(!ftp_completed.wait(c_res,300000)) { /* 5 minutes timeout */
    odlog(1)<<"delete_ftp: globus_ftp_client_delete timeout"<<endl;
    globus_ftp_client_abort(&ftp_handle);
    ftp_completed.wait(c_res,-1);
    return false;
  };
  return (c_res==0);
}

static bool remove_last_dir(string &dir) {
  // dir also contains proto and server
  int nn = string::npos;
  if(!strncasecmp(dir.c_str(),"ftp://",6)) { nn=dir.find('/',6); }
  else if(!strncasecmp(dir.c_str(),"gsiftp://",9)) { nn=dir.find('/',9); };
  if(nn == string::npos) return false;
  int n;
  if((n=dir.rfind('/')) == string::npos) return false;
  if(n < nn) return false;
  dir.resize(n);
  return true;
}

static bool add_last_dir(string &dir,string &path) {
  int l = dir.length();
  int n = path.find('/',l+1);
  if(n == string::npos) return false;
  dir=path; dir.resize(n);
  return true;
}

bool DataHandle::mkdir_ftp() {
  ftp_dir_path=c_url;
  for(;;) if(!remove_last_dir(ftp_dir_path)) break;
  /*
  for(;;) {
    if(!remove_last_dir(ftp_dir_path)) {
      return false;
    };
    odlog(2)<<"mkdir_ftp: checking for "<<ftp_dir_path<<endl;
    globus_result_t res=globus_ftp_client_exists(&ftp_handle,ftp_dir_path.c_str(),&ftp_opattr,&ftp_complete_callback,this);
    if(res != GLOBUS_SUCCESS) {
      odlog(1)<<"Globus error: "<<GlobusResult(res)<<endl; return false;
    };
    int c_res;
    if(!ftp_completed.wait(c_res,300000)) {
      odlog(1)<<"mkdir_ftp: timeout waiting for exists"<<endl;
      // timeout - have to cancel operation here
      globus_ftp_client_abort(&ftp_handle);
      ftp_completed.wait(c_res,-1);
      return false;
    };
    if(c_res == RESULT_CREDENTIALS_EXPIRED) return false;
    if(c_res == RESULT_SUCCESS) break;
  };
  // have this object - go back with creating directories 
  */
  bool result = false;
  for(;;) {
    if(!add_last_dir(ftp_dir_path,c_url)) break;
    odlog(2)<<"mkdir_ftp: making "<<ftp_dir_path<<endl;
    globus_result_t res=globus_ftp_client_mkdir(&ftp_handle,ftp_dir_path.c_str(),&ftp_opattr,&ftp_complete_callback,this);
    if(res != GLOBUS_SUCCESS) {
      odlog(1)<<"Globus error: "<<GlobusResult(res)<<endl; return false;
    };
    int c_res;
    if(!ftp_completed.wait(c_res,300000)) {
      odlog(1)<<"mkdir_ftp: timeout waiting for mkdir"<<endl;
      /* timeout - have to cancel operation here */
      globus_ftp_client_abort(&ftp_handle);
      ftp_completed.wait(c_res,-1);
      return false;
    };
    if(c_res == RESULT_CREDENTIALS_EXPIRED) return false;
    // if(c_res != 0) return false;
    result = result || (c_res == 0);
  };
  return result;
  // return true;
}

bool DataHandle::start_reading_ftp(DataBufferPar &buf) {
  buffer=&buf;
  /* size of file first */
  globus_off_t size = 0;
  bool limit_length = false;
  unsigned long long int range_length;
  if(range_end > range_start) {
    range_length = range_end - range_start;
    limit_length = true;
  };
  odlog(2)<<"start_reading_ftp"<<endl;
  ftp_completed.reset(); ftp_eof_flag=false;
  globus_ftp_client_handle_cache_url_state(&ftp_handle,c_url.c_str());
  globus_result_t res;
  int c_res;      
  if(!no_checks) {
    odlog(2)<<"start_reading_ftp: size: url: "<<c_url.c_str()<<endl;
    res = globus_ftp_client_size(&ftp_handle,c_url.c_str(),
                     &ftp_opattr,&size,&ftp_complete_callback,this);
    if(res != GLOBUS_SUCCESS) {
      odlog(0)<<"start_reading_ftp: failure"<<endl;
      odlog(1)<<"Globus error: "<<GlobusResult(res)<<endl;
      globus_ftp_client_handle_flush_url_state(&ftp_handle,c_url.c_str());
      buffer->error_read(true);
      return false;
    };
    if(!ftp_completed.wait(c_res,300000)) {
      odlog(0)<<"start_reading_ftp: timeout waiting for file size"<<endl;
      /* timeout - have to cancel operation here */
      odlog(1)<<"Timeout waiting for FTP file size - cancel transfer"<<endl;
      globus_ftp_client_abort(&ftp_handle);
      /* have to do something in addition if complete callback will be called */
      ftp_completed.wait(c_res,-1);
      globus_ftp_client_handle_flush_url_state(&ftp_handle,c_url.c_str());
      buffer->error_read(true);
      return false;
    };
    if(c_res != 0) return false;
    /* provide some metadata */
    odlog(1)<<"start_reading_ftp: obtained size: "<<size<<endl;
    url->meta_size(size);
    odlog(2)<<"start_reading_ftp: set size: "<<url->meta_size()<<endl;
    globus_abstime_t gl_modify_time;
    res = globus_ftp_client_modification_time(&ftp_handle,c_url.c_str(),
                     &ftp_opattr,&gl_modify_time,&ftp_complete_callback,this);
    if(res != GLOBUS_SUCCESS) {
      odlog(2)<<"start_reading_ftp: globus_ftp_client_modification_time failed"<<endl;
      odlog(1)<<"Globus error"<<GlobusResult(res)<<endl;
      buffer->error_read(true);
      return false;
    };
    if(!ftp_completed.wait(c_res,300000)) { /* 5 minutes timeout */
      odlog(1)<<"start_reading_ftp: timeout waiting for modification_time"<<endl;
      globus_ftp_client_abort(&ftp_handle);
      ftp_completed.wait(c_res,-1);
      globus_ftp_client_handle_flush_url_state(&ftp_handle,c_url.c_str());
      buffer->error_read(true);
      return false;
    };
    if(c_res != 0) {
      odlog(1)<<"start_reading_ftp: failed to get file's modification time"<<endl;
      buffer->error_read(true);
      return false;
    };
    time_t modify_time;
    int modify_utime;
    GlobusTimeAbstimeGet(gl_modify_time,modify_time,modify_utime);
    url->meta_created(modify_time);
    odlog(2)<<"start_reading_ftp: set creation time: "<<url->meta_created()<<endl;
    if(limit_length) {
      if(size < range_end) {
        if(size <= range_start) { // report eof immediately
          odlog(2)<<"start_reading_ftp: range is out of size"<<endl;
          buffer->eof_read(true);
          ftp_completed.signal(RESULT_SUCCESS);
          return true;
        };
        range_length=size-range_start;
      };
    };
  };
  odlog(2)<<"start_reading_ftp: globus_ftp_client_get"<<endl;
  if(limit_length) {
    res=globus_ftp_client_partial_get(&ftp_handle,c_url.c_str(),&ftp_opattr,
                         GLOBUS_NULL,range_start,range_start+range_length+1,&ftp_get_complete_callback,this);
  } else {
    res=globus_ftp_client_get(&ftp_handle,c_url.c_str(),&ftp_opattr,
                         GLOBUS_NULL,&ftp_get_complete_callback,this);
  };
  if(res != GLOBUS_SUCCESS) {
    odlog(2)<<"start_reading_ftp: globus_ftp_client_get failed"<<endl;
    odlog(1)<<"Globus error: "<<GlobusResult(res)<<endl;
    globus_ftp_client_handle_flush_url_state(&ftp_handle,c_url.c_str());
    buffer->error_read(true);
    return false;
  };
  if(globus_thread_create(&ftp_control_thread,GLOBUS_NULL,&ftp_read_thread,this) != 0) {
    odlog(2)<<"start_reading_ftp: globus_thread_create failed"<<endl;
    globus_ftp_client_abort(&ftp_handle);
    ftp_completed.wait(c_res,-1);
    globus_ftp_client_handle_flush_url_state(&ftp_handle,c_url.c_str());
    buffer->error_read(true);
    return false;
  };
  /* make sure globus has thread for handling network/callbacks */
  globus_thread_blocking_will_block();
  return true;
}

bool DataHandle::stop_reading_ftp(void) {
  if(!buffer->eof_read()) {
    odlog(2)<<"stop_reading_ftp: aborting connection"<<endl;
    globus_ftp_client_abort(&ftp_handle);
  };
  int c_res;
  odlog(2)<<"stop_reading_ftp: waiting for transfer to finish"<<endl;
  ftp_completed.wait(c_res,-1); 
  odlog(2)<<"stop_reading_ftp: exiting: "<<c_url<<endl;
  globus_ftp_client_handle_flush_url_state(&ftp_handle,c_url.c_str());
  return true;
}

void* DataHandle::ftp_read_thread(void *arg) {
  DataHandle* it = (DataHandle*)arg;
  int h;
  unsigned int l;
  globus_result_t res;
  int registration_failed = 0;
  odlog(1)<<"ftp_read_thread: get and register buffers"<<endl;
  int n_buffers = 0;
  for(;;) {
    if(it->buffer->eof_read()) break;
    if(!it->buffer->for_read(h,l,true)) { /* eof or error */
      if(it->buffer->error()) { /* error -> abort reading */
        odlog(2)<<"ftp_read_thread: for_read failed - aborting: "<<it->c_url<<endl;
        globus_ftp_client_abort(&(it->ftp_handle));
      };
      break;
    };
    res=globus_ftp_client_register_read(&(it->ftp_handle),
              (globus_byte_t*)((*(it->buffer))[h]),l,
              &(it->ftp_read_callback),it);
    if(res != GLOBUS_SUCCESS) {
      odlog(3)<<"ftp_read_thread: Globus error: "<<GlobusResult(res)<<endl;
      globus_object_t* err = globus_error_get(res);
      registration_failed++;
      if(registration_failed>=10) {
        it->buffer->is_read(h,0,0);
        it->buffer->error_read(true);
        it->buffer->eof_read(true); // can set eof here because no callback will be called (I guess).
        odlog(3)<<"ftp_read_thread: too many registration failures - abort: "<<it->c_url<<endl;
      } else {
        odlog(3)<<"ftp_read_thread: failed to register globus buffer - will try later: "<<it->c_url<<endl;
        it->buffer->is_read(h,0,0); sleep(1);
      };
    }
    else { n_buffers++; };
  };
  /* make sure complete callback is called */
  odlog(2)<<"ftp_read_thread: waiting for eof"<<endl;
  it->buffer->wait_eof_read();
  odlog(2)<<"ftp_read_thread: exiting"<<endl;
  it->ftp_completed.signal(it->buffer->error_read()?1:0);
  return NULL;
}

void DataHandle::ftp_read_callback(void* arg,globus_ftp_client_handle_t* handle,globus_object_t* error,globus_byte_t* buffer,globus_size_t length,globus_off_t offset,globus_bool_t eof) {
  DataHandle* it = (DataHandle*)arg;
  if(error != GLOBUS_SUCCESS) {
    it->buffer->is_read((char*)buffer,0,0);
    return;
  };
  it->buffer->is_read((char*)buffer,length,offset);
  if(eof) it->ftp_eof_flag=true;
  return;
}

void DataHandle::ftp_get_complete_callback(void* arg,globus_ftp_client_handle_t* handle,globus_object_t* error) {
  odlog(2)<<"ftp_get_complete_callback"<<endl;
  DataHandle* it = (DataHandle*)arg;
  /* data transfer finished */
  if(error != GLOBUS_SUCCESS) {
    odlog(1)<<"Failed to get ftp file."<<endl;
    odlog(2)<<"Globus error: "<<error<<endl;
    if(it->is_secure) {
      CertInfo ci;
      if(ci.TimeLeft() <= 0) { // proxy expired
        odlog(0)<<"ftp_get_complete_callback: proxy expired"<<endl;
        it->failure_code=credentials_expired_failure;
      };
    };
    it->buffer->error_read(true);
    return;
  };
  it->buffer->eof_read(true);
  return;
}

bool DataHandle::start_writing_ftp(DataBufferPar &buf) {
  buffer=&buf;
  /* size of file first */
  globus_off_t size;
  bool limit_length = false;
  unsigned long long int range_length;
  if(range_end > range_start) {
    range_length = range_end - range_start;
    limit_length = true;
  };
  odlog(2)<<"start_writing_ftp"<<endl;
  ftp_completed.reset(); ftp_eof_flag=false; ftp_counter.reset();
  globus_result_t res;
  int c_res;
/* !!!! TODO - preallocate here */

  globus_ftp_client_handle_cache_url_state(&ftp_handle,c_url.c_str());
  if(!no_checks) {
    odlog(2)<<"start_writing_ftp: mkdir"<<endl;
    if(!DataHandle::mkdir_ftp()) {
      odlog(2)<<"start_writing_ftp: mkdir failed - still trying to write"<<endl;
    };
  };
  odlog(2)<<"start_writing_ftp: put"<<endl;
  if(limit_length) {
    res=globus_ftp_client_partial_put(&ftp_handle,c_url.c_str(),&ftp_opattr,
                          GLOBUS_NULL,range_start,range_start+range_length,&ftp_put_complete_callback,this);

  } else {
    res=globus_ftp_client_put(&ftp_handle,c_url.c_str(),&ftp_opattr,
                         GLOBUS_NULL,&ftp_put_complete_callback,this);
  };
  if(res != GLOBUS_SUCCESS) {
    odlog(2)<<"start_writing_ftp: put failed"<<endl;
    odlog(1)<<"Globus error: "<<GlobusResult(res)<<endl;
    globus_ftp_client_handle_flush_url_state(&ftp_handle,c_url.c_str());
    buffer->error_write(true);
    return false;
  };
  if(globus_thread_create(&ftp_control_thread,GLOBUS_NULL,&ftp_write_thread,this) != 0) {
    odlog(2)<<"start_writing_ftp: globus_thread_create failed"<<endl;
    globus_ftp_client_handle_flush_url_state(&ftp_handle,c_url.c_str());
    buffer->error_write(true);
    return false;
  };
  /* make sure globus has thread for handling network/callbacks */
  globus_thread_blocking_will_block();
  return true;
}

bool DataHandle::stop_writing_ftp(void) {
  if(!buffer->eof_write()) {
    globus_ftp_client_abort(&ftp_handle);
  };
  int c_res;
  ftp_completed.wait(c_res,-1);
  globus_ftp_client_handle_flush_url_state(&ftp_handle,c_url.c_str());
  return true;
}

void* DataHandle::ftp_write_thread(void *arg) {
  DataHandle* it = (DataHandle*)arg;
  int h;
  unsigned int l;
  unsigned long long int o;
  globus_result_t res;
  globus_bool_t eof = GLOBUS_FALSE;
  odlog(1)<<"ftp_write_thread: get and register buffers"<<endl;
  int n_buffers = 0;
  for(;;) {
    if(!it->buffer->for_write(h,l,o,true)) {
      if(it->buffer->error()) {
        odlog(2)<<"ftp_write_thread: for_write failed - aborting"<<endl;
        globus_ftp_client_abort(&(it->ftp_handle));
        break;
      };
      // no buffers and no errors - must be pure eof
      eof = GLOBUS_TRUE;
      char dummy;
      o = it->buffer->eof_position();
      res=globus_ftp_client_register_write(&(it->ftp_handle),
                   (globus_byte_t*)(&dummy),0
                   ,o,eof,&ftp_write_callback,it);
      break;
      // if(res == GLOBUS_SUCCESS) break;
      // sleep(1); continue;
    };
    res=globus_ftp_client_register_write(&(it->ftp_handle),
                   (globus_byte_t*)((*(it->buffer))[h]),l
                   ,o,eof,&ftp_write_callback,it);
    if(res != GLOBUS_SUCCESS) { 
      it->buffer->is_notwritten(h); sleep(1);
    };
  };
  /* make sure complete callback is called */
  it->buffer->wait_eof_write();
  it->ftp_completed.signal(it->buffer->error_write()?1:0);
  return NULL;
}

void DataHandle::ftp_write_callback(void* arg,globus_ftp_client_handle_t* handle,globus_object_t* error,globus_byte_t* buffer,globus_size_t length,globus_off_t offset,globus_bool_t eof) {
  DataHandle* it = (DataHandle*)arg;
  // if(it->ftp_eof_flag) return; /* eof callback */
  if(error != GLOBUS_SUCCESS) {
    it->buffer->is_written((char*)buffer);
    return;
  };
  it->buffer->is_written((char*)buffer);
  return;
}

void DataHandle::ftp_put_complete_callback(void* arg,globus_ftp_client_handle_t* handle,globus_object_t* error) {
  odlog(2)<<"ftp_put_complete_callback"<<endl;
  DataHandle* it = (DataHandle*)arg;
  /* data transfer finished */
  if(error != GLOBUS_SUCCESS) {
    odlog(1)<<"Failed to store ftp file."<<endl;
    odlog(2)<<"Globus error: "<<error<<endl;
    if(it->is_secure) {
      CertInfo ci;
      if(ci.TimeLeft() <= 0) { // proxy expired
        odlog(0)<<"ftp_put_complete_callback: proxy expired"<<endl;
        it->failure_code=credentials_expired_failure;
      };
    };
    it->buffer->error_write(true);
    return;
  };
  it->buffer->eof_write(true);
  return;
}

bool DataHandle::list_files_ftp(list<DataPoint::FileInfo> &files,bool resolve) {
  Lister lister; 
  if(lister.retrieve_dir(c_url) != 0) {
    odlog(0)<<"Failed to obtain listing from ftp: "<<c_url<<endl;
    return false;
  };
  lister.close_connection();
  bool result = true;
  for(list<ListerFile>::iterator i = lister.begin();i!=lister.end();++i) {
    list<DataPoint::FileInfo>::iterator f = 
          files.insert(files.end(),DataPoint::FileInfo(i->GetName().c_str()));
    if(resolve) {
      globus_result_t res;
      globus_off_t size = 0;
      globus_abstime_t gl_modify_time;
      time_t modify_time;
      int modify_utime;
      int c_res;
      // Here comes a tricky part: how to recognize object itself
      string f_url = c_url;
      if((lister.size() != 1) || (i->GetName()[0] != '/')) {
        f_url+="/"+i->GetName();
      };
      f->type=(DataPoint::FileInfo::Type)(i->GetType());
      if(i->CheckSize()) {
        f->size=i->GetSize(); f->size_available=true;
      } else if(i->GetType() != ListerFile::file_type_dir) {
        res = globus_ftp_client_size(&ftp_handle,f_url.c_str(),&ftp_opattr,
                       &size,&ftp_complete_callback,this);
        if(res != GLOBUS_SUCCESS) {
          odlog(2)<<"list_files_ftp: globus_ftp_client_size failed"<<endl;
          odlog(1)<<"Globus error"<<GlobusResult(res)<<endl; result=false;
        } else if(!ftp_completed.wait(c_res,300000)) { /* 5 minutes timeout */
          odlog(1)<<"list_files_ftp: timeout waiting for size"<<endl;
          globus_ftp_client_abort(&ftp_handle);
          ftp_completed.wait(c_res,-1); result=false;
        } else if(c_res != 0) {
          odlog(1)<<"list_files_ftp: failed to get file's size"<<endl;
          result=false;
        } else {
          f->size=size; f->size_available=true;
        };
      };
      if(i->CheckCreated()) {
        f->created=i->GetCreated(); f->created_available=true;
      } else {
        res = globus_ftp_client_modification_time(&ftp_handle,f_url.c_str(),
                     &ftp_opattr,&gl_modify_time,&ftp_complete_callback,this);
        if(res != GLOBUS_SUCCESS) {
          odlog(2)<<"list_files_ftp: globus_ftp_client_modification_time failed"<<endl;
          odlog(1)<<"Globus error"<<GlobusResult(res)<<endl; result=false;
        } else if(!ftp_completed.wait(c_res,300000)) { /* 5 minutes timeout */
          odlog(1)<<"list_files_ftp: timeout waiting for modification_time"<<endl;
          globus_ftp_client_abort(&ftp_handle);
          ftp_completed.wait(c_res,-1); result=false;
        } else if(c_res != 0) {
          odlog(1)<<"list_files_ftp: failed to get file's modification time"<<endl;
          result=false;
        } else {
          GlobusTimeAbstimeGet(gl_modify_time,modify_time,modify_utime);
          f->created=modify_time; f->created_available=true;
        };
      };
    };
  };
  return result;
}

