#include "../std.h"
#include "../misc/url_options.h"
#include "../misc/checkfile.h"
#include "../misc/checksum.h"
#include "../misc/mkdir_recursive.h"
#include "../misc/globus_modules.h"
#include "../misc/log_time.h"

#include "datapoint.h"
#include "datahandle.h"
#include "databufferpar.h"

#include "datamove.h"

#ifndef BUFLEN
#define BUFLEN 1024
#endif

DataMove::DataMove() {
  be_verbose=false;
  force_secure=false;
  force_passive=false;
  force_registration=false;
  do_retries=true;
  do_checks=true;
  default_min_speed=0;
  default_min_speed_time=0;
  default_min_average_speed=0;
  default_max_inactivity_time=300;
  show_progress=NULL;
}

DataMove::~DataMove() {
}

bool DataMove::verbose(void) { return be_verbose; };

void DataMove::verbose(bool val) { be_verbose=val; };

void DataMove::verbose(const string &prefix) { 
  be_verbose=true; verbose_prefix=prefix;
};

bool DataMove::retry(void) { return do_retries; };

void DataMove::retry(bool val) { do_retries=val; };

bool DataMove::checks(void) { return do_checks; };

void DataMove::checks(bool val) { do_checks=val; };

typedef struct {
  DataPoint *source;
  DataPoint *destination;
  DataCache *cache;
  const UrlMap *map;
  unsigned long long int min_speed;
  time_t min_speed_time;
  unsigned long long int min_average_speed;
  time_t max_inactivity_time;
  DataMove::callback cb;
  DataMove* it;
  void* arg;
  const char* prefix;
} transfer_struct;

void* transfer_func(void* arg) {
  transfer_struct* param = (transfer_struct*)arg;
  DataMove::result res = param->it->Transfer(
         *(param->source),*(param->destination),*(param->cache),*(param->map),
         param->min_speed,param->min_speed_time,
         param->min_average_speed,param->max_inactivity_time,
         NULL,NULL,param->prefix);
  (*(param->cb))(param->it,res,param->arg);
  if(param->prefix) free((void*)(param->prefix));
  delete param->cache;
  free(param);
  return NULL;
}

/* transfer data from source to destination */
DataMove::result DataMove::Transfer(
      DataPoint &source,DataPoint &destination,
      DataCache &cache,const UrlMap &map,
      DataMove::callback cb,void* arg,const char* prefix) {
  return Transfer(source,destination,cache,map,
           default_min_speed,default_min_speed_time,
           default_min_average_speed,
           default_max_inactivity_time,
           cb,arg,prefix);
}

DataMove::result DataMove::Transfer(
      DataPoint &source,DataPoint &destination,
      DataCache &cache,const UrlMap &map,
      unsigned long long int min_speed,time_t min_speed_time,
      unsigned long long int min_average_speed,
      time_t max_inactivity_time,
      DataMove::callback cb,void* arg,const char* prefix) {
  if(cb != NULL) {
    odlog(2)<<"DataMove::Transfer : starting new thread"<<endl;
    pthread_t thread;
    pthread_attr_t thread_attr;
    pthread_attr_init(&thread_attr);
    pthread_attr_setdetachstate(&thread_attr,PTHREAD_CREATE_DETACHED);
    transfer_struct* param = (transfer_struct*)malloc(sizeof(transfer_struct));
    if(param==NULL) {
      pthread_attr_destroy(&thread_attr); return system_error;
    };
    param->source=&source;
    param->destination=&destination;
    param->cache=new DataCache(cache);
    param->map=&map;
    param->min_speed=min_speed;
    param->min_speed_time=min_speed_time;
    param->min_average_speed=min_average_speed;
    param->max_inactivity_time=max_inactivity_time;
    param->cb=cb;
    param->it=this;
    param->arg=arg;
    param->prefix=NULL;
    if(prefix) param->prefix=strdup(prefix);
    if(param->prefix == NULL) param->prefix=strdup(verbose_prefix.c_str());
    if(pthread_create(&thread,&thread_attr,&transfer_func,param) != 0) {
      free(param); pthread_attr_destroy(&thread_attr); return system_error;
    };
    pthread_attr_destroy(&thread_attr);
    return success;
  };
  odlog(1)<<"Transfer from "<<source.canonic_url()<<" to "<<destination.canonic_url()<<endl;
  if(!source) {
    odlog(0)<<"Not valid source"<<endl; return read_acquire_error;
  };
  if(!destination) {
    odlog(0)<<"Not valid destination"<<endl; return write_acquire_error;
  };
  int try_num;
  for(try_num=0;;try_num++) {
    if(source.meta_resolve(true,map)) {
      if(source.have_locations()) break;
      odlog(0)<<"No locations for source found: "<<source<<endl;
    } else odlog(0)<<"Failed to resolve source: "<<source<<endl;
    source.next_location(); /* try again */
    if((try_num != 0) && (!do_retries)) return read_resolve_error;
    if(!source.have_location()) return read_resolve_error;
  };
  for(try_num=0;;try_num++) {
    if(destination.meta_resolve(false,UrlMap())) {
      if(destination.have_locations()) break;
      odlog(0)<<"No locations for destination found: "<<destination<<endl;
    } else odlog(0)<<"Failed to resolve destination: "<<destination<<endl;
    destination.next_location(); /* try again */
    if((try_num != 0) && (!do_retries)) return write_resolve_error;
    if(!destination.have_location()) return write_resolve_error;
  };
  bool replication=false;
  if(source.meta() && destination.meta()) {
    // check for possible replication
    if(source.canonic_url() == destination.canonic_url()) {
      replication=true;
      // we do not want to replicate to same site
      destination.remove_locations(source);
      if(!destination.have_locations()) {
        odlog(0)<<"No locations for destination different from source found: "<<destination<<endl;
        return write_resolve_error;
      };
    };
  };
  bool destination_meta_initially_stored = destination.meta_stored();
  result res = transfer_error;
  for(try_num=0;;try_num++) {  /* cycle for retries */
    odlog(2)<<"DataMove: cycle"<<endl;
    if((try_num != 0) && (!do_retries)) {
      odlog(2)<<"DataMove: no retries requested - exit"<<endl;
      return res;
    };
    if( (!source.have_location()) || (!destination.have_location()) ) {
      odlog(2)<<"DataMove: ";
      if(!source.have_location()) odlog_(2)<<" source ";
      if(!destination.have_location()) odlog_(2)<<" destination ";
      odlog_(2)<<" out of tries - exit"<<endl;
      /* out of tries */
      return res;
    };
    odlog(1)<<"Real transfer from "<<source.current_location()<<" to "<<destination.current_location()<<endl;
    /* creating handler for transfer */
    DataHandle source_h(&source);
    DataHandle destination_h(&destination);
    source_h.secure(force_secure);
    source_h.passive(force_passive);
    source_h.additional_checks(do_checks);
    destination_h.secure(force_secure);
    destination_h.passive(force_passive);
    destination_h.additional_checks(do_checks);
    /* take suggestion from DataHandle about buffer, etc. */
    DataHandle::analyze_t hint_read;
    DataHandle::analyze_t hint_write;
    bool cacheable=false;
    long int bufsize;
    int bufnum;
    source_h.analyze(hint_read);
    destination_h.analyze(hint_write);
    if(hint_read.cache && hint_write.local) {
      if(cache) cacheable=true;
    };
    /* tune buffers */
    bufsize=65536; /* have reasonable buffer size */
    bool seekable = destination_h.out_of_order();
    source_h.out_of_order(seekable);
    bufnum=1;
    if(hint_read.bufsize  > bufsize) bufsize=hint_read.bufsize;
    if(hint_write.bufsize > bufsize) bufsize=hint_write.bufsize;
    if(seekable) {
      if(hint_read.bufnum  > bufnum)  bufnum=hint_read.bufnum;
      if(hint_write.bufnum > bufnum)  bufnum=hint_read.bufnum;
    };
    bufnum=bufnum*2;
    odlog(2)<<"Creating buffer: "<<bufsize<<" x "<<bufnum<<endl;
    /* prepare crc */
    CheckSumAny crc;
    // Shold we trust indexing service or always compute checksum ?
    // Let's trust.
    if(destination.accepts_meta()) { // may need to compute crc
      // Let it be CRC32 by default.
      string crc_type;
      {
        string d_url = destination.base_url();
        if(get_url_option(d_url,"checksum",0,crc_type) != 0) {
          crc_type="cksum";
        };
        odlog(2)<<"DataMove::Transfer: checksum type is "<<crc_type<<endl;
      };
      if(!source.meta_checksum_available()) {
        crc=crc_type.c_str();
        odlog(2)<<"DataMove::Transfer: will try to compute crc"<<endl;
      } else if(CheckSumAny::Type(crc_type.c_str()) != CheckSumAny::Type(source.meta_checksum())) {
        crc=crc_type.c_str();
        odlog(2)<<"DataMove::Transfer: will try to compute crc"<<endl;
      };
    };
    /* create buffer and tune speed control */
    DataBufferPar buffer(&crc,bufsize,bufnum);
    if(!buffer) {
      odlog(1)<<"Buffer creation failed !"<<endl;
    };
    buffer.speed.set_min_speed(min_speed,min_speed_time);
    buffer.speed.set_min_average_speed(min_average_speed);
    buffer.speed.set_max_inactivity_time(max_inactivity_time);
    buffer.speed.verbose(be_verbose);
    if(be_verbose) {
      if(prefix) { buffer.speed.verbose(string(prefix)); }
      else { buffer.speed.verbose(verbose_prefix); };
      buffer.speed.set_progress_indicator(show_progress);
    };
    /* checking if current source should be mapped to different location */
    /*
       TODO: make mapped url to be handled by source handle directly
    */
    bool mapped = false;
    string mapped_url = "";
    if(hint_write.local) {
      mapped_url = source.current_location();
      ::canonic_url(mapped_url);
      mapped=map.map(mapped_url);
      /* TODO: copy options to mapped_url */
      if(!mapped) {
        mapped_url="";
      } else {
        odlog(2)<<"Url is mapped to: "<<mapped_url<<endl;
        if((strncasecmp(mapped_url.c_str(),"link:/",6) == 0) ||
           (strncasecmp(mapped_url.c_str(),"file:/",6) == 0)) {
          /* can't cache links */
          cacheable=false;
        };
      };
    };
    // Do not link if user asks. Replace link:// with file://
    if(hint_read.readonly && mapped) {
      if(strncasecmp(mapped_url.c_str(),"link:/",6) == 0) {
        mapped_url.replace(0,4,"file");
      };
    };
    DataPoint mapped_p(mapped_url.c_str());
    DataHandle mapped_h(&mapped_p);
    mapped_h.secure(force_secure);
    mapped_h.passive(force_passive);
    mapped_h.additional_checks(do_checks);
    /* Try to initiate cache (if needed) */
    if(cacheable) {
      res=success;
      for(;;) { /* cycle for outdated cache files */
        bool is_in_cache = false;
        if(!cache.start(source.canonic_url().c_str(),is_in_cache)) {
          cacheable=false;
          odlog(1)<<"Failed to initiate cache"<<endl;
          break;
        };
        if(is_in_cache) { /* just need to check permissions */
          odlog(1)<<"File is cached - checking permissions"<<endl;
          if(!source_h.check()) {
            odlog(0)<<"Permission checking failed: "<<source<<endl;
            cache.stop(true,false);
            source.next_location(); /* try another source */
            odlog(2)<<"source.next_location"<<endl;
            res=read_start_error;
            break;
          };
          odlog(2)<<"Permission checking passed"<<endl;
          /* check if file is fresh enough */
          bool outdated = true;
          if(source.meta_created_available() && cache.created_available()) {
            if(source.meta_created() <= cache.created()) outdated=false;
          }
          else if(cache.validtill_available()) {
            if(cache.validtill() > time(NULL)) outdated=false;
          };
          if(outdated) {
            cache.stop(false,true);
            odlog(1)<<"Cached file is outdated"<<endl;
            continue;
          };
          if(hint_read.readonly) {
            odlog(2)<<"Linking/copying cached file"<<endl;
            if(!cache.link(get_url_path(destination.current_location()))){
              /* failed cache link is unhandable */
              cache.stop(true,false);
              return cache_error;
            };
          } else {
            odlog(2)<<"Copying cached file"<<endl;
            if(!cache.copy(get_url_path(destination.current_location()))){
              /* failed cache copy is unhandable */
              cache.stop(true,false);
              return cache_error;
            };
          };
          cache.stop(false,false);
          return success; // Leave here. Rest of code below is for transfer.
        };
        break;
      };
      if(cacheable) {
        if(res != success) continue;
      };
    };
    if(mapped) {
      if((strncasecmp(mapped_url.c_str(),"link:/",6) == 0) ||
         (strncasecmp(mapped_url.c_str(),"file:/",6) == 0)) {
        /* check permissions first */
        odlog(1)<<"URL is mapped to local access - checking permissions on original URL"<<endl;
        if(!source_h.check()) {
          odlog(0)<<"Permission checking on original URL failed: "<<source<<endl;
          source.next_location(); /* try another source */
          odlog(2)<<"source.next_location"<<endl;
          res=read_start_error;
          if(cacheable) cache.stop(true,false);
          continue;
        };
        odlog(2)<<"Permission checking passed"<<endl;
        if(strncasecmp(mapped_url.c_str(),"link:/",6) == 0) {
          odlog(2)<<"Linking local file"<<endl;
          const char* file_name = get_url_path(mapped_url.c_str());
          const char* link_name = get_url_path(destination.current_location());
          // create directory structure for link_name
          {
            uid_t uid=get_user_id();
            gid_t gid=get_user_group(uid);
            string dirpath = link_name;
            int n = dirpath.rfind('/');
            if(n == string::npos) n=0;
            if(n == 0) { dirpath="/"; } else { dirpath.resize(n); };
            if(mkdir_recursive(NULL,dirpath.c_str(),S_IRWXU,uid,gid) != 0) {
              if(errno != EEXIST) {
               char* err;
               char errbuf[BUFLEN];
               err=strerror_r(errno,errbuf,sizeof(errbuf));
                odlog(0)<<"Failed to create/find directory "<<dirpath<<" : "<<err<<endl;
                source.next_location(); /* try another source */
                odlog(2)<<"source.next_location"<<endl;
                res=read_start_error;
                if(cacheable) cache.stop(true,false);
                continue;
              };
            };
          };
          // make link
          if(symlink(file_name,link_name) == -1) {
            char* err;
            char errbuf[BUFLEN];
            err=strerror_r(errno,errbuf,sizeof(errbuf));
            odlog(0)<<"Failed to make symbolic link "<<link_name<<" to "<<file_name<<" : "<<err<<endl;
            source.next_location(); /* try another source */
            odlog(2)<<"source.next_location"<<endl;
            res=read_start_error;
            if(cacheable) cache.stop(true,false);
            continue;
          };
          {
            uid_t uid = get_user_id();
            gid_t gid = get_user_group(uid);
            lchown(link_name,uid,gid);
          };
          if(cacheable) cache.stop(false,false);
          return success; // Leave after making a link. Rest moves data.
        };
      };
    };
    if(!mapped) {
      if(!source_h.start_reading(buffer)) {
        odlog(0)<<"Failed to start reading from source: "<<source<<endl;
        res=read_start_error;
        if(source_h.failure_reason()==DataHandle::credentials_expired_failure) 
                        res=credentials_expired_error;
        source.next_location(); /* try another source */
        odlog(2)<<"(Re)Trying next source"<<endl;
        if(cacheable) cache.stop(true,false);
        continue;
      };
    }
    else {
      if(!mapped_h.start_reading(buffer)) {
        odlog(0)<<"Failed to start reading from mapped source: "<<mapped_p<<endl;
        res=read_start_error;
        if(mapped_h.failure_reason()==DataHandle::credentials_expired_failure) 
                        res=credentials_expired_error;
        source.next_location(); /* try another source */
        odlog(2)<<"(Re)Trying next source"<<endl;
        if(cacheable) cache.stop(true,false);
        continue;
      };
      destination.meta(mapped_p);
    };
    if(force_registration && destination.meta()) { // at least compare metadata
      if(!destination.meta_compare(source)) {
        odlog(0)<<"Metadata of source and destination are different"<<endl;
        source.next_location(); /* not exactly sure if this would help */
        res=preregister_error;
        if(cacheable) cache.stop(true,false);
        continue;
      };
    };
    destination.meta(source); // pass metadata gathered during start_reading()
                              // from source to destination
    if(destination.meta_size_available()) {
      buffer.speed.set_max_data(destination.meta_size());
    };
    if(!destination.meta_preregister(replication,force_registration)) {
      odlog(0)<<"Failed to preregister destination: "<<destination<<endl;
      destination.next_location(); /* not exactly sure if this would help */
      odlog(2)<<"destination.next_location"<<endl;
      res=preregister_error;
      if(cacheable) cache.stop(true,false);
      continue;
    };
    buffer.speed.reset();
    DataHandle::failure_reason_t read_failure = DataHandle::common_failure;
    DataHandle::failure_reason_t write_failure = DataHandle::common_failure;
    if(!cacheable) {
      if(!destination_h.start_writing(buffer)) {
        odlog(0)<<"Failed to start writing to destination: "<<destination<<endl;
        if(!mapped) { source_h.stop_reading(); }
        else { mapped_h.stop_reading(); };
        if(!destination.meta_preunregister(replication || destination_meta_initially_stored)) {
          odlog(0)<<"Failed to unregister preregistered lfn, You may need to unregister it manually: "<<destination<<endl;
        };
        destination.next_location();
        odlog(2)<<"destination.next_location"<<endl;
        res=write_start_error;
        if(destination_h.failure_reason()==
         DataHandle::credentials_expired_failure) res=credentials_expired_error;
        continue;
      };
      odlog(2)<<"Waiting for buffer"<<endl;
      for(;(!buffer.eof_read() || !buffer.eof_write()) && !buffer.error();) {
        buffer.wait();
      };
      odlog(1)<<"buffer: read eof : "<<buffer.eof_read()<<endl;
      odlog(1)<<"buffer: write eof: "<<buffer.eof_write()<<endl;
      odlog(1)<<"buffer: error    : "<<buffer.error()<<endl;
      odlog(2)<<"Closing read channel"<<endl;
      if(!mapped) {
        source_h.stop_reading();
        read_failure=source_h.failure_reason();
      } else {
        mapped_h.stop_reading();
        read_failure=mapped_h.failure_reason();
      };
      odlog(2)<<"Closing write channel"<<endl;
      destination_h.stop_writing();
      write_failure=destination_h.failure_reason();
    }
    else {  /* create new destination for cache file */
      string churl = "file://"+cache.file();
      odlog(1)<<"cache file: "<<churl<<endl;
      DataPoint chdest(churl.c_str());
      DataHandle chdest_h(&chdest);
      chdest_h.secure(force_secure);
      chdest_h.passive(force_passive);
      chdest_h.additional_checks(do_checks);
      chdest.meta(destination); // share metadata
      if(!chdest_h.start_writing(buffer,&cache)) {
        odlog(0)<<"Failed to start writing to cache"<<endl;
        if(!mapped) { source_h.stop_reading(); }
        else { mapped_h.stop_reading(); };
        cache.stop(true,false);
        if(!destination.meta_preunregister(replication || destination_meta_initially_stored)) {
          odlog(0)<<"Failed to unregister preregistered lfn, You may need to unregister it manually"<<endl;
        };
        return cache_error;  /* repeating won't help here */
      };
      odlog(2)<<"Waiting for buffer"<<endl;
      for(;(!buffer.eof_read() || !buffer.eof_write()) && !buffer.error();) {
        buffer.wait();
      };
      odlog(1)<<"buffer: read eof : "<<buffer.eof_read()<<endl;
      odlog(1)<<"buffer: write eof: "<<buffer.eof_write()<<endl;
      odlog(1)<<"buffer: error    : "<<buffer.error()<<endl;
      odlog(2)<<"Closing read channel"<<endl;
      if(!mapped) {
        source_h.stop_reading();
        read_failure=source_h.failure_reason();
      } else {
        mapped_h.stop_reading();
        source.meta(mapped_p); // pass more metadata (checksum)
        read_failure=mapped_h.failure_reason();
      };
      odlog(2)<<"Closing write channel (cache)"<<endl;
      chdest_h.stop_writing();
      write_failure=destination_h.failure_reason();
      bool download_error = buffer.error();
      if(!download_error) {
        if(source.meta_created_available()) cache.created_force(source.meta_created());
        if(source.meta_validtill_available()) cache.validtill_force(source.meta_validtill());
        odlog(2)<<"Linking/copying cached file"<<endl;
        if(!cache.link(get_url_path(destination.current_location()))) {
          buffer.error_write(true);
          cache.stop(download_error,false);
          if(!destination.meta_preunregister(replication || destination_meta_initially_stored)) {
            odlog(0)<<"Failed to unregister preregistered lfn, You may need to unregister it manually"<<endl;
          };
          return cache_error;  /* retry won't help */
        };
      };
      cache.stop(download_error,false);
    };
    if(buffer.error()) {
      if(!destination.meta_preunregister(replication || destination_meta_initially_stored)) {
        odlog(0)<<"Failed to unregister preregistered lfn, You may need to unregister it manually"<<endl;
      };
      // analyze errors
      if(buffer.error_read())  {
        odlog(2)<<"(Re)Trying next source"<<endl;
        source.next_location();
        if(read_failure == DataHandle::credentials_expired_failure) {
          res=credentials_expired_error;
        } else { res=read_error; };
      } else if(buffer.error_write()) {
        destination.next_location();
        odlog(2)<<"(Re)Trying next destination"<<endl;
        if(write_failure == DataHandle::credentials_expired_failure) {
          res=credentials_expired_error;
        } else { res=write_error; };
      } else if(buffer.error_transfer()) {
        res=transfer_error;
        if(destination.local()) { 
          odlog(2)<<"source.next_location"<<endl;
          source.next_location();
        }
        else { 
          odlog(2)<<"destination.next_location"<<endl;
          destination.next_location(); 
        };
      };
      continue;
    };
    if(crc) {
      if(buffer.checksum_valid()) {
        // source.meta_checksum(crc.end());
        char buf[100];
        crc.print(buf,100);
        source.meta_checksum(buf);
        odlog(2)<<"DataMove::Transfer: have valid checksum"<<endl;
      };
    };
    destination.meta(source); // pass more metadata (checksum)
    if(!destination.meta_postregister(replication,buffer.error())) {
      odlog(0)<<"Failed to postregister destination"<<destination<<endl;
      if(!destination.meta_preunregister(replication || destination_meta_initially_stored)) {
        odlog(0)<<"Failed to unregister preregistered lfn, You may need to unregister it manually"<<destination<<endl;
      };
      destination.next_location(); /* not sure if this can help */
      odlog(2)<<"destination.next_location"<<endl;
      res=postregister_error;
      continue;
    };
    if(buffer.error()) continue; // should never happen - keep just in case
    break;
  };
  return success;
}

void DataMove::secure(bool val) {
  force_secure=val;
}

void DataMove::passive(bool val) {
  force_passive=val;
}

void DataMove::force_to_meta(bool val) {
  force_registration=val;
}


static const char* result_string[] = {
    "success",  // 0
    "bad source URL", // 1
    "bad destination URL", // 2
    "failed to resolve source locations", // 3
    "failed to resolve destination locations", // 4
    "failed to register new destination file", // 5
    "can't start reading from source", // 6
    "can't start writing to destination", // 7
    "can't read from source", // 8
    "can't write to destination", // 9
    "data transfer was too slow", // 10
    "failed while closing connection to source", // 11
    "failed while closing connection to destination", // 12
    "failed to register new location", // 13
    "can't use local cache", // 14
    "system error", // 15
    "delegated credentials expired" // 16
};

const char* DataMove::get_result_string(result r) {
  if(r == undefined_error) return "unexpected error";
  if((r<0) || (r>16)) return "unknown error";
  return result_string[r];
}

