#ifndef GRID_MANAGER_INFO_TYPES_H
#define GRID_MANAGER_INFO_TYPES_H

#include "../std.h"
#include <string>
#include <list>
#include <iostream>

#ifdef stdin
#undef stdin
#endif
#ifdef stdout
#undef stdout
#endif
#ifdef stderr
#undef stderr
#endif

/* 
  Defines few data types used by grid-manager to store information
  about jobs.
*/

/*
  Pair of values containing file's path (pfn - physical file name)
  and it's source or destination on the net (lfn - logical file name)
*/
class FileData {
 public:
  typedef list<FileData>::iterator iterator;
  FileData(void);
  FileData(const char *pfn_s,const char *lfn_s);
  string pfn;  /* path relative to session dir */
  string lfn;  /* input/output url or size.checksum */
  FileData& operator= (const char* str);
  bool operator== (const char* name);
  bool operator== (const FileData& data);
  bool has_lfn(void);
};
istream &operator>> (istream &i,FileData &fd);
ostream &operator<< (ostream &o,const FileData &fd);

/*
  Time with visual representation suitable for puting into
  MDS fields Mds-Valid-from and Mds-Valid-to. It looks like
  YYYYMMDDhhmmssZ . Here Z stands for Z. All values are GMT.
  Internally it's represented by Unix timestamp.
*/
class mds_time {
 private:
  time_t t;
 public:
  mds_time(void) { t=(time_t)(-1); };
  mds_time(int tm) { t=(time_t)tm; };
  mds_time(time_t tm) { t=tm; };
  mds_time& operator= (time_t tm) { t=tm; return (*this); };
  mds_time& operator= (string s);
  mds_time& operator= (const char* s);
  operator time_t (void) const { return t; };
  bool defined(void) const { return (t != -1); };
  bool operator>  (time_t tm) { return (t>tm);  };
  bool operator>= (time_t tm) { return (t>=tm); };
  bool operator<= (time_t tm) { return (t<=tm); };
  bool operator<  (time_t tm) { return (t<tm);  };
  bool operator== (time_t tm) { return (t==tm); };
};

istream &operator>> (istream &i,mds_time &t);
ostream &operator<< (ostream &o,const mds_time &t);

/*
  Most important information about job extracted from different sources
  (mostly from job description) and stored in separate file for 
  relatively quick and simple access.
*/
class JobLocalDescription {
 /* all values are public, this class is just for convenience */
 public:
  JobLocalDescription(void):lrms(""),queue(""),jobid(""),localid(""),
                           DN(""),starttime(),lifetime(""),
                           notify(""),exectime(),reruns(0),jobname(""),
                           cleanuptime(),gsiftpthreads(1),clientname(""),
                           clientsoftware(""),
                           dryrun(false),diskspace(0),downloads(-1),uploads(-1),
                           failedstate(""),jobreport(""),expiretime()
  {
  };
  string jobid;              /* job's unique identificator */
/* attributes stored in job.ID.local */
  string lrms;               /* lrms type to use - pbs */
  string queue;              /* queue name  - default */
  string localid;            /* job's id in lrms */
  list<string> arguments;    /* executable + arguments */
  string DN;                 /* user's distinguished name aka subject name */
  mds_time starttime;        /* job submission time */
  string lifetime;           /* time to live for submission directory */
  string notify;             /* notification flags used and email address */
  mds_time processtime;      /* time to start job processing (downloading) */
  mds_time exectime;         /* time to start execution */
  string clientname;         /* IP+port of user interface + info given by ui */
  string clientsoftware;     /* Client's version */
  int    reruns;             /* number of allowed reruns left */
  int    downloads;          /* number of downloadable files requested */
  int    uploads;            /* number of uploadable files requested */
  string jobname;            /* name of job given by user */
  string jobreport;          /* URL of user's/VO's logger */
  mds_time cleanuptime;      /* time to remove job completely */
  mds_time expiretime;       /* when delegation expires */
  string stdlog;             /* dirname to which log messages will be
                                put after job finishes */
  string sessiondir;         /* job's session directory */
  string failedstate;        /* state at which job failed, used for rerun */
/* attributes stored in other files */
  list<FileData> inputdata;  /* input files */
  list<FileData> outputdata; /* output files */
/* attributes taken from RSL */
  string action;             /* what to do - must be 'request' */
  string rc;                 /* url to contact replica collection */
  string stdin;              /* file name for stdin handle */
  string stdout;             /* file name for stdout handle */
  string stderr;             /* file name for stderr handle */
  string cache;              /* cache default, yes/no */
  int    gsiftpthreads;      /* number of parallel connections to use 
                                during gsiftp down/uploads */
  bool   dryrun;             /* if true, this is test job */
  unsigned long long int diskspace;  /* anount of requested space on disk */
};

/* Information stored in job.#.lrms_done file */
class LRMSResult {
 private:
  int code_;
  string description_;
  bool set(const char* s);
 public:
  LRMSResult(void):code_(-1),description_("") { };
  LRMSResult(const string& s) { set(s.c_str()); };
  LRMSResult(int c):code_(c),description_("") { };
  LRMSResult(const char* s) { set(s); };
  void operator=(const string& s) { set(s.c_str()); };
  void operator=(const char* s) { set(s); };
  int code(void) const { return code_; }; 
  const string& description(void) const { return description_; };
};
istream& operator>>(istream& i,LRMSResult &r);
ostream& operator<<(ostream& i,const LRMSResult &r);

/*
  Writes to 'o' string 'str' prepending each blank space with \ .
*/
void output_escaped_string(ostream &o,const string &str);

#endif
