#include "../../std.h"

#include <iostream>
#include <fstream>

#include "../../misc/stringtoint.h"
#include "../../misc/escaped.h"
#include "../../misc/time_utils.h"
#include "../../misc/log_time.h"
#include "files.h"


SEFileHandle::SEFileHandle(SEFile& f,uint64_t o,bool for_read):file(f) {
  active=true;
  offset=o;
  open_read=for_read;
  if(file.open(open_read) != 0) active=false;
  odlog(3)<<"SEFileHandle constructor: active: "<<active<<endl;
}

SEFileHandle::~SEFileHandle(void) {
  if(active) file.close(open_read);
}

uint64_t SEFileHandle::read(void* buf,uint64_t size) {
  odlog(3)<<"SEFileHandle::read - size: "<<size<<endl;
  if(!active) return 0;
  uint64_t l = file.read(buf,offset,size);
  offset+=l;
  odlog(3)<<"SEFileHandle::read - l: "<<l<<endl;
  return l;
}

uint64_t SEFileHandle::write(void* buf,uint64_t size) {
  if(!active) return 0;
  uint64_t l = file.write(buf,offset,size);
  offset+=l;
  return l;
}

SEFileHandle* SEFileHandle::open(const char* id,uint64_t o,bool for_read,SEFiles& files) {
  SEFile* file = files.get(id);
  if(!file) return NULL;
  SEFileHandle* f = new SEFileHandle(*file,o,for_read);
  if(!(*f)) { delete f; return NULL; };
  return f;
}

SEFiles::SEFiles(void) {
  odlog(4)<<"SEFiles::SEFiles(void)"<<endl;
//  files=NULL;
  ns=NULL;
//  nfiles=0;
  valid=false;
  reg_type=registration_immediate;
}

SEFiles::SEFiles(const char* dirpath):path(dirpath),space(dirpath) {
  odlog(4)<<"SEFiles::SEFiles(const char* dirpath)"<<endl;
  valid=false;
//  files=NULL;
  ns=NULL;
  reg_type=registration_immediate;
//  nfiles=0;
  struct dirent file_;
  struct dirent *file;
  DIR *dir=opendir(dirpath);
  if(dir == NULL) {
    odlog(0)<<"Failed opening directory: "<<dirpath<<endl; return;
  };
  for(;;) {
    readdir_r(dir,&file_,&file);
    if(file == NULL) break;
    int l=strlen(file->d_name);
    if(l>5) {
      if(!strcmp(file->d_name+l-5,".attr")) {
        string name(file->d_name); name.resize(l-5);
        name="/"+name; name=dirpath+name;
        odlog(2)<<"SEFiles: creating SEFile: "<<name<<endl;
        SEFile* f = new SEFile(name.c_str(),space);
        if(!(*f)) {
          odlog(0)<<"SEFiles: failed to acquire SEFile: "<<name<<endl;
        } else {
          odlog(2)<<"SEFiles: adding SEFile: "<<name<<endl;
          add(*f);
          odlog(1)<<"Added file: "<<f->id()<<endl;
        };
      };
    };
  };
  closedir(dir);
  valid=true;
}

SEFiles::~SEFiles() {
  lock.block();
//  for(int i=0;i<nfiles;i++) {
//    if(files[i]) delete files[i];
//  };
//  free(files); files=NULL;
  valid=false;
  lock.unblock();
}

SEFile* SEFiles::add(SEFile& file) {
  odlog(2)<<"SEFiles::add"<<endl;
  lock.block(); // TODO: to remove ?!
  if(files) {
    SEFile* f = find_by_id(file.id());
    if(f) {
      // File with this id already exists. Allow (re)creation if 
      // owner is a same person (DN) and other attributes are same
      // too (if set).
      bool same = true;
      if(file.creator() != f->creator()) same=false;
      if(same && file.size_available() && f->size_available()) {
        if(file.size() != f->size()) same=false;
      };
      if(same && file.checksum_available() && f->checksum_available()) {
        if(file.checksum() != f->checksum()) same=false;
      };
      if(same && file.created_available() && f->created_available()) {
        if(memcmp(file.created(),f->created(),sizeof(struct tm))!=0)same=false;
      };
      if(!same) {
        odlog(0)<<"SEFiles::add: file already exists: "<<file.id()<<endl;
        file.destroy();
        lock.unblock();
        return NULL;
      };
      // files look identical, destroy new and replace it by new
      file.destroy();
      lock.unblock();
      return f;
    };
  };
  odlog(2)<<"SEFiles::add: new file: "<<file.id()<<endl;
  files.add(file);
  lock.unblock();
  return &file;
}

bool SEFiles::remove(SEFile& file) {
  for(SEFiles::iterator i=begin();i!=end();++i) {
    if(&(*i) == &file) {
      return remove(i);
    };
  };
  return false;
}

bool SEFiles::remove(SEFiles::iterator& file) {
   odlog(1)<<"SEFiles::remove: "<<file->id()<<endl;
  if(!file) return true;
  lock.block(); // TODO: to remove
  file->destroy(); 
  files.remove(file);
  lock.unblock();
  return true;
}

SEFile* SEFiles::find_by_id(const char* id) {
  for(SEFiles::iterator i = begin();i!=end();++i) {
    if(i->check_id(id)) {
      SEFile* f = &(*i);
      return f;
    };
  };
  return NULL;
}

SEFile* SEFiles::get(const char* id) {
  lock.block(); // TODO: to remove
  SEFile* f = find_by_id(id);
  lock.unblock();
  return f;
}

// Register all files, which require registration
int SEFiles::Register(void) {
  odlog(2)<<"SEFiles::Register"<<endl;
  // TODO: Be careful with NS - it can be non-thread-safe
  int failures = 0;
  if(files && ns) {
    for(SEFiles::iterator f = files.begin();f!=files.end();++f) {
      ns->Connect(); // take advantage of possibly less reconnections
      // analyze situation
      f->acquire();
      if((f->state_reg() == REG_STATE_UNREGISTERING) ||
         (f->state_reg() == REG_STATE_REGISTERING)) {
        f->release(); continue; // file is alredy claimed
      };
      if(f->state_reg()!=REG_STATE_LOCAL) {  // already/yet registered 
        f->release(); continue;
      };
      if(!register_retry()) { // this function only serves retries
        f->release(); continue;
      };
      if(!(
           (f->state_file()==FILE_STATE_VALID) ||
           (register_immediately() && (f->state_file()==FILE_STATE_COLLECTING))
          )) { // not proper for register
        f->release(); continue;
      };
      // claim file being registered
      odlog(1)<<"Registering: "<<f->id()<<endl;
      if(!f->state_reg(REG_STATE_REGISTERING)) {
        odlog(0)<<"Registering: "<<f->id()<<" - change state failed"<<endl;
        failures++;
        f->release(); continue;
      };
      f->release();
      if(ns->Register(*f) != 0) {
        failures++;
        if(!f->state_reg(REG_STATE_LOCAL)) {
          // ???????????
        };
        continue;
      };
      if(f->state_reg(REG_STATE_ANNOUNCED)) {
        // ???????????
        continue;
      };
    };
    ns->Disconnect();
  };
  return failures;
}

// Unregister all files being deleted
int SEFiles::Unregister(void) {
  odlog(3)<<"SEFiles::Unregister"<<endl;
  int failures = 0;
  if(files && ns) {
    for(SEFiles::iterator f = files.begin();f!=files.end();++f) {
      ns->Connect(); // take advantage of possibly less reconnections
      // analyze situation
      f->acquire(); // lock file to prevent race while analyzing state
      if((f->state_reg() == REG_STATE_UNREGISTERING) ||
         (f->state_reg() == REG_STATE_REGISTERING)) {
        f->release(); continue; // file is already claimed
      };
      if(f->state_file()!=FILE_STATE_DELETING) { // not proper for unregister
        f->release(); continue;
      };
      if(f->state_reg()==REG_STATE_LOCAL) {  // stuck file 
        f->release(); f->destroy(); files.remove(f); continue;
      };
      if(f->state_reg()!=REG_STATE_ANNOUNCED) {  // already/yet not registered 
        f->release(); continue;
      };
      if(!register_retry()) { // this function only serves retries
        f->release(); continue;
      };
      // claim file being unregistered
      odlog(1)<<"Unregistering: "<<f->id()<<endl;
      if(!f->state_reg(REG_STATE_UNREGISTERING)) {
        odlog(0)<<"Unregistering: "<<f->id()<<" - change state failed"<<endl;
        failures++;
        f->release(); continue;
      };
      f->release();
      if(ns->Unregister(*f) != 0) {
        odlog(0)<<"Unregistering - Unregister failed"<<endl;
        failures++;
        if(!f->state_reg(REG_STATE_ANNOUNCED)) {
          // ???????????
        };
        continue;
      };
      odlog(2)<<"Unregistering - Unregister succeeded"<<endl;
      if(!f->state_reg(REG_STATE_LOCAL)) {
        failures++;
        // ???????????
        continue;
      };
      // unregistered file in deleting state  - destroy and remove
      f->destroy();
      files.remove(f);
    };
    ns->Disconnect();
  };
  return failures;
}

// There MUST be only one thread which uses this function.
// Verify content of file (and do immediate (re)registration)
int SEFiles::Verify(void) {
  odlog(3)<<"SEFiles::Verify"<<endl;
  int failures = 0;
  if(files) {
    for(SEFiles::iterator f = files.begin();f!=files.end();++f) {
      // Verification
      f->acquire();
      if(f->state_file()!=FILE_STATE_COMPLETE) { f->release(); continue; };
      if(!f->complete()) { f->release(); continue; }; // metadata not complete
      f->release();
      int res = f->verify();
      if(res == 1) {
        odlog(0)<<"File has incomplete (meta)data, validation failed: "<<f->id()<<endl;
      } else if(res != 0) {
        odlog(0)<<"Failed to validate file: "<<f->id()<<endl;
      };
      f->acquire();
      if(res != 0) { // destroy file
        if(!f->state_file(FILE_STATE_DELETING)) {
          odlog(0)<<"Failed to change file state (what to do?): "<<f->id()<<endl;
        };
        f->release(); continue;
      };
      // Attempt of immediate registration
      if(ns) {
        if(f->state_reg(REG_STATE_REGISTERING)) {
          f->release();
          if(ns->Register(*f) != 0) {
            if(register_retry()) {
              odlog(0)<<"Failed to register (will retry)"<<endl;
            } else {
              odlog(0)<<"Failed to register (what to do ?)"<<endl;
            };
            f->state_reg(REG_STATE_LOCAL);
          } else {
            f->state_reg(REG_STATE_ANNOUNCED);
          };
          f->acquire();
        } else {
          if(register_retry()) {
            odlog(0)<<"Failed to set REGISTERING state (will retry)"<<endl;
          } else {
            odlog(0)<<"Failed to set REGISTERING state (what to do ?)"<<endl;
          };
        };
      };
      if(!f->state_file(FILE_STATE_VALID)) {
        odlog(0)<<"Verify: "<<f->id()<<" - change state failed"<<endl;
        failures++; continue;
      };
      f->release();
    };
  };
  return failures;
}

int SEFiles::RemoveStuck(void) {
  odlog(3)<<"SEFiles::RemoveStuck"<<endl;
  int failures = 0;
  if(files) {
    for(SEFiles::iterator f = files.begin();f!=files.end();++f) {
      // analyze situation
      f->acquire();
      if(f->state_file() == FILE_STATE_COLLECTING) {
        if((time(NULL) - f->state_file_changed()) > (24*60*60)) {
          if((time(NULL) - f->data_changed()) > (12*60*60)) {
            // file was not touched for long time - probably forgotten
            odlog(0)<<"Removing file which stayed too long in COLLECTING state: "<<f->id()<<endl;
            bool unregistered = true;
            if(f->state_reg() == REG_STATE_ANNOUNCED) {
              unregistered=false;
              if(!f->state_reg(REG_STATE_UNREGISTERING)) {
                odlog(0)<<"SEFiles::RemoveStuck: failed to set REG_STATE_UNREGISTERING"<<endl;
              } else {
                f->release();
                if(NS()->Unregister(*f) != 0) {
                  f->acquire();
                  f->state_reg(REG_STATE_ANNOUNCED);
                  odlog(0)<<"SEFiles::RemoveStuck: failed to unregister"<<endl;
                } else {
                  f->acquire();
                  f->state_reg(REG_STATE_LOCAL);
                  unregistered=true;
                };
              };
            };
            if(unregistered) {
              remove(*f);
            };
          };
        };
      };
      f->release();
    };
  };
  return failures;
}

int SEFiles::Replicate(void) {
  odlog(3)<<"SEFiles::Replicate"<<endl;
  int failures = 0;
  if(files) {
    for(SEFiles::iterator f = files.begin();f!=files.end();++f) {
      // analyze situation
      f->acquire();
      if(f->state_file() == FILE_STATE_REQUESTED) {
        // claim file being downloaded
        odlog(0)<<"Downloading: "<<f->id()<<endl;
        if(!f->state_file(FILE_STATE_DOWNLOADING)) {
          odlog(0)<<"Downloading: "<<f->id()<<" - change state failed"<<endl;
          failures++; f->release(); continue;
        };
        f->release();
        bool downloaded = true;
        SEReplicator replicator;
        odlog(3)<<"Downloading - calling Obtain"<<endl;
        if(replicator.Obtain(*f) != 0) {
          odlog(0)<<"Downloading - Obtain failed"<<endl;
          failures++; downloaded=false;
        };
        if(downloaded) { // validate file if possible
          int res;
          if(f->checksum_available()) {
            res = f->verify();
            if(res == 1) {
              odlog(0)<<"File downloaded but has incomplete (meta)data: "<<f->id()<<endl;
            } else if(res != 0) {
              odlog(0)<<"Failed to validate downloaded file: "<<f->id()<<endl;
            };
          } else { // Can't validate - at least compute checksum
            odlog(0)<<"WARNING: Checksum for file is not known, computing for later usage."<<endl;
            res=f->checksum_compute("md5");
            if(res != 0) {
              odlog(0)<<"Failed to compute checksum."<<endl;
            } else if(f->write_attr() != 0) {
              odlog(0)<<"Failed to write attributes."<<endl;
              res=-1;
            };
          };
          if(res != 0) { // destroy file and go to next
              f->acquire();
              downloaded=false;
              if(!f->state_file(FILE_STATE_DELETING)) {
              odlog(0)<<"Failed to change file state: "<<f->id()<<endl;
            } else {
              f->release();
              continue;
            };
            f->release();
          };
        };
        f->acquire();
        if(downloaded) { // here means validated
          if(!f->state_file(FILE_STATE_VALID)) {
            odlog(0)<<"Downloading: "<<f->id()<<" - change state failed"<<endl;
            failures++; f->release(); continue;
          };
          // Try immediate registration
          if(ns) {
            if(f->state_reg(REG_STATE_REGISTERING)) {
              f->release();
              if(ns->Register(*f) != 0) {
                if(register_retry()) {
                  odlog(0)<<"Failed to register (will retry)"<<endl;
                } else {
                  odlog(0)<<"Failed to register (what to do ?)"<<endl;
                };
                f->state_reg(REG_STATE_LOCAL);
              } else {
                f->state_reg(REG_STATE_ANNOUNCED);
              };
              f->acquire();
            } else {
              if(register_retry()) {
                odlog(0)<<"Failed to set REGISTERING state (will retry)"<<endl;
              } else {
                odlog(0)<<"Failed to set REGISTERING state (what to do?)"<<endl;
              };
            };
          };
        } else {  // download failed
          if(!f->state_file(FILE_STATE_REQUESTED)) {
            odlog(0)<<"Downloading: "<<f->id()<<" - change state failed"<<endl;
            failures++; f->release(); continue;
          };
        };
      };
      f->release();
    };
  };
  return failures;
}

