#include "../../std.h"

extern "C" {
#include <globus_rls_client.h>
#include <globus_rls_rpc.h>
};

#include "../../misc/time_utils.h"
#include "../../misc/inttostring.h"
#include "../../misc/escaped.h"
#include "../../misc/log_time.h"

#include "se_ns_lrc.h"


SENameServerLRC::SENameServerLRC(const char* contact,const char* se_url):SENameServer(contact,se_url),valid(false) {
  for(;;) {
    string url;
    int n = input_escaped_string(contact,url);
    if(n==0) break;
    urls.push_back(url);
    contact+=n;
  };
  valid=true;
}

SENameServerLRC::~SENameServerLRC(void) {

}

int SENameServerLRC::Get(SEAttributes& file) {
  if(!valid) return -1;
  return 0;
}

int SENameServerLRC::Maintain(void) {
  if(!valid) return -1;
  list<string>::iterator u = urls.begin();
  int registered = 0;
  for(;u!=urls.end();++u) {
    string url = *u;
    url.replace(0,3,"rls");
    string lfn("__storage_service__");
    string pfn(se_url());
    int n = pfn.find(':'); if(n != string::npos) pfn.replace(0,n,"se");
    // TODO: https and httpg too
    globus_result_t r = GLOBUS_SUCCESS;
    globus_rls_handle_t *h = NULL;
    char errbuf[MAXERRMSG];
    int rc;
    if((r=globus_rls_client_connect((char*)(url.c_str()),&h)) != 
                                            GLOBUS_SUCCESS) goto error1;
    odlog(3)<<"Mapping: "<<lfn<<" -> "<<pfn<<endl;
    if((r=globus_rls_client_lrc_create(h,(char*)(lfn.c_str()),
                                  (char*)(pfn.c_str()))) != GLOBUS_SUCCESS){
      globus_rls_client_error_info(r,&rc,errbuf,sizeof(errbuf),GLOBUS_FALSE);
      if(rc == GLOBUS_RLS_LFN_EXIST) { // have lfn - try to add pfn only
        if((r=globus_rls_client_lrc_add(h,(char*)(lfn.c_str()),
                                   (char*)(pfn.c_str()))) != GLOBUS_SUCCESS) {
          globus_rls_client_error_info(r,&rc,errbuf,sizeof(errbuf),GLOBUS_FALSE);
          if(rc != GLOBUS_RLS_MAPPING_EXIST) goto error2;
        };
      } else if(rc != GLOBUS_RLS_MAPPING_EXIST) goto error2;
    };
    globus_rls_client_close(h);
    registered++;
    continue;
error1:
    globus_rls_client_error_info(r,NULL,errbuf,sizeof(errbuf),GLOBUS_FALSE);
error2:
    odlog(2)<<"SENameServerLRC::Maintain failed ("<<url<<"): "<<errbuf<<endl;
    globus_rls_client_close(h);
    continue;
  };
  if(registered) return 0;
  odlog(0)<<"SENameServerLRC::Maintain failed: none site accepted registeration"<<endl;
  return -1;
}


static int compare_attr(globus_rls_handle_t *h,const char* key,const char* name,const char* value) {
  globus_result_t r = GLOBUS_SUCCESS;
  globus_list_t* attr_list = GLOBUS_NULL;

  r=globus_rls_client_lrc_attr_value_get(h,(char*)key,(char*)name,
          globus_rls_obj_lrc_lfn,&attr_list);
  if(r != GLOBUS_SUCCESS) {
    int rc;
    char errbuf[MAXERRMSG];
    globus_rls_client_error_info(r,&rc,errbuf,sizeof(errbuf),GLOBUS_FALSE);
    if(rc == GLOBUS_RLS_ATTR_NEXIST) return 1;
    odlog(0)<<"attr_value_get failed: "<<errbuf<<endl;
    return -1;
  };
  if(attr_list == GLOBUS_NULL) return 1;
  // it should be only one value - take first
  globus_rls_attribute_t* attr =
                       (globus_rls_attribute_t*)globus_list_first(attr_list);
  if(attr->type != globus_rls_attr_type_str) { 
    globus_rls_client_free_list(attr_list);
    return -1;
  };
  const char* p = attr->val.s; for(;*p;p++) if(!isspace(*p)) break;
  if(strcmp(value,p) != 0) {
    odlog(0)<<"compare_attr: mismatch: "<<name<<": "<<value<<" != "<<attr->val.s<<endl;
    globus_rls_client_free_list(attr_list);
    return -1;
  };
  globus_rls_client_free_list(attr_list);
  return 0;
}

static int add_attr(globus_rls_handle_t *h,const char* key,const char* name,const char* value) {
  globus_result_t r = GLOBUS_SUCCESS;
  globus_rls_attribute_t attr;
  int rc;
  char errbuf[MAXERRMSG];

  if((r=globus_rls_client_lrc_attr_create(h,(char*)name,
       globus_rls_obj_lrc_lfn,globus_rls_attr_type_str)) != GLOBUS_SUCCESS) {
    globus_rls_client_error_info(r,&rc,errbuf,sizeof(errbuf),GLOBUS_FALSE);
    if(rc != GLOBUS_RLS_ATTR_EXIST) {
      odlog(0)<<"attr_create failed: "<<errbuf<<endl;
      return -1;
    };
  };
  attr.objtype=globus_rls_obj_lrc_lfn;
  attr.type=globus_rls_attr_type_str;
  attr.name=(char*)name; attr.val.s=(char*)value; 
  r=globus_rls_client_lrc_attr_add(h,(char*)key,&attr);
  if(r != GLOBUS_SUCCESS){
    globus_rls_client_error_info(r,&rc,errbuf,sizeof(errbuf),GLOBUS_FALSE);
    if(rc != GLOBUS_RLS_ATTR_EXIST) {
      odlog(0)<<"attr_add failed: "<<errbuf<<endl;
      return -1;
    };
    rc=compare_attr(h,key,name,value);
    if(rc == -1) return -1;
  };
  return 0;
}

static int compare_size(globus_rls_handle_t *h,SEFile& file) {
  string s = inttostring(file.size());
  return compare_attr(h,file.id(),"size",s.c_str());
}

static int compare_checksum(globus_rls_handle_t *h,SEFile& file) {
  if(!file.checksum_available()) return 1;
  return compare_attr(h,file.id(),"checksum",file.checksum().c_str());
}

static int compare_created(globus_rls_handle_t *h,SEFile& file) {
  if(!file.created_available()) return 1;
  string s(""); timetostring(*file.created(),s);
  return compare_attr(h,file.id(),"created",s.c_str());
}
static int add_size(globus_rls_handle_t *h,SEFile& file) {
  string s = inttostring(file.size());
  return add_attr(h,file.id(),"size",s.c_str());
}

static int add_checksum(globus_rls_handle_t *h,SEFile& file) {
  if(!file.checksum_available()) return 0;
  return add_attr(h,file.id(),"checksum",file.checksum().c_str());
}

static int add_created(globus_rls_handle_t *h,SEFile& file) {
  if(!file.created_available()) return 0;
  string s(""); timetostring(*file.created(),s);
  return add_attr(h,file.id(),"created",s.c_str());
}
 

int SENameServerLRC::Register(SEFile& file,bool require_unique) {
  if(!valid) return -1;
  string pfn = se_url();
  int n = pfn.find(':');
  if(n == string::npos) {
    odlog(0)<<"No protocol in url: "<<pfn<<endl;
    return -1;
  };
  pfn.replace(0,n,"se");
  pfn+="?"; pfn+=file.id();
  int registered = 0;
  list<string>::iterator u = urls.begin();
  for(;u!=urls.end();++u) {
    string url = *u;
    url.replace(0,3,"rls");

    globus_result_t r = GLOBUS_SUCCESS;
    globus_rls_handle_t *h = NULL;
    char errbuf[MAXERRMSG];
    int rc;

    if((r=globus_rls_client_connect((char*)(url.c_str()),&h)) != GLOBUS_SUCCESS)
       goto error1;
    odlog(3)<<"Mapping: "<<file.id()<<" -> "<<pfn<<endl;
    r=globus_rls_client_lrc_create(h,(char*)(file.id()),(char*)(pfn.c_str()));
    if(r != GLOBUS_SUCCESS){
      globus_rls_client_error_info(r,&rc,errbuf,sizeof(errbuf),GLOBUS_FALSE);
      if(rc == GLOBUS_RLS_LFN_EXIST) { // have lfn - try to add pfn only
        r=globus_rls_client_lrc_add(h,(char*)(file.id()),(char*)(pfn.c_str()));
        if(r != GLOBUS_SUCCESS) {
          globus_rls_client_error_info(r,&rc,errbuf,sizeof(errbuf),GLOBUS_FALSE);
          if(rc != GLOBUS_RLS_MAPPING_EXIST) goto error2;
        };
      } else if(rc != GLOBUS_RLS_MAPPING_EXIST) goto error2;
    };
  // store metadata
    rc=add_size(h,file); if(rc==-1) goto error3;
    rc=add_checksum(h,file); if(rc==-1) goto error3;
    rc=add_created(h,file); if(rc==-1) goto error3;
    registered++;
    globus_rls_client_close(h);
    continue;
error3:
    odlog(0)<<"SENameServerLRC::Register failed"<<endl;
    if(h) globus_rls_client_close(h);
    continue;
error2:
    odlog(0)<<"SENameServerLRC::Register failed: "<<errbuf<<endl;
    if(h) globus_rls_client_close(h);
    continue;
error1:
    if(r != GLOBUS_SUCCESS) {
      globus_rls_client_error_info(r,NULL,errbuf,sizeof(errbuf),GLOBUS_FALSE);
      odlog(0)<<"SENameServerLRC::Register failed: "<<errbuf<<endl;
    };
    if(h) globus_rls_client_close(h);
    continue;
  };
  if(registered) {
    file.state_reg(REG_STATE_ANNOUNCED);
    return 0;
  };
  return -1;
}

int SENameServerLRC::Unregister(SEFile& file,bool allow_last) {
  if(!valid) return -1;
  string pfn = se_url();
  int n = pfn.find(':');
  if(n == string::npos) {
    odlog(0)<<"No protocol in url: "<<pfn<<endl;
    return -1;
  };
  pfn.replace(0,n,"se");
  pfn+="?"; pfn+=file.id();
  int registered = 0;
  list<string>::iterator u = urls.begin();
  for(;u!=urls.end();++u) {
    string url = *u;
    url.replace(0,3,"rls");

    globus_result_t r = GLOBUS_SUCCESS;
    globus_rls_handle_t *h = NULL;
    char errbuf[MAXERRMSG];
    int rc;

    if((r=globus_rls_client_connect((char*)(url.c_str()),&h)) != GLOBUS_SUCCESS)
       goto error1;

    r=globus_rls_client_lrc_delete(h,(char*)(file.id()),(char*)(pfn.c_str()));
    if(r != GLOBUS_SUCCESS){
      globus_rls_client_error_info(r,&rc,errbuf,sizeof(errbuf),GLOBUS_FALSE);
      if((rc != GLOBUS_RLS_LFN_NEXIST) && 
         (rc != GLOBUS_RLS_MAPPING_NEXIST) && 
         (rc != GLOBUS_RLS_PFN_NEXIST)) goto error2; 
    };
    globus_rls_client_close(h);
    continue;
error2:
    registered++;
    odlog(0)<<"SENameServerLRC::Unregister failed: "<<errbuf<<endl;
    if(h) globus_rls_client_close(h);
    continue;
error1:
    registered++;
    if(r != GLOBUS_SUCCESS) {
      globus_rls_client_error_info(r,NULL,errbuf,sizeof(errbuf),GLOBUS_FALSE);
      odlog(0)<<"SENameServerLRC::Unregister failed: "<<errbuf<<endl;
    };
    if(h) globus_rls_client_close(h);
    continue;
  };
  if(!registered) return 0;
  return -1;
}

int SENameServerLRC::Modify(SEFile& file) {
  return -1;
}

string SENameServerLRC::url(SEAttributes& file) {
  string url = contact();
  url.replace(0,3,"rls");
  url+="/"; url+=file.id();
  return url;  
};

