#include "../../std.h"
#include <dlfcn.h>

#include <fstream>
#ifdef HAVE_SSTREAM
#include <sstream>
#else
#include <strstream>
#endif


#include "../../auth/auth.h"
#include "../../run/run.h"
#include "../../misc/escaped.h"
#include "../../misc/inttostring.h"
#include "../../misc/condition.h"
#include "../../transfer/proxy.h"
#include "../../misc/log_time.h"
#include "../../misc/globus_error_utils.h"
#include "../../config/environment.h"
#include "../../config/config_sections.h"
#include "../../config/daemon.h"

#include "../httpsd.h"

unsigned int conections_processed = 0;

//  Attention - ugly hack

#ifdef BUILD_LOGGER
#define SOAP_ENV__Header LOGGER_SOAP_ENV__Header
#define SOAP_ENV__Code   LOGGER_SOAP_ENV__Code
#define SOAP_ENV__Fault  LOGGER_SOAP_ENV__Fault
#define SOAP_ENV__Detail LOGGER_SOAP_ENV__Detail
#include "../logger.h"
#undef SOAP_ENV__Header
#undef SOAP_ENV__Code
#undef SOAP_ENV__Fault
#undef SOAP_ENV__Detail
#endif

#ifdef BUILD_SE
#define SOAP_ENV__Header SE_SOAP_ENV__Header
#define SOAP_ENV__Code   SE_SOAP_ENV__Code
#define SOAP_ENV__Fault  SE_SOAP_ENV__Fault
#define SOAP_ENV__Detail SE_SOAP_ENV__Detail
#include "../se.h"
#undef SOAP_ENV__Header
#undef SOAP_ENV__Code
#undef SOAP_ENV__Fault
#undef SOAP_ENV__Detail
#endif

static int curr_connections = 0;
static int pre_connections = 2;
static int max_connections = 100;
static int curr_pid = 1;
// static string default_base_url("");
string base_gsi_url("");
string base_ssl_url("");

globus_cond_t conn_cond;
globus_mutex_t conn_lock;

HTTP_Services services;

list<AuthEvaluator*> auths;

globus_io_handle_t s_gsi;
globus_io_handle_t s_ssl;

typedef struct {
  const char* name;
  service_creator      creat;
  service_configurator config;
} service_list_t;

service_item_t service_list_internal[] = {
#ifdef BUILD_LOGGER
 { "logger", &logger_service_creator, &logger_service_configurator },
#endif
#ifdef BUILD_SE
 { "se", &se_service_creator, &se_service_configurator },
#endif
 { NULL, NULL, NULL }
};

list<service_item_t> service_list;

LockSimple service_lock;

void add_services(service_item_t* s_list) {
  for(service_item_t* s=s_list;s->name;++s) {
    list<service_item_t>::iterator p=service_list.insert(service_list.end(),*s);    odlog(0)<<"Loaded service: "<<p->name<<endl;
  };
}

bool add_library(const char* libname) {
  service_lock.block();
  void* handle = dlopen(libname,RTLD_NOW | RTLD_GLOBAL);
  if(handle == NULL) {
    odlog(0)<<"Library '"<<libname<<"' could not be loaded: "<<dlerror()<<endl;
    service_lock.unblock();
    return false;
  };
  service_item_t* s_list = (service_item_t*)dlsym(handle,"service_list");
  if(s_list == NULL) {
    odlog(0)<<"Library '"<<libname<<"' has no services"<<endl;
    dlclose(handle);
    service_lock.unblock();
    return false;
  };
  add_services(s_list);
  service_lock.unblock();
  return true;
}

service_item_t* find_service(const char* name) {
  service_lock.block();
  // look among already loaded services
  list<service_item_t>::iterator s = service_list.begin();
  service_item_t* s_;
  for(;s!=service_list.end();++s) if(strcmp(s->name,name) == 0) break;
  if(s!=service_list.end()) {
    s_=&(*s); service_lock.unblock();
    return s_;
  };
  // look for library with <name>.so
  string libname = string(name)+".so";
  void* handle = dlopen(libname.c_str(),RTLD_NOW | RTLD_GLOBAL);
  if(handle == NULL) {
    // look for library with lib<name>.so
    libname = string("lib")+name+".so";
    handle = dlopen(libname.c_str(),RTLD_NOW | RTLD_GLOBAL);
  };
  if(handle == NULL) {
    odlog(0)<<"Library '"<<libname<<"' could not be loaded: "<<dlerror()<<endl;
    service_lock.unblock();
    return NULL;
  };
  service_item_t* s_list = (service_item_t*)dlsym(handle,"service_list");
  if(s_list == NULL) {
    odlog(0)<<"Library '"<<libname<<"' has no services"<<endl;
    dlclose(handle);
    service_lock.unblock();
    return NULL;
  };
  for(s_=s_list;s_->name;++s_) {
    if(strcmp(s_->name,name) == 0) break;
  };
  if(!(s_->name)) {
    odlog(0)<<"Library '"<<libname<<"' has no service named "<<name<<endl;
    dlclose(handle);
    service_lock.unblock();
    return NULL;
  };
  // Add new services
  s=service_list.end();
  for(service_item_t* s__=s_list;s__->name;++s__) {
    list<service_item_t>::iterator p = 
                  service_list.insert(service_list.end(),*s__);
    odlog(0)<<"Loaded service: "<<p->name<<endl;
    if(s_ == s__) s=p;
  };
  if(s==service_list.end()) {
    service_lock.unblock();
    return NULL;
  };
  s_=&(*s); service_lock.unblock();
  return s_;
}

int configure_common(const char* fname,Daemon& daemon,unsigned int& gsiport,unsigned int& sslport,unsigned int& port) {
  ifstream f(fname);
  if(!f.is_open()) {
    olog<<"Failed to read configuration file "<<fname<<endl;
    return -1;
  };
  if(!central_configuration) {
    for(;!f.eof();) {
      char buf[1024];
      istream_readline(f,buf,sizeof(buf));
      char* p = buf;
      for(;*p;p++) if(!isspace(*p)) break;
      char* command = p;
      if(*command == '#') continue;
      for(;*p;p++) if(isspace(*p)) break;
      int command_len = p-command;
      if(!command_len) continue;
      int r;
      {
        string cmd(command); cmd.resize(command_len);
        string rest(command+command_len);
        r=daemon.config(cmd,rest);
        if(r==0) continue;
        if(r==-1) { f.close(); return -1; };
      };
      if((command_len==7) && (strncmp(command,"gsiport",command_len)==0)) {
        if(gsiport) if(sscanf(command+command_len,"%u",gsiport) != 1) {
          olog<<"Wrong GSI port number in configuration\n";
          f.close(); return 1;
        };
      }
      else if((command_len==7) && (strncmp(command,"sslport",command_len)==0)) {
        if(sslport) if(sscanf(command+command_len,"%u",sslport) != 1) {
          olog<<"Wrong SSL port number in configuration\n";
          f.close(); return 1;
        };
      }
      else if((command_len==4) && (strncmp(command,"port",command_len)==0)) {
        if(port) if(sscanf(command+command_len,"%u",port) != 1) {
          olog<<"Wrong port number in configuration\n";
          f.close(); return 1;
        };
      }
      else {
      };
    };
  } else {
    ConfigSections cf(f);
    cf.AddSection("common");
    cf.AddSection("httpsd");
    for(;;) {
      string rest;
      string command;
      cf.ReadNext(command,rest);
      if((command.length() == 0) && (rest.length() == 0)) break; /* EOF */
      if(cf.SubSection()[0]) continue;
      int r;
      {
        r=daemon.config(command,rest);
        if(r==0) continue;
        if(r==-1) { f.close(); return -1; };
      };
      if(command == "gsiport") {
        if(gsiport) if(sscanf(rest.c_str(),"%u",gsiport) != 1) {
          olog<<"Wrong GSI port number in configuration\n";
          f.close(); return 1;
        };
      }
      else if(command == "sslport") {
        if(sslport) if(sscanf(rest.c_str(),"%u",sslport) != 1) {
          olog<<"Wrong SSL port number in configuration\n";
          f.close(); return 1;
        };
      }
      else if(command == "port") {
        if(port) if(sscanf(rest.c_str(),"%u",port) != 1) {
          olog<<"Wrong port number in configuration\n";
          f.close(); return 1;
        };
      }
      else {
      };
    };
  };
  f.close();
  return 0;
}

int configure_services(const char* fname) {
  ifstream f(fname);
  if(!f.is_open()) {
    olog<<"Failed to read configuration file "<<fname<<endl;
    return -1;
  };
  // Add internal services
  service_lock.block();
  add_services(service_list_internal);
  service_lock.unblock();
  if(!central_configuration) {
    for(;!f.eof();) {
      char buf[1024];
      istream_readline(f,buf,sizeof(buf));
      char* p = buf;
      for(;*p;p++) if(!isspace(*p)) break;
      char* command = p;
      if(*command == '#') continue;
      for(;*p;p++) if(isspace(*p)) break;
      int command_len = p-command;
      if(!command_len) continue;
      {
        string cmd(command); cmd.resize(command_len);
        if(Daemon::skip_config(cmd) == 0) continue;
      };
      if((command_len==5) && (strncmp(command,"group",command_len)==0)) {
        string name("");
        int n = input_escaped_string(p,name);
        if(n==0) {
          odlog(0)<<"Group name is missing"<<endl; f.close(); return -1;
        };
        p+=n;
        AuthEvaluator* auth = new AuthEvaluator(name.c_str());
        if(!auth) {
          odlog(0)<<"Failed to create AuthEvaluator"<<endl; f.close();
          if(auth) delete auth;
          return -1;
        };
        auths.push_back(auth);
        for(;!f.eof();) {
          istream_readline(f,buf,sizeof(buf));
          p = buf;
          for(;*p;p++) if(!isspace(*p)) break;
          command = p;
          if(*command == '#') continue;
          for(;*p;p++) if(isspace(*p)) break;
          command_len = p-command;
          if(!command_len) continue;
          if((command_len==3) && (strncmp(command,"end",command_len)==0)) break;
          auth->add(command);
        };
        odlog(0)<<"Added authorization group: "<<name<<endl;
      } else if((command_len==6) && (strncmp(command,"plugin",command_len)==0)) {
        string libname("");
        int n;
        for(;;) {
          n=input_escaped_string(p,libname);
          if(n==0) break;
          if(libname.length() == 0) continue;
          p+=n;
          add_library(libname.c_str());
        };
      } else if((command_len==7) && (strncmp(command,"service",command_len)==0)) {
        /* service name path */
        string name("");
        int n = input_escaped_string(p,name);
        if(n==0) {
          odlog(0)<<"Service name is missing"<<endl; f.close(); return -1;
        };
        p+=n;
        string path("");
        n = input_escaped_string(p,path);
        if(path[path.length()-1] == '/') path.resize(path.length()-1);
        service_item_t* s = find_service(name.c_str());
        if(!s) {
          odlog(0)<<"Service '"<<name<<"' is not supported"<<endl; f.close();
          return -1;
        };
        HTTP_Service_Properties prop;
        if(s->config(f,path.c_str(),prop)) {
          services.add(path.c_str(),s->creat,prop);
          odlog(0)<<"Added service "<<s->name<<" at "<<path<<endl;
        } else {
          odlog(0)<<"Service '"<<s->name<<"' at "<<path<<" failed to configure"<<endl;
        };
      }
      else if((command_len==7) && (strncmp(command,"gsiport",command_len)==0)) {
      }
      else if((command_len==7) && (strncmp(command,"sslport",command_len)==0)) {
      }
      else if((command_len==4) && (strncmp(command,"port",command_len)==0)) {
      } else {
        *p=0;
        odlog(0)<<"Unknown command: "<<command<<endl; f.close(); return -1;
      };
    };
  } else {
    typedef enum {
      conf_state_single,
      conf_state_group,
      conf_state_service
    } config_state_t;
    config_state_t st = conf_state_single;
    ConfigSections cf(f);
    cf.AddSection("common");
    cf.AddSection("group");
    cf.AddSection("httpsd");
    string group_name;
    list<string> group_rules;
    string service_name;
    string service_path;
    string service_config;
    for(;;) {
      string rest;
      string command;
      cf.ReadNext(command,rest);
      if(cf.SectionNew()) {
        switch(st) {
          case conf_state_group: {
            if(group_name.length() == 0) {
              odlog(0)<<"Group name is missing"<<endl; f.close();
              return -1;
            };
            AuthEvaluator* auth = new AuthEvaluator(group_name.c_str());
            if(!auth) {
              odlog(0)<<"Failed to create AuthEvaluator"<<endl; f.close();
              if(auth) delete auth;
              return -1;
            };
            auths.push_back(auth);
            for(list<string>::iterator a = group_rules.begin();
                                a!=group_rules.end();++a) {
              auth->add(a->c_str());
            };
            odlog(0)<<"Added authorization group: "<<group_name<<endl;
          }; break;
          case conf_state_service: {
            if(service_name.length() == 0) {
              odlog(0)<<"Service name is missing"<<endl; f.close(); return -1;
            };
            if(service_path.length() == 0) {
              odlog(0)<<"Service path/url is missing"<<endl;
              f.close(); return -1;
            };
            if(service_path[service_path.length()-1] == '/') 
                        service_path.resize(service_path.length()-1);
            service_item_t* s = find_service(service_name.c_str());
            if(!s) {
              odlog(0)<<"Service '"<<service_name<<"' is not supported"<<endl; f.close();
              f.close(); return -1;
            };
            /* find service */
            HTTP_Service_Properties prop;
#ifdef HAVE_SSTREAM
            stringstream fake_f(service_config);
#else
            strstream fake_f;
            fake_f<<service_config;
#endif
            if(s->config(fake_f,service_path.c_str(),prop)) {
              services.add(service_path.c_str(),s->creat,prop);
              odlog(0)<<"Added service "<<s->name<<" at "<<service_path<<endl;
            } else {
              odlog(0)<<"Service '"<<s->name<<"' at "<<service_path<<
                                               " failed to configure"<<endl;
            };
          }; break;
        };
        // prepare for new section
        service_name="";
        service_path="";
        service_config="";
        group_name="";
        group_rules.resize(0);
        // 0 - common, 1 - group, 2 - httpsd
        if(cf.SectionNum() == 1) {
          st=conf_state_group;
          group_name=cf.SubSection();
        } else if(cf.SubSection()[0] == 0) {
          st=conf_state_single;
        } else {
          service_name=cf.SubSection();
          st=conf_state_service;
        };
      };
      if((command.length() == 0) && (rest.length() == 0)) break; /* EOF */
      switch(st) {
        case conf_state_single: {
          if(Daemon::skip_config(command) == 0) {
          } else if(command == "gsiport") {
          } else if(command == "sslport") {
          } else if(command == "port") {
          } else if(command == "plugin") {
            string libname("");
            int n;
            const char* p = rest.c_str();
            for(;;) {
              n=input_escaped_string(p,libname);
              if(n==0) break;
              if(libname.length() == 0) continue;
              p+=n;
              add_library(libname.c_str());
            };
          };
        }; break;
        case conf_state_group: {
          if(command == "name") {
            group_name=rest;
          } else {
            rest=command+" "+rest;
            group_rules.push_back(rest);
          };
        }; break;
        case conf_state_service: {
          if(command == "name") {
            service_name=rest;
          } else if(command == "path") {
            service_path=rest;
          } else {
            service_config+=command;
            service_config+=" ";
            service_config+=rest;
            service_config+="\n";
          };
        }; break;
      };
    };
  };
  f.close();
  return 0;
}

#if GLOBUS_IO_VERSION<4
static globus_bool_t auth_cb(void* arg,globus_io_handle_t* h,globus_result_t result,char* identity,gss_ctx_id_t* context_handle) {
#else
static globus_bool_t auth_cb(void* arg,globus_io_handle_t* h,globus_result_t result,char* identity,gss_ctx_id_t context_handle) {
#endif
/*
  {
  gss_ctx_id_desc *context;
  gss_ctx_id_t context_handle;
  SSL_CTX *ctx;
  SSL* s;
  STACK_OF(SSL_CIPHER) *ciphers;
  SSL_CIPHER *cipher;
  char buf[1024];

  context_handle=h->context;
  context =(gss_ctx_id_desc *)context_handle;
  if(context == NULL) {
    cerr<<"!!!!!!!! context is empty"<<endl;
  } else {
    s=context->gss_ssl;
    if(s == NULL) {
      cerr<<"!!!!!!! SSL is empty"<<endl;
    } else {
      cerr<<"!!!!!!! SSL: "<<(unsigned int)s<<endl;
      ciphers=SSL_get_ciphers(s);
      if(ciphers==NULL) {
        ERR_print_errors_fp(stderr);
        printf("SSL_get_ciphers failed\n"); exit(1);
      };
      for(;;) {
        cipher=SKM_sk_pop(SSL_CIPHER,ciphers);
        if(cipher==NULL) break;
        printf("name: %s\n",SSL_CIPHER_get_name(cipher));
        printf("version: %s\n",SSL_CIPHER_get_version(cipher));
        printf("description: %s\n",SSL_CIPHER_description(cipher,buf,sizeof(buf)));
      };
      cipher=SSL_get_current_cipher(s);
      if(cipher) {
        printf("current cipher: %s\n",SSL_CIPHER_get_name(cipher));
      };
    };
  };
  };
*/
  if(result != GLOBUS_SUCCESS) {
    odlog(0)<<"auth failed: "<<result<<endl;
    return GLOBUS_FALSE;
  }; 
  HTTPS_Connector::identity(h,identity);
  return GLOBUS_TRUE;
}

void listen_cb(void *callback_arg,globus_io_handle_t *handle,globus_result_t result) {
  if(result != GLOBUS_SUCCESS) {
    odlog(0)<<"listen failed: "<<result<<endl;
    return;
  };
  HTTPS_Connector* c = NULL;
  if(handle == &s_ssl) {
    c = new HTTPS_Connector(handle,base_ssl_url.c_str(),services,auths);
  } else if(handle == &s_gsi) {
    c = new HTTPS_Connector(handle,base_gsi_url.c_str(),services,auths);
  };
  if(c) {
    c->pid=(curr_pid++);
    if(!(*c)) {
      delete c;
      conections_processed++;
    };
  };
  result=globus_io_tcp_register_listen(handle,&listen_cb,NULL); 
  if(result != GLOBUS_SUCCESS) {
    odlog(0)<<"Failed to listen: "<<result<<endl;
  };
}

#define DEFAULT_GSI_PORT 8000
#define DEFAULT_SSL_PORT 8001
#define DEFAULT_LOG_FILE "/var/log/httpsd.log"
#define DEFAULT_PID_FILE "/var/run/httpsd.pid"

int main(int argc,char* argv[]) {
  unsigned short server_port = 0;
  int handle;
  int handle_ssl;
  struct sockaddr_in myaddr;
  LogTime::level = 1;

  nordugrid_config_loc="httpsd.conf";

  read_env_vars(true);

  string config_file;

  int n;
  Daemon daemon;
  while((n=daemon.getopt(argc,argv,"hp:c:n:")) != -1) {
    switch(n) {
      case ':': { olog<<"Missing argument\n"; return 1; };
      case 'h': {
        cout<<"httpsd [-p port_to_listen] [-c config_file] [-n maximal_connections] "<<daemon.short_help()<<endl;
         return 0;
      };
      case 'p': {
        if(sscanf(optarg,"%hu",&server_port) != 1) {
          olog<<"Wrong port number\n";
          return 1;
        };
      }; break;
      case 'c': {
        config_file=optarg;
      }; break;
      case 'n': {
        if((sscanf(optarg,"%i",&max_connections) != 1) ||
           (max_connections < 0)) {
          olog<<"Wrong number of connections\n";
          return 1;
        };
      }; break;
      default: break;
    };
  };
  if(config_file.length() == 0) {
    if(!central_configuration) {
      config_file=nordugrid_loc+"/etc/httpsd.conf";
    } else {
      config_file="/etc/nordugrid.conf";
    };
//    olog<<"Missing configuration file"<<endl;
//    return 1;
  };
  unsigned int gsiport=0;
  unsigned int sslport=0;
  unsigned int port=0;
  if(configure_common(config_file.c_str(),daemon,gsiport,sslport,port) != 0) {
    return 1;
  };
  if(server_port != 0) {
    gsiport=server_port; sslport=server_port+1;
  };
  if(gsiport == 0) gsiport=DEFAULT_GSI_PORT;
  if(sslport == 0) sslport=DEFAULT_SSL_PORT;

  daemon.logfile(DEFAULT_LOG_FILE);
  daemon.pidfile(DEFAULT_PID_FILE);
  if(daemon.daemon() != 0) {
    perror("daemonization failed");
  };

  // Initialize children signal handling environemnt
  Run run;

  if(globus_module_activate(GLOBUS_IO_MODULE) != GLOBUS_SUCCESS) {
    olog<<"Failed to activate Globus IO module"<<endl;
    return 1;
  };
  char hbuf[1024];
  hbuf[sizeof(hbuf)-1]=0;
  if(globus_libc_gethostname(hbuf,sizeof(hbuf)-1) != 0) {
    strcpy(hbuf,"localhost");
  };

  base_gsi_url=string("httpg://")+hbuf+":"+inttostring(gsiport);
  base_ssl_url=string("https://")+hbuf+":"+inttostring(sslport);

  if(configure_services(config_file.c_str()) != 0) {
    olog<<"Configuration failed. Exiting"<<endl;
    globus_module_deactivate(GLOBUS_IO_MODULE);
    exit(1);
  };

  globus_io_attr_t attr;
  globus_io_secure_authorization_data_t auth;
  globus_result_t res;

  globus_cond_init(&conn_cond,GLOBUS_NULL);
  globus_mutex_init(&conn_lock,GLOBUS_NULL);

  globus_io_secure_authorization_data_initialize(&auth);
  globus_io_secure_authorization_data_set_callback(&auth,&auth_cb,NULL);
  globus_io_tcpattr_init(&attr);
  globus_io_attr_set_secure_authentication_mode(&attr,GLOBUS_IO_SECURE_AUTHENTICATION_MODE_GSSAPI,GSS_C_NO_CREDENTIAL);
  globus_io_attr_set_secure_authorization_mode(&attr,GLOBUS_IO_SECURE_AUTHORIZATION_MODE_CALLBACK,&auth);
  globus_io_attr_set_secure_channel_mode(&attr,GLOBUS_IO_SECURE_CHANNEL_MODE_GSI_WRAP);
  globus_io_attr_set_secure_protection_mode(&attr,GLOBUS_IO_SECURE_PROTECTION_MODE_PRIVATE);
  globus_io_attr_set_secure_delegation_mode(&attr,GLOBUS_IO_SECURE_DELEGATION_MODE_NONE);
  globus_io_attr_set_secure_proxy_mode(&attr,GLOBUS_IO_SECURE_PROXY_MODE_MANY);
  server_port=gsiport;
  res=globus_io_tcp_create_listener(&server_port,-1,&attr,&s_gsi);   
  if(res != GLOBUS_SUCCESS) {
    olog<<"Failed to create GSI listener: "<<res<<endl;
    return 1; 
  };
  odlog(0)<<"Running on: "<<base_gsi_url<<endl;
  globus_io_attr_set_secure_channel_mode(&attr,GLOBUS_IO_SECURE_CHANNEL_MODE_SSL_WRAP);
  globus_io_attr_set_secure_protection_mode(&attr,GLOBUS_IO_SECURE_PROTECTION_MODE_SAFE);  // allow SSL conenction without encryption (decision by client)
  server_port=sslport;
  res=globus_io_tcp_create_listener(&server_port,-1,&attr,&s_ssl);
  if(res != GLOBUS_SUCCESS) {
    olog<<"Failed to create SSL listener: "<<res<<endl;
    return 1; 
  };
  odlog(0)<<"Running on: "<<base_ssl_url<<endl;

  res=globus_io_tcp_register_listen(&s_gsi,&listen_cb,NULL); 
  if(res != GLOBUS_SUCCESS) {
    olog<<"Failed to listen (GSI): "<<res<<endl;
    return 1; 
  };
  res=globus_io_tcp_register_listen(&s_ssl,&listen_cb,NULL); 
  if(res != GLOBUS_SUCCESS) {
    olog<<"Failed to listen (SSL): "<<res<<endl;
    return 1; 
  };

  globus_abstime_t timeout;
  globus_mutex_lock(&conn_lock);
  while(1) {  // forever
    GlobusTimeAbstimeSet(timeout,10,0);
    globus_cond_timedwait(&conn_cond,&conn_lock,&timeout);
//    odlog(0)<<"conections_processed: "<<conections_processed<<endl;
//    if(conections_processed >= 10) break;
//    globus_cond_wait(&conn_cond,&conn_lock);
//    globus_thread_blocking_will_block();
  };
  globus_mutex_unlock(&conn_lock);

  globus_io_close(&s_gsi);
  globus_io_close(&s_ssl);
  globus_io_secure_authorization_data_destroy(&auth);
  globus_io_tcpattr_destroy(&attr);

  globus_cond_destroy(&conn_cond);
  globus_mutex_destroy(&conn_lock);

  globus_module_deactivate(GLOBUS_IO_MODULE);

  return 0;
}

/*
int stat_file(const char* fname,unsigned long long int &size) {
  struct stat64 st;
  if(stat64(fname,&st) != 0) return -1;
  if(!S_ISREG(st.st_mode)) return -1;
  size=st.st_size;
  return 0;
}

int send_file(const char* fname,HTTPS_Connector &c) {
  int h = ::open(fname,O_RDONLY);
  if(h == -1) return 1;
  char buf[1024];
  for(;;) {
    int l = read(h,buf,sizeof(buf));
    if(l == 0) break;
    if(l == -1) { close(h); return 1; };
    if(c.write(buf,l)) { close(h); return 1; };
  };
  close(h);
  return 0;
}
*/
