/*
  Filename: states.cc
  keeps list of states
  acts on states
*/

#include "../std.h"
#include "../files/info_files.h"
#include "../rsl/parse_rsl.h"
#include "../run/run_parallel.h"
#include "../config/environment.h"
#include "../misc/inttostring.h"
#include "../misc/stringtoint.h"
#include "../mail/send_mail.h"
#include "../misc/log_time.h"
#include "../misc/job_log.h"
#include "../config/config_file.h"
#include "../jobs/users.h"
#include "../jobs/job.h"
#include "../jobs/plugins.h"
#include "states.h"


long int JobsList::jobs_num[JOB_STATE_NUM] = { 0, 0, 0, 0, 0, 0, 0, 0 };
long int JobsList::max_jobs_processing=DEFAULT_MAX_JOBS;
long int JobsList::max_jobs_processing_emergency=1;
long int JobsList::max_jobs_running=-1;
long int JobsList::max_jobs=-1;
int JobsList::max_downloads=-1;
unsigned long long int JobsList::min_speed=0;
time_t JobsList::min_speed_time=300;
unsigned long long int JobsList::min_average_speed=0;
time_t JobsList::max_inactivity_time=300;
bool JobsList::use_secure_transfer=false; /* secure data transfer is OFF by default !!! */
bool JobsList::use_local_transfer=false;
bool JobsList::cache_registration=false;
unsigned int JobsList::wakeup_period = 120; // default wakeup every 3 min.


#ifdef NO_GLOBUS_CODE
ContinuationPlugins::ContinuationPlugins(void) { };
ContinuationPlugins::~ContinuationPlugins(void) { };
bool ContinuationPlugins::add(const char* state,unsigned int timeout,const char* command) { return true; };
bool ContinuationPlugins::add(job_state_t state,unsigned int timeout,const char* command) { return true; };
bool ContinuationPlugins::add(const char* state,const char* options,const char* command) { return true; };
bool ContinuationPlugins::add(job_state_t state,const char* options,const char* command) { return true; };
ContinuationPlugins::action_t ContinuationPlugins::run(const JobDescription &job,const JobUser& user,int& result,string& response) { return act_pass; };
void RunPlugin::set(const string& cmd) { };
void RunPlugin::set(char const * const * args) { };
#endif

JobsList::JobsList(JobUser &user,ContinuationPlugins &plugins) {
  this->user=&user;
  this->plugins=&plugins;
  jobs.clear();
}

JobsList::~JobsList(void){
}

JobsList::iterator JobsList::FindJob(const JobId &id){
  iterator i;
  for(i=jobs.begin();i!=jobs.end();++i) {
    if((*i) == id) break;
  };
  return i;
}

bool JobsList::AddJobNoCheck(const JobId &id,uid_t uid,gid_t gid){
  iterator i;
  return AddJobNoCheck(id,i,uid,gid);
}

bool JobsList::AddJobNoCheck(const JobId &id,JobsList::iterator &i,uid_t uid,gid_t gid){
  i=jobs.insert(jobs.end(),JobDescription(id,user->SessionRoot() + "/" + id));
  i->keep_finished=user->KeepFinished();
  i->keep_deleted=user->KeepDeleted();
  i->set_uid(uid,gid);
  return true;
}

bool JobsList::AddJob(const JobId &id,uid_t uid,gid_t gid){
  /* jobs should be unique */
  if(FindJob(id) != jobs.end()) return false;
  olog << id << ": Added" << endl;
  iterator i=jobs.insert(jobs.end(),
         JobDescription(id,user->SessionRoot() + "/" + id));
  i->keep_finished=user->KeepFinished();
  i->keep_deleted=user->KeepDeleted();
  i->set_uid(uid,gid);
  return true;
}

bool JobsList::AddJob(JobUser &user,const JobId &id,uid_t uid,gid_t gid){
  if((&user) != NULL) {
    if((this->user) == NULL) { this->user = &user; }
    else {
      if(this->user != &user) { /* incompatible user */
        return false;
      };
    };
  };
  return AddJob(id,uid,gid);
};

#ifndef NO_GLOBUS_CODE

bool JobsList::ActJob(const JobId &id,bool hard_job)  {
  iterator i=FindJob(id);
  if(i == jobs.end()) return false;
  return ActJob(i,hard_job);
}

bool JobsList::ActJobs(bool hard_job) {
  bool res = true;
  bool once_more = false;
  bool postpone_preparing = false;
  bool postpone_finishing = false;
  if((max_jobs_processing != -1) && 
     (!use_local_transfer) && 
     ((JOB_NUM_PROCESSING*3) > (max_jobs_processing*2))) {
    if(JOB_NUM_PREPARING > JOB_NUM_FINISHING) { 
      postpone_preparing=true; 
    } else if(JOB_NUM_PREPARING < JOB_NUM_FINISHING) {
      postpone_finishing=true;
    };
  };
  // first pass - optionally skipping some states
  for(iterator i=jobs.begin();i!=jobs.end();) {
    if(i->job_state == JOB_STATE_UNDEFINED) { once_more=true; }
    else if(((i->job_state == JOB_STATE_ACCEPTED) && postpone_preparing) ||
            ((i->job_state == JOB_STATE_INLRMS) && postpone_finishing)  ) {
      once_more=true;
      i++; continue;
    };
    res &= ActJob(i,hard_job);
  };
  // second pass - process skipped states and new jobs
  if(once_more) for(iterator i=jobs.begin();i!=jobs.end();) {
    res &= ActJob(i,hard_job);
  };
  return res;
}

bool JobsList::DestroyJobs(bool finished,bool active) {
  bool res = true;
  for(iterator i=jobs.begin();i!=jobs.end();) {
    res &= DestroyJob(i,finished,active);
  };
  return res;
}

/* returns false if had to run external process */
bool JobsList::DestroyJob(JobsList::iterator &i,bool finished,bool active) {
  olog << i->job_id << ": Destroying" <<  endl;
  job_state_t new_state=i->job_state;
  if(new_state == JOB_STATE_UNDEFINED) {
    if((new_state=job_state_read_file(i->job_id,*user))==JOB_STATE_UNDEFINED) {
      olog << i->job_id << ": Can't read state - no comments, just cleaning" << endl;
      job_clean_final(i->job_id,*user);
      if(i->local) { delete i->local; }; i=jobs.erase(i);
      return true;
    };
  };
  i->job_state = new_state;
  if((new_state == JOB_STATE_FINISHED) && (!finished)) { ++i; return true; };
  if(!active) { ++i; return true; };
  if((new_state != JOB_STATE_INLRMS) || 
     (job_lrms_mark_check(i->job_id,*user))) {
    olog << i->job_id << ": Cleaning control and session directories" << endl;
    job_clean_final(i->job_id,*user);
    if(i->local) { delete i->local; }; i=jobs.erase(i);
    return true;
  };
  olog << i->job_id << ": This job may be still running - canceling" << endl;
  bool state_changed = false;
  if(!state_submiting(i,state_changed,true)) {
    olog << i->job_id << ": Cancelation failed (probably job finished) - cleaning anyway" << endl;
    job_clean_final(i->job_id,*user);
    if(i->local) { delete i->local; }; i=jobs.erase(i);
    return true;
  };
  if(!state_changed) { ++i; return false; }; /* child still running */
  olog << i->job_id << ": Cancelation probably succeeded - cleaning" << endl;
  job_clean_final(i->job_id,*user);
  if(i->local) { delete i->local; };
  i=jobs.erase(i);
  return true;
}

/* do processing necessary in case of failure */
bool JobsList::FailedJob(const JobsList::iterator &i) {
  /* put mark - failed */
  if(!job_failed_mark_put(*i,*user,i->failure_reason)) return false;
  /* make all output files unuploadable */
  list<FileData> fl;
  if(!job_output_read_file(i->job_id,*user,fl)) return true; /* no file - no error */
  for(list<FileData>::iterator ifl=fl.begin();ifl!=fl.end();++ifl) {
    ifl->lfn="";
  };
  if(!job_output_write_file(*i,*user,fl)) return false;
  if(!(i->local)) {
    JobLocalDescription *job_desc = new JobLocalDescription;
    if(!job_local_read_file(i->job_id,*user,*job_desc)) {
      olog<<i->job_id<<": Failed reading local information."<<endl;
      delete job_desc;
    }
    else {
      i->local=job_desc;
    };
  };
  if(i->local) {
    i->local->uploads=0;
    job_local_write_file(*i,*user,*(i->local));
  };
  return true;
}

bool JobsList::GetLocalDescription(JobsList::iterator &i) {
  if(!i->GetLocalDescription(*user)) {
    olog << i->job_id << ": Failed reading local information." << endl;
    return false;
  };
  return true;
}

bool JobsList::state_submiting(const JobsList::iterator &i,bool &state_changed,bool cancel) {
  if(i->child == NULL) {
    /* no child was running yet, or recovering from fault */
    /* write grami file for globus-script-X-submit */
    JobLocalDescription* job_desc;
    if(i->local) { job_desc=i->local; }
    else { job_desc=new JobLocalDescription; };
    if(i->local == NULL) {
      if(!job_local_read_file(i->job_id,*user,*job_desc)) {
        olog << i->job_id << ": Failed reading local information." << endl;
        if(!cancel) i->AddFailure("Internal error: can't read local file");
        delete job_desc;
        return false;
      };
      i->local=job_desc;
    };
    if(!cancel) { olog<<i->job_id<<": state SUBMITING: starting child"<<endl; }
    else { olog<<i->job_id<<": state CANCELING: starting child"<<endl; };
    if(!cancel) {  /* in case of cancel all preparations are already done */
      const char *local_transfer_s = NULL;
      if(use_local_transfer) { 
        local_transfer_s="joboption_localtransfer=yes";
      };
      if(!write_grami(*i,*user,i->SessionDir(),*job_desc,local_transfer_s)) {
        olog << i->job_id << ": Failed creating grami file." << endl; return false;
      };
      if(!set_execs(i->job_id,*user,i->SessionDir())) {
        olog << i->job_id << ": Failed setting executable permissions." << endl; return false;
      };
      /* precreate file to store diagnostics from lrms */
      job_diagnostics_mark_put(*i,*user);
      job_lrmsoutput_mark_put(*i,*user);
    };
    /* submit/cancel job to LRMS using submit/cancel-X-job */
    string cmd;
    if(cancel) { cmd=nordugrid_libexec_loc+"/cancel-"+job_desc->lrms+"-job"; }
    else { cmd=nordugrid_libexec_loc+"/submit-"+job_desc->lrms+"-job"; };
    string grami = user->ControlDir()+"/job."+(*i).job_id+".grami";
    char* args[3] ={ (char*)cmd.c_str(), (char*)grami.c_str(), NULL };
    job_errors_mark_put(*i,*user);
    if(!RunParallel::run(*user,*i,args,&(i->child))) {
      if(!cancel) {
        i->AddFailure("Failed initiating job submission to LRMS");
        olog << i->job_id << ": Failed running submission process." << endl;
      } else {
        olog << i->job_id << ": Failed running cancel process." << endl;
      };
      return false;
    };
    return true;
  }
  else {
    /* child was run - check exit code */
    if(i->child->get_exit_code() == -1) { 
      /* child is running - come later */
      return true;
    };
    if(!cancel) { olog<<i->job_id<<": state SUBMITING: child exited"<<endl; }
    else { olog<<i->job_id<<": state CANCELING: child exited"<<endl; };
    if(cancel) job_diagnostics_mark_move(*i,*user);
    if(i->child->get_exit_code() != 0) { 
      if(!cancel) { olog<<i->job_id<<": Job submission to LRMS failed."<<endl; }
      else { olog<<i->job_id<<": Failed to cancel running job."<<endl; };
      RunParallel::release(i->child); i->child=NULL;
      if(!cancel) i->AddFailure("Job submission to LRMS failed");
      return false;
    };
    RunParallel::release(i->child); i->child=NULL;
    if(!cancel) {
      /* success code - get LRMS job id */
      string local_id=read_grami(i->job_id,*user,i->SessionDir());
      if(local_id.length() == 0) {
        olog<<i->job_id<<": Failed obtaining lrms id."<<endl;
        i->AddFailure("Failed extracting LRMS ID due to some internal error");
        return false;
      };
      /* put id into local information file */
      if(!i->GetLocalDescription(*user)) {
        olog << i->job_id << ": Failed reading local information." << endl;
        i->AddFailure("Internal error");
        return false;
      };   
      /*
      JobLocalDescription *job_desc;
      if(i->local) { job_desc=i->local; }
      else { job_desc=new JobLocalDescription; };
      if(i->local == NULL) {
        if(!job_local_read_file(i->job_id,*user,*job_desc)) {
          olog << i->job_id << ": Failed reading local information." << endl;
          i->AddFailure("Internal error");
          delete job_desc; return false;
        };
        i->local=job_desc;
      };
      */
      i->local->localid=local_id;
      if(!job_local_write_file(*i,*user,*(i->local))) {
        i->AddFailure("Internal error");
        olog<<i->job_id<<": Failed writing local information."<<endl;
        return false;
      };
    };
    /* move to next state */
    state_changed=true;
    return true;
  };
}

bool JobsList::state_loading(const JobsList::iterator &i,bool &state_changed,bool up) {
  /* RSL was analyzed/parsed - now run child process downloader
     to download job input files and to wait for user uploaded ones */
  if(i->child == NULL) { /* no child started */
    odlog(3)<<i->job_id<<": state: PREPARING/FINISHING: starting new child"<<endl;
    /* no child was running yet, or recovering from fault */
    /* run it anyway and exit code will give more inforamtion */
    bool switch_user = true;
    string cmd; 
    if(up) { cmd=nordugrid_libexec_loc+"/uploader"; }
    else { cmd=nordugrid_libexec_loc+"/downloader"; };
    if(!(user->CachePrivate())) { switch_user=false; };
    uid_t user_id = user->get_uid();
    if(user_id == 0) user_id=i->get_uid();
    string user_id_s = inttostring(user_id);
    string max_files_s;
    int argn=3;
    char* args[16] = {
      (char*)(cmd.c_str()),
      "-U",
      (char*)(user_id_s.c_str()),
      NULL,
      NULL,
      NULL,
      NULL,
      NULL,
      NULL,
      NULL,
      NULL,
      NULL,
      NULL,
      NULL,
      NULL,
      NULL
    };
    if(JobsList::max_downloads > 0) {
      max_files_s=inttostring(JobsList::max_downloads);
      args[argn]="-n"; argn++;
      args[argn]=(char*)(max_files_s.c_str()); argn++;
    };
    if(!use_secure_transfer) { 
      args[argn]="-c"; argn++;
    };
    if(use_local_transfer) { 
      args[argn]="-l"; argn++;
    };
    if(central_configuration) { 
      args[argn]="-Z"; argn++;
    };
    args[argn]=(char*)(i->job_id.c_str()); argn++;
    args[argn]=(char*)(user->ControlDir().c_str()); argn++;
    args[argn]=(char*)(i->SessionDir().c_str()); argn++;
    if(user->CacheDir().length() != 0) {
      args[argn]=(char*)(user->CacheDir().c_str()); argn++;
      if(user->CacheDataDir().length() != 0) {
        args[argn]=(char*)(user->CacheDataDir().c_str()); argn++;
      }
      else {
        args[argn]=(char*)(user->CacheDir().c_str()); argn++;
      };
      if(user->CacheLinkDir().length() != 0) {
        args[argn]=(char*)(user->CacheLinkDir().c_str()); argn++;
      };
    };
    if(!up) { olog<<i->job_id<<": State PREPARING: starting child: "<<args[0]<<endl; }
    else { olog<<i->job_id<<": State FINISHING: starting child: "<<args[0]<<endl; };
    job_errors_mark_put(*i,*user);
    job_restart_mark_remove(i->job_id,*user);
    if(!RunParallel::run(*user,*i,args,&(i->child),switch_user)) {
      olog << i->job_id << ": Failed starting down/uploader process." << endl;
      return false;
    };
    return true;
  } 
  else {
    if(i->child->get_exit_code() == -1) {
      odlog(3)<<i->job_id<<": State: PREPARING/FINISHING: child is running"<<endl;
      /* child is running - come later */
      return true;
    };
    /* child was run - check exit code */
    if(!up) { odlog(3)<<i->job_id<<": State: PREPARING: child exited with code: "<<i->child->get_exit_code()<<endl; }
    else { odlog(3)<<i->job_id<<": State: FINISHING: child exited with code: "<<i->child->get_exit_code()<<endl; };
    if(i->child->get_exit_code() != 0) { 
      if(i->child->get_exit_code() == 1) { 
        /* unrecoverable failure detected - all we can do is to kill the job */
        if(up) {
          olog<<i->job_id<<": State: FINISHING unrecoverable error detected"<<endl;
          i->AddFailure("Failed in files upload (post-processing)");
        } else {
          olog<<i->job_id<<": State: PREPARING: unrecoverable error detected"<<endl;
          i->AddFailure("Failed in files download (pre-processing)");
        };
      } else if(i->child->get_exit_code() == 3) {
        if(up) {
          olog<<i->job_id<<": State: FINISHING: credentials probably expired"<<endl;
          JobFailStateRemember(i,JOB_STATE_FINISHING);
          i->AddFailure("Failed in files upload due to expired credentials - try to renew");
        } else {
          olog<<i->job_id<<": State: PREPARING: credentials probably expired"<<endl;
          JobFailStateRemember(i,JOB_STATE_PREPARING);
          i->AddFailure("Failed in files download due to expired credentials - try to renew");
        };
      } else {
        if(up) {
          olog<<i->job_id<<": State: FINISHING: some kind of recoverable error detected. Recover from such type of errors is not supported yet."<<endl;
          i->AddFailure("Failed in files upload (post-processing)");
        } else {
          olog<<i->job_id<<": State: PREPARING: some kind of recoverable error detected. Recover from such type of errors is not supported yet."<<endl;
          i->AddFailure("Failed in files download (pre-processing)");
        };
      };
      RunParallel::release(i->child); i->child=NULL;
      return false;
    };
    /* success code - move to next state */
    state_changed=true;
    RunParallel::release(i->child); i->child=NULL;
    return true;
  };
}

bool JobsList::JobPending(JobsList::iterator &i) {
  if(i->job_pending) return true;
  i->job_pending=true; 
  return job_state_write_file(*i,*user,i->job_state,true);
}

job_state_t JobsList::JobFailStateGet(const JobsList::iterator &i) {
  if(!(i->local)) {
    JobLocalDescription *job_desc = new JobLocalDescription;
    if(!job_local_read_file(i->job_id,*user,*job_desc)) {
      olog<<i->job_id<<": Failed reading local information."<<endl;
      delete job_desc; return JOB_STATE_UNDEFINED;
    }
    else {
      i->local=job_desc;
    };
  };
  if(i->local->failedstate.length() == 0) { return JOB_STATE_UNDEFINED; };
  for(int n = 0;states_all[n].name != NULL;n++) {
    if(!strcmp(states_all[n].name,i->local->failedstate.c_str())) {
      i->local->failedstate="";
      job_local_write_file(*i,*user,*(i->local));
      return states_all[n].id;
    };
  };
  i->local->failedstate="";
  job_local_write_file(*i,*user,*(i->local));
  return JOB_STATE_UNDEFINED;
}

bool JobsList::RecreateTransferLists(const JobsList::iterator &i) {
  list<FileData> fl_old;
  list<FileData> fl_new;
  // keep current list
  if(!job_output_read_file(i->job_id,*user,fl_old)) {
    olog<<i->job_id<<": Failed to read output list."<<endl;
    return false;
  };
  // recreate list by reprocessing RSL 
  JobLocalDescription job_desc; // palceholder
  if(!process_rsl(*user,*i,job_desc)) {
    olog<<i->job_id<<": Reprocessing RSL failed."<<endl;
    return false;
  };
  if(!job_output_read_file(i->job_id,*user,fl_new)) {
    olog<<i->job_id<<": Failed to read reprocessed output list."<<endl;
    return false;
  };
  // remove uploaded files
  for(list<FileData>::iterator i_new = fl_new.begin();
                                   i_new!=fl_new.end();) {
    if(!(i_new->has_lfn())) { ++i_new; continue; }; // user file - keep
    list<FileData>::iterator i_old = fl_old.begin();
    for(;i_old!=fl_old.end();++i_old) {
      if((*i_new) == (*i_old)) break;
    };
    if(i_old == fl_old.end()) { ++i_new; continue; };
    i_new=fl_new.erase(i_new);
    fl_old.erase(i_old);
  };
  if(!job_output_write_file(*i,*user,fl_new)) return false;
  return true;
}

bool JobsList::JobFailStateRemember(const JobsList::iterator &i,job_state_t state) {
  if(!(i->local)) {
    JobLocalDescription *job_desc = new JobLocalDescription;
    if(!job_local_read_file(i->job_id,*user,*job_desc)) {
      olog<<i->job_id<<": Failed reading local information."<<endl;
      delete job_desc; return false;
    }
    else {
      i->local=job_desc;
    };
  };
  if(state == JOB_STATE_UNDEFINED) {
    i->local->failedstate="";
    return job_local_write_file(*i,*user,*(i->local));
  };
  if(i->local->failedstate.length() == 0) {
    i->local->failedstate=states_all[state].name;
    return job_local_write_file(*i,*user,*(i->local));
  };
  return true;
}

/* Do job's processing: check&change state, run necessary external
   programs, do necessary things. Also advance pointer and/or delete
   slot if necessary */
bool JobsList::ActJob(JobsList::iterator &i,bool hard_job) {
  bool once_more     = true;
  bool delete_job    = false;
  bool job_error     = false;
  bool state_changed = false;
  job_state_t old_state = i->job_state;
  bool old_pending = i->job_pending;
  while(once_more) {
    once_more     = false;
    delete_job    = false;
    job_error     = false;
    state_changed = false;
 /* some states can not be canceled (or there is no sense to do that) */
/*
       (i->job_state != JOB_STATE_FINISHING) &&
*/
    if((i->job_state != JOB_STATE_CANCELING) &&
       (i->job_state != JOB_STATE_FINISHED)) {
      if(job_cancel_mark_check(i->job_id,*user)) {
        olog<<i->job_id<<": Canceling job ("<<user->UnixName()<<") because of user request"<< endl;
        /* kill running child */
        if(i->child) { 
          i->child->kill();
          RunParallel::release(i->child);
          i->child=NULL;
        };
        /* put some explanation */
        i->AddFailure("User requested to cancel the job");
        /* behave like if job failed */
        if(!FailedJob(i)) {
          /* DO NOT KNOW WHAT TO DO HERE !!!!!!!!!! */
        };
        /* special processing for INLRMS case */
        if(i->job_state == JOB_STATE_INLRMS) {
          i->job_state = JOB_STATE_CANCELING;
        }
        else if(i->job_state == JOB_STATE_FINISHING) {
          i->job_state = JOB_STATE_FINISHED;
        }
        else {
          i->job_state = JOB_STATE_FINISHING;
        };
        job_cancel_mark_remove(i->job_id,*user);
        state_changed=true;
        once_more=true;
      };
    };
    if(!state_changed) switch(i->job_state) {
    /* undefined state - not actual state - job was just added but
       not analyzed yet */
      case JOB_STATE_UNDEFINED: { /* read state from file */
        /* undefined means job just detected - read it's status */
        /* but first check if it's not too many jobs in system  */
        if((JOB_NUM_ACCEPTED < max_jobs) || (max_jobs == -1)) {
          job_state_t new_state=job_state_read_file(i->job_id,*user);
          if(new_state == JOB_STATE_UNDEFINED) { /* something failed */
            olog<<i->job_id<<": Reading status of new job failed."<<endl;
            job_error=true; i->AddFailure("Failed reading status of the job");
            break;
          };
          //  By keeping once_more==false jobs does not cycle here but
          // goes out and registers it's state in counters. This allows
          // to maintain limits properly after restart. Except FINISHED
          // jobs because they are not kept in memory and should be 
          // processed immediately.
          i->job_state = new_state; /* this can be any state, if we are
                                         recovering after failure */
          if(new_state == JOB_STATE_ACCEPTED) {
            // parse request (do it here because any other processing can 
            // read 'local' and then we never know if it was new job)
            JobLocalDescription *job_desc;
            job_desc = new JobLocalDescription;
            job_desc->sessiondir=i->session_dir;
            /* first phase of job - just  accepted - parse request */
            olog<<i->job_id<<": State: ACCEPTED: parsing RSL"<<endl;
            if(!process_rsl(*user,*i,*job_desc)) {
              olog<<i->job_id<<": Processing RSL failed."<<endl;
              job_error=true; i->AddFailure("Could not process RSL");
              delete job_desc;
              break; /* go to next job */
            };
            i->local=job_desc;
            // prepare information for logger
            job_log.make_file(*i,*user);
          } else if(new_state == JOB_STATE_FINISHED) {
            once_more=true;
          } else if(new_state == JOB_STATE_DELETED) {
            once_more=true;
          } else {
            olog<<i->job_id<<": "<<JobDescription::get_state_name(new_state)<<
                ": New job belongs to "<<
                i->get_uid()<<"/"<<i->get_gid()<< endl;
            // Make it clean state after restart
            job_state_write_file(*i,*user,i->job_state);
          };
        }; // Not doing JobPending here because that job kind of does not exist.
      }; break;
  /* accepted state - job was just accepted by jobmager-ng and we already
     know that it is accepted - now we are analyzing/parsing request */
  /* or it can also happen we are waiting for user specified time */
      case JOB_STATE_ACCEPTED: {
        odlog(3)<<i->job_id<<": State: ACCEPTED"<<endl;
        if(!i->GetLocalDescription(*user)) {
          olog << i->job_id << ": Failed reading local information." << endl;
          job_error=true; i->AddFailure("Internal error");
          break; /* go to next job */
        };
        if(i->local->dryrun) {
          olog<<i->job_id<<": State: ACCEPTED: dryrun"<<endl;
          job_error=true; 
          i->AddFailure("User requested dryrun. Job skiped.");
        };
        if((max_jobs_processing == -1) ||
           (use_local_transfer) ||
           (i->local->downloads == 0) ||
           (JOB_NUM_PROCESSING < max_jobs_processing) ||
           ((JOB_NUM_FINISHING >= max_jobs_processing) && 
            (JOB_NUM_PREPARING < max_jobs_processing_emergency))) {
          /* check for user specified time */
          if(i->local->processtime.defined()) {
            olog<<i->job_id<<": State: ACCEPTED: have processtime "<<
                  i->local->processtime<<endl;
            if((i->local->processtime) <= time(NULL)) {
              odlog(3)<<i->job_id<<": State: ACCEPTED: moving to PREPARING"<<endl;
              state_changed=true; once_more=true;
              i->job_state = JOB_STATE_PREPARING;
            };
          }
          else {
            odlog(3)<<i->job_id<<": State: ACCEPTED: moving to PREPARING"<<endl;
            state_changed=true; once_more=true;
            i->job_state = JOB_STATE_PREPARING;
          };
        } else JobPending(i);
      }; break;
/* preparing state - means job is parsed and we are going to download or
   already downloading input files. process downloader is run for
   that. it also checks for files user interface have to upload itself */
      case JOB_STATE_PREPARING: {
        odlog(3)<<i->job_id<<": State: PREPARING"<<endl;
        if(i->job_pending || state_loading(i,state_changed,false)) {
          if(i->job_pending || state_changed) {
            if((JOB_NUM_RUNNING<max_jobs_running) || (max_jobs_running==-1)) {
              i->job_state = JOB_STATE_SUBMITING;
              state_changed=true; once_more=true;
            } else {
              state_changed=false;
              JobPending(i);
            };
          };
        } else {
          i->AddFailure("Failed to run downloader (preprocessing)");
          job_error=true;
        };
      }; break;
 /* state submitting - everything is ready for submission - so run submission */
      case JOB_STATE_SUBMITING: {
        odlog(3)<<i->job_id<<": State: SUBMITTING"<<endl;
        if(state_submiting(i,state_changed)) {
          if(state_changed) {
            i->job_state = JOB_STATE_INLRMS;
            once_more=true;
          };
        }
        else { job_error=true; };
      }; break;
 /* This state is like submitting, only -rm instead of -submit */
      case JOB_STATE_CANCELING: {
        odlog(3)<<i->job_id<<": State: CANCELING"<<endl;
        if(state_submiting(i,state_changed,true)) {
          if(state_changed) {
            i->job_state = JOB_STATE_FINISHING;
            once_more=true;
          };
        }
        else { job_error=true; };
      }; break;
      case JOB_STATE_INLRMS: {
        odlog(3)<<i->job_id<<": State: INLRMS"<<endl;
        if(!i->GetLocalDescription(*user)) {
          olog << i->job_id << ": Failed reading local information." << endl;
          job_error=true;
          i->AddFailure("Failed reading local job information");
          return false;
        };
        if(i->job_pending || job_lrms_mark_check(i->job_id,*user)) {
          if(!i->job_pending) {
            olog<<i->job_id<<": Job finished."<<endl;
            job_diagnostics_mark_move(*i,*user);
            LRMSResult ec = job_lrms_mark_read(i->job_id,*user);
            if(ec.code() != 0) {
              odlog(2)<<i->job_id<<": State: INLRMS: exit message is "<<ec<<endl;
              /* check if not asked to rerun job */
              JobLocalDescription *job_desc = i->local;
              if(job_desc->reruns > 0) { /* rerun job once more */
                job_desc->reruns--;
                job_desc->localid="";
                job_local_write_file(*i,*user,*job_desc);
                job_lrms_mark_remove(i->job_id,*user);
                olog<<i->job_id<<": State: INLRMS: job restarted"<<endl;
                i->job_state = JOB_STATE_SUBMITING; 
                // INLRMS slot is already taken by this job, so resubmission
                // can be done without any checks
              } else {
                i->AddFailure("LRMS error: ("+
                      inttostring(ec.code())+") "+ec.description());
                job_error=true;
                i->job_state = JOB_STATE_FINISHING;
                // This does not require any special postprocessing and
                // can go to next state directly
              };
              state_changed=true; once_more=true;
              break;
            } else {
              // i->job_state = JOB_STATE_FINISHING;
            };
          };
          if((max_jobs_processing == -1) ||
             (use_local_transfer) ||
             (i->local->uploads == 0) ||
             (JOB_NUM_PROCESSING < max_jobs_processing) ||
             ((JOB_NUM_PREPARING >= max_jobs_processing) &&
              (JOB_NUM_FINISHING < max_jobs_processing_emergency))) {
            state_changed=true; once_more=true;
            i->job_state = JOB_STATE_FINISHING;
          } else JobPending(i);
        };
      }; break;
      case JOB_STATE_FINISHING: {
        odlog(3)<<i->job_id<<": State: FINISHING"<<endl;
        if(state_loading(i,state_changed,true)) {
          if(state_changed) {
            i->job_state = JOB_STATE_FINISHED;
            once_more=true; hard_job=true;
          };
        }
        else {
          i->job_state = JOB_STATE_FINISHED;
          state_changed=true; /* to send mail */
          once_more=true; hard_job=true;
          i->AddFailure("Failed to run uploader (postprocessing)");
          job_error=true;
        };
      }; break;
      case JOB_STATE_FINISHED: {
        if(job_clean_mark_check(i->job_id,*user)) {
          olog<<i->job_id<<": Job is requested to clean - deleting."<<endl;
          /* delete everything */
          job_clean_final(i->job_id,*user);
        } else {
          if(job_restart_mark_check(i->job_id,*user)) { 
            job_restart_mark_remove(i->job_id,*user); 
            /* request to rerun job - check if can */
            job_state_t state_ = JobFailStateGet(i);
            if(state_ == JOB_STATE_PREPARING) {
              job_failed_mark_remove(i->job_id,*user);
              // state_changed=true;
              i->job_state = JOB_STATE_ACCEPTED; // This will recreate lists
              i->job_pending = true;
              JobPending(i); // make it go to end of state immediately
              break;
            } else if(state_ == JOB_STATE_FINISHING) {
              if(RecreateTransferLists(i)) {
                job_failed_mark_remove(i->job_id,*user);
                // state_changed=true;
                i->job_state = JOB_STATE_INLRMS;
                JobPending(i); // make it go to end of state immediately
                break;
              };
            };
          };
          if(hard_job) { /* try to minimize load */
            time_t t = -1;
            if(!job_local_read_cleanuptime(i->job_id,*user,t)) {
              /* must be first time - create cleanuptime */
              JobLocalDescription job_desc;
              /* read lifetime - if empty it wont be overwritten */
              job_local_read_file(i->job_id,*user,job_desc);
              if(!stringtoint(job_desc.lifetime,t)) t = i->keep_finished;
              if(t > i->keep_finished) t = i->keep_finished;
              time_t last_changed=job_state_time(i->job_id,*user);
              t=last_changed+t; job_desc.cleanuptime=t;
              job_local_write_file(*i,*user,job_desc);
            };
            /* check if it is not time to remove that job completely */
            if((time(NULL)-t) >= 0) {
              olog<<i->job_id<<": Job is too old - deleting."<<endl;
              if(i->keep_deleted) {
                job_clean_deleted(i->job_id,*user);
                i->job_state = JOB_STATE_DELETED;
                state_changed=true;
              } else {
                /* delete everything */
                job_clean_final(i->job_id,*user);
              };
            };
          };
        };           
      }; break;
      case JOB_STATE_DELETED: {
        if(hard_job) { /* try to minimize load */
          time_t t = -1;
          if(!job_local_read_cleanuptime(i->job_id,*user,t)) {
            /* should not happen - delete job */
            JobLocalDescription job_desc;
            /* read lifetime - if empty it wont be overwritten */
            job_clean_final(i->job_id,*user);
          } else {
            /* check if it is not time to remove remnants of that */
            if((time(NULL)-(t+i->keep_deleted)) >= 0) {
              olog<<i->job_id<<": Job is ancient - delete rest of information."<<endl;
              /* delete everything */
              job_clean_final(i->job_id,*user);
            };
          };
        };
      }; break;
      default: { // should destroy job with unknown state ?!
      };
    };
    if(state_changed) {
      olog<<i->job_id<<": State: "<<JobDescription::get_state_name(i->job_state)<<" from "<<JobDescription::get_state_name(old_state)<<endl;
      i->job_pending=false;
      /* talk to external plugin to ask if we can proceed */
      if(plugins) {
        int result;
        string response;
        ContinuationPlugins::action_t act = 
                     plugins->run(*i,*user,result,response);
        // analyze result
        if(act == ContinuationPlugins::act_fail) {
          olog<<i->job_id<<": Plugin in state "<<
                states_all[i->get_state()].name<<" : "<< response << endl;
          i->AddFailure(string("Plugin at state ")+
               states_all[i->get_state()].name+" failed: "+response);
          job_error=true;
        } else if(act == ContinuationPlugins::act_log) {
          // Scream but go ahead
          olog<<i->job_id<<": Plugin at state "<<
                states_all[i->get_state()].name<<" : "<< response << endl;
        } else if(act == ContinuationPlugins::act_pass) {
          // Just continue quietly 
        } else {
          olog<<i->job_id<<": Plugin execution failed"<<endl;
          i->AddFailure(string("Failed running plugin at state ")+
                                          states_all[i->get_state()].name);
          job_error=true;
        };
      };
      if(!job_error) if(!job_state_write_file(*i,*user,i->job_state)) {
        job_error=true; i->AddFailure("Failed writing job status");
        once_more=false;
      };
      if(!job_error) {
        if(i->job_state == JOB_STATE_FINISHED) {
          if(i->GetLocalDescription(*user)) {
            job_stdlog_move(*i,*user,i->local->stdlog);
          };
          job_clean_finished(i->job_id,*user);
          job_log.make_file(*i,*user);
          job_log.finish_info(*i,*user);
        }
        else if(i->job_state == JOB_STATE_PREPARING) {
          job_log.start_info(*i,*user);
        };
      };
    };
    if(job_error) { /* something failed during processing of the job */
      olog<<i->job_id<<": Request to cancel failed job"<<endl;
      if(!FailedJob(i)) { /* something is really wrong */
        i->AddFailure("Failed during processing failure");
        delete_job=true;
      }
      else { /* just move job to state FINISHING */
        if((i->job_state) != JOB_STATE_FINISHED) {
          i->job_state = JOB_STATE_FINISHING;
        };
        i->job_pending=false;
        if(!job_state_write_file(*i,*user,i->job_state)) {
          delete_job=true;
        };
        once_more=true;
      };
    };
    /* send mail after error is processed */
    /* do not send if something really wrong happened to avoid email DoS */
    if(state_changed && !delete_job) {
      send_mail(*i,*user);
    };
    if(delete_job) { 
      olog << i->job_id << ": Delete request " << endl;
//      DestroyJob(i);
      i->job_state = JOB_STATE_FINISHED; /* move to finished in order to 
                                            remove from list */
      i->job_pending=false;
      job_state_write_file(*i,*user,i->job_state); 
      i->AddFailure("Serious troubles (problems during processing problems)");
      FailedJob(i);  /* put some marks */
      if(i->GetLocalDescription(*user)) {
        job_stdlog_move(*i,*user,i->local->stdlog);
      };
      job_clean_finished(i->job_id,*user);  /* clean status files */
      once_more=true; hard_job=true; /* to process some things in local */
    };
  };
  /* FINISHED+DELETED jobs are not kept in list - only in files */
  /* if job managed to get here with state UNDEFINED - 
     means we are overloaded with jobs - do not keep them in list */
  if((i->job_state == JOB_STATE_FINISHED) ||
     (i->job_state == JOB_STATE_DELETED) ||
     (i->job_state == JOB_STATE_UNDEFINED)) {
    /* this is the ONLY place there jobs are removed from memory */
    /* update counters */
    if(!old_pending) jobs_num[old_state]--;
    if(i->local) { delete i->local; };
    i=jobs.erase(i);
  }
  else {
    /* update counters */
    if(!old_pending) jobs_num[old_state]--;
    if(!i->job_pending) jobs_num[i->job_state]++;
    ++i;
  };
  return true;
}

#endif //  NO_GLOBUS_CODE

class JobFDesc {
 public:
  JobId id;
  uid_t uid;
  gid_t gid;
  time_t t;
  JobFDesc(char* s,unsigned int n):id(s,n),uid(0),gid(0),t(-1) { };
  bool operator<(JobFDesc &right) { return (t < right.t); };
};

/* find new jobs - sort by date to implement FIFO */
bool JobsList::ScanNewJobs(bool hard_job) {
  struct dirent file_;
  struct dirent *file;
  string cdir=user->ControlDir();
  DIR *dir=opendir(cdir.c_str());
  if(dir == NULL) {
    olog << "Failed reading control directory: " << user->ControlDir() << endl;
    return false;
  };
  list<JobFDesc> ids;
  for(;;) {
    readdir_r(dir,&file_,&file);
    if(file == NULL) break;
    int l=strlen(file->d_name);
    if(l>(4+7)) {  /* job id contains at least 1 character */
      if(!strncmp(file->d_name,"job.",4)) {
        if(!strncmp((file->d_name)+(l-7),".status",7)) {
          JobFDesc id((file->d_name)+4,l-7-4);
          if(FindJob(id.id) == jobs.end()) {
            string fname=cdir+'/'+file->d_name;
            uid_t uid;
            gid_t gid;
            time_t t;
            if(check_file_owner(fname,*user,uid,gid,t)) {
              /* add it to the list */
              id.uid=uid; id.gid=gid; id.t=t;
              ids.push_back(id);
            };
          };
        };
      };
    };
  };
  closedir(dir);
  /* sorting by date */
  ids.sort();
  for(list<JobFDesc>::iterator id=ids.begin();id!=ids.end();++id) {
    iterator i;
    /* adding job with file's uid/gid */
    if(AddJobNoCheck(id->id,i,id->uid,id->gid)) {
/*    ActJob(i,hard_job);  */
    };
 /* failed AddJob most probably means it is already in list or limit exceeded */
  };
  return true;
}

