#ifndef GRID_MANAGER_USERS_H
#define GRID_MANAGER_USERS_H

#include <sys/types.h>
#include <string>
#include <list>

class JobsList;
class RunElement;
class RunPlugin;

class JobUser;

/*
  Object to run external processes associated with user (helper)
*/
class JobUserHelper {
 private:
  /* command beeing run */
  string command;
  /* object representing running process */
  RunElement *proc;
 public:
  JobUserHelper(const string &cmd);
  ~JobUserHelper(void);
  /* start process if it is not running yet */
  bool run(JobUser &user);
};

/*
  Description of user - owner of jobs
*/
class JobUser {
 private:
  /* directory where files explaining jobs are stored */
  string control_dir;
  /* directory where directories used to run jobs are created */
  string session_root;
  /* directories to store cache information and cached files */
  string cache_dir;
  string cache_data_dir;
  /* directory used to create soft-links to cached files */
  string cache_link_dir;
  /* if cache is owned ONLY by this user */
  bool   private_cache;
  /* upper and lower watermarks for cache */
  long long int cache_max;
  long long int cache_min;
  /* deafault LRMS and queue to use */
  string default_lrms;
  string default_queue;
  /* local name, home directory, uid and gid of this user */
  string unix_name;
  string home;
  uid_t uid;
  gid_t gid;
  /* How long jobs are kept after job finished. This is 
     default and maximal value. */
  time_t keep_finished;
  time_t keep_deleted;
  /* Maximal value of times job is allowed to be rerun. 
     Default is 0. */
  int reruns;
  /* not used */
  unsigned long long int diskspace;
  /* if this object is valid */
  bool valid;
  /* list of associated external processes */
  list<JobUserHelper> helpers;
  /* List of jobs belonging to this user */
  JobsList *jobs;    /* filled by external functions */
  RunPlugin* cred_plugin;
 public:
  JobsList* get_jobs() const { 
    return jobs;
  };
  void operator=(JobsList *jobs_list) { 
    jobs=jobs_list;
  };
  JobUser(void);
  JobUser(const string &unix_name,RunPlugin* cred_plugin = NULL);
  JobUser(uid_t uid,RunPlugin* cred_plugin = NULL);
  JobUser(const JobUser &user);
  ~JobUser(void);
  /* Set and get corresponding private values */
  void SetControlDir(const string &dir);
  void SetSessionRoot(const string &dir);
  void SetCacheDir(const string &dir,const string &data_dir,bool priv = false);
  void SetCacheDir(const string &dir,const string &data_dir,const string &link_dir,bool priv = false);
  void SetCacheSize(long long int cache_max,long long int cache_min = 0);
  void SetLRMS(const string &lrms_name,const string &queue_name);
  void SetKeepFinished(time_t ttl) { keep_finished=ttl; };
  void SetKeepDeleted(time_t ttr) { keep_deleted=ttr; };
  void SetReruns(int n) { reruns=n; };
  void SetDiskSpace(unsigned long long int n) { diskspace=n; };
  bool CreateDirectories(void);
  bool is_valid(void) { return valid; };
  const string & ControlDir(void) const { return control_dir; };
  const string & SessionRoot(void) const { return session_root; };
  const string & CacheDir(void) const { return cache_dir; };
  const string & CacheDataDir(void) const { return cache_data_dir; };
  const string & CacheLinkDir(void) const { return cache_link_dir; };
  long long int CacheMaxSize(void) const { return cache_max; };
  long long int CacheMinSize(void) const { return cache_min; };
  bool CachePrivate(void) const { return private_cache; };
  const string & DefaultLRMS(void) const { return default_lrms; };
  const string & DefaultQueue(void) const { return default_queue; };
  const string & UnixName(void) const { return unix_name; };
  const string & Home(void) const { return home; };
  time_t KeepFinished(void) const { return keep_finished; };
  time_t KeepDeleted(void) const { return keep_deleted; };
  uid_t get_uid(void) const { return uid; };
  gid_t get_gid(void) const { return gid; };
  int Reruns(void) const { return reruns; };
  RunPlugin* CredPlugin(void) { return cred_plugin; };
  unsigned long long int DiskSpace(void) { return diskspace; };
  bool operator==(string name) { return (name == unix_name); };
  /* Change owner of the process to this user if su=true.
     Otherwise just set environment variables USER_ID and USER_NAME */
  bool SwitchUser(bool su = true) const;
  void add_helper(const string &helper) {
    helpers.push_back(JobUserHelper(helper));
  };
  bool has_helpers(void) { return (helpers.size() != 0); };
  /* Start/restart all helper processes */
  bool run_helpers(void);
  bool substitute(string& param) const;
};

/*
  List of users serviced by grid-manager.
  Just wrapper around 'list<JobUser>' with little additional
  functionality.
*/
class JobUsers {
 private:
  list<JobUser> users;
 public:
  typedef list<JobUser>::iterator iterator;
  typedef list<JobUser>::const_iterator const_iterator;
  JobUsers(void);
  ~JobUsers(void);
  iterator begin(void) { return users.begin(); };
  const_iterator begin(void) const { return users.begin(); };
  iterator end(void) { return users.end(); };
  const_iterator end(void) const { return users.end(); };
  size_t size(void) const { return users.size(); };
  /* Create new user object and push into list */
  iterator AddUser(const string &unix_name,RunPlugin* cred_plugin = NULL,const string &control_dir = "",const string &session_root = "");
  iterator DeleteUser(iterator &user) { return users.erase(user); };
  /* Check existance of user of given unix name */
  bool HasUser(const string &name) {
    for(iterator i=users.begin();i!=users.end();++i)
      if((*i) == name) return true;
    return false;
  };
  string ControlDir(iterator user);
  string ControlDir(const string user);
  /* Find user of given unix name */
  iterator find(const string user);
  /* Start/restart all associated processes of all users */
  bool run_helpers(void);
  bool substitute(string& param) const;
};

#endif
