#include "../std.h"
#include <string>
#include <iostream>
#include <fstream>
#include <globus_common.h>
#include <globus_rsl.h>
#include <globus_gram_protocol.h>
#include <globus_symboltable.h>
#include "../jobs/users.h"
#include "../files/info_files.h"
#include "../files/info_types.h"
#include "../files/info_types.h"
#include "../transfer/replica_utils.h"
#include "../misc/canonical_dir.h"
#include "../misc/stringtoint.h"
#include "../misc/inttostring.h"
#include "../misc/url_options.h"
#include "../misc/log_time.h"
#include "../config/environment.h"
#include "subst_rsl.h"
#include "parse_rsl.h"

static void rsl_value_to_grami(ostream &o,globus_rsl_value_t* value);
static void rsl_print_to_grami(ostream &o,globus_rsl_t *cur);

/* function for jobmanager-ng */
bool parse_rsl_for_action(const char* fname,
               string &action,string &jobid,
               string &lrms,string &queue) {
  JobLocalDescription job_desc;
  string filename(fname);
  if(parse_rsl(filename,job_desc)) {
    action=job_desc.action;
    jobid=job_desc.jobid;
    lrms=job_desc.lrms;
    queue=job_desc.queue;
    return true;
  };
  return false;
}

/* parse RSL and write few informational files */
bool process_rsl(JobUser &user,const JobDescription &desc) {
  JobLocalDescription job_desc;
  return process_rsl(user,desc,job_desc);
}

bool process_rsl(JobUser &user,const JobDescription &desc,JobLocalDescription &job_desc) {
  /* read local first to get some additional info pushed here by script */
  job_local_read_file(desc.get_id(),user,job_desc);
  /* some default values */
  job_desc.lrms=user.DefaultLRMS();
  job_desc.queue=user.DefaultQueue();
  string filename;
  filename = user.ControlDir() + "/job." + desc.get_id() + ".description";
  if(!parse_rsl(filename,job_desc)) return false;
  if(job_desc.reruns>user.Reruns()) job_desc.reruns=user.Reruns();
  if((job_desc.diskspace>user.DiskSpace()) || (job_desc.diskspace==0)) {
    job_desc.diskspace=user.DiskSpace();
  };
  if(job_desc.rc.length() != 0) {
    for(FileData::iterator i=job_desc.outputdata.begin();
                         i!=job_desc.outputdata.end();++i) {
      insert_RC_to_url(i->lfn,job_desc.rc);
    };
    for(FileData::iterator i=job_desc.inputdata.begin();
                         i!=job_desc.inputdata.end();++i) {
      insert_RC_to_url(i->lfn,job_desc.rc);
    };
  };
  if(job_desc.gsiftpthreads > 1) {
    string v = inttostring(job_desc.gsiftpthreads);
    for(FileData::iterator i=job_desc.outputdata.begin();
                         i!=job_desc.outputdata.end();++i) {
      add_url_option(i->lfn,"threads",v.c_str(),-1);
    };
    for(FileData::iterator i=job_desc.inputdata.begin();
                         i!=job_desc.inputdata.end();++i) {
      add_url_option(i->lfn,"threads",v.c_str(),-1);
    };
  };
  if(job_desc.cache.length() != 0) {
    string value;
    for(FileData::iterator i=job_desc.outputdata.begin();
                         i!=job_desc.outputdata.end();++i) {
      get_url_option(i->lfn,"cache",-1,value);
      if( value.length() == 0 )
	add_url_option(i->lfn,"cache",job_desc.cache.c_str(),-1);
    };
    for(FileData::iterator i=job_desc.inputdata.begin();
                         i!=job_desc.inputdata.end();++i) {
      get_url_option(i->lfn,"cache",-1,value);
      if( value.length() == 0 )
	add_url_option(i->lfn,"cache",job_desc.cache.c_str(),-1);
    };
  };
  if(!job_local_write_file(desc,user,job_desc)) return false;
  if(!job_input_write_file(desc,user,job_desc.inputdata)) return false;
  if(!job_output_write_file(desc,user,job_desc.outputdata)) return false;
  return true;  
}

void add_non_cache(const char *fname,list<FileData> &inputdata) {
  for(list<FileData>::iterator i=inputdata.begin();i!=inputdata.end();++i) {
    if(i->has_lfn()) if((*i) == fname) {
      add_url_option(i->lfn,"cache","no",-1);
      add_url_option(i->lfn,"exec","yes",-1);
    };
  };
}

/* parse rsl and set specified file permissions to executable */
bool set_execs(const JobId &job_id,const JobUser &user,const string &session_dir) {
  bool res=false;
  string frsl = user.ControlDir() + "/job." + job_id + ".description";
  globus_rsl_t *rsl_tree = NULL;
  rsl_tree=read_rsl(frsl);
  if(rsl_tree == NULL) return false;
  char** tmp_param;
  if (globus_rsl_param_get(rsl_tree,GLOBUS_RSL_PARAM_MULTI_LITERAL,
                            NG_RSL_ARGUMENTS_PARAM,&tmp_param) != 0) {
    olog << "Broken RSL" << endl; goto exit;
  };
  if(tmp_param[0] == NULL) {
    olog << "Missing arguments in RSL" << endl; goto exit;
  };
  /* executable can be external, so first check for initial slash */
  if(( tmp_param[0][0] != '/' ) && ( tmp_param[0][0] != '$' )){
    fix_file_permissions(session_dir+"/"+tmp_param[0],true);
  };
  if (globus_rsl_param_get(rsl_tree,GLOBUS_RSL_PARAM_MULTI_LITERAL,
                            NG_RSL_EXECUTABLES_PARAM,&tmp_param) != 0) {
    olog << "Broken RSL" << endl; goto exit;
  };
  for(int i=0;tmp_param[i]!=NULL;i++) { fix_file_permissions(session_dir+"/"+tmp_param[i],true);
  };
  res=true;
exit:
  if(rsl_tree) globus_rsl_free_recursive(rsl_tree);
  return res;
}

/* parse rsl (and local configuration) and write grami file.
   it also returns job local description (for not to parse twice) */

class value_for_shell {
 friend ostream& operator<<(ostream&,const value_for_shell&);
 private:
  const char* str;
  bool quote;
 public:
  value_for_shell(const char *str_,bool quote_):str(str_),quote(quote_) { };
  value_for_shell(const string &str_,bool quote_):str(str_.c_str()),quote(quote_) { };
};

ostream& operator<<(ostream &o,const value_for_shell &s) {
  if(s.str == NULL) return o;
  if(s.quote) o<<"'";
  const char* p = s.str;
  for(;;) {
    char* pp = strchr(p,'\'');
    if(pp == NULL) { o<<p; if(s.quote) o<<"'"; break; };
    o.write(p,pp-p); o<<"'\\''"; p=pp+1;
  };
  return o;
}

class numvalue_for_shell {
 friend ostream& operator<<(ostream&,const numvalue_for_shell&);
 private:
  long int n;
 public:
  numvalue_for_shell(const char *str) {
    n=0;
    sscanf(str,"%li",&n);
  };
};

ostream& operator<<(ostream &o,const numvalue_for_shell &s) {
  o<<s.n;
  return o;
};
  
bool write_grami(const JobDescription &desc,const JobUser &user,const string &session_dir,JobLocalDescription &job_desc,const char* opt_add) {
  bool res=false;
  string fgrami = user.ControlDir() + "/job." + desc.get_id() + ".grami";
  string frsl = user.ControlDir() + "/job." + desc.get_id() + ".description";
  /* read local parameters */
  /* !!! most probaly there is no need to read file - should be read already */
  if(!job_local_read_file(desc.get_id(),user,job_desc)) {
    olog << "Can't read job local description" << endl; return false;
  };
  /* read RSL and write grami */
  globus_rsl_t *rsl_tree = NULL;
  rsl_tree=read_rsl(frsl);
  if(rsl_tree == NULL) return false;
  ofstream f(fgrami.c_str(),ios::out | ios::trunc);
  if(!f.is_open()) return false;
  if(!fix_file_owner(fgrami,desc,user)) { goto exit; };
  char** tmp_param;
  f<<"joboption_directory='"<<session_dir<<"'"<<endl;

#define GLOBUS_RSL_PARAM_GET(RSL,TYPE,NAME,PARAM)              \
        if(globus_rsl_param_get(RSL,TYPE,NAME,&PARAM) != 0) {  \
          olog<<"Broken RSL in NAME"<<endl; goto exit;             \
        }

  GLOBUS_RSL_PARAM_GET(rsl_tree,GLOBUS_RSL_PARAM_MULTI_LITERAL,
                            NG_RSL_ARGUMENTS_PARAM,tmp_param);
  if(tmp_param[0] == NULL) {
    olog<<"Missing arguments in RSL"<<endl; goto exit;
  };
  f<<"joboption_arg_0='";
  if((tmp_param[0][0] != '/') && (tmp_param[0][0] != '$')) f<<"./";
  f<<value_for_shell(tmp_param[0],false)<<"'"<<endl;
  for(int i=1;tmp_param[i]!=NULL;i++) { 
    f<<"joboption_arg_"<<i<<"="<<value_for_shell(tmp_param[i],true)<<endl;
  };
  GLOBUS_RSL_PARAM_GET(rsl_tree,GLOBUS_RSL_PARAM_SEQUENCE,
                           GLOBUS_GRAM_PROTOCOL_ENVIRONMENT_PARAM,tmp_param);
  {
    int i;
    for(i=0;tmp_param[i]!=NULL;i++) { 
      f<<"joboption_env_"<<(i/2)<<"='"<<value_for_shell(tmp_param[i],false)<<"=";
      i++; if(tmp_param[i]==NULL) { f<<"'"<<endl; break; };
      f<<value_for_shell(tmp_param[i],false)<<"'"<<endl;
    };
/*
    f<<joboption_env_"<<(i/2)<<"='X509_CERT_DIR=/etc/grid-security/certificates'"<<endl; i+=2;
    f<<joboption_env_"<<(i/2)<<"='GLOBUS_LOCATION="<<globus_loc<<"'"<<endl; i+=2;
    f<<joboption_env_"<<(i/2)<<"='X509_USER_PROXY="<<user.ControlDir()+"/job."+desc.get_id()+".proxy'"<<endl;
*/
  };
  GLOBUS_RSL_PARAM_GET(rsl_tree,GLOBUS_RSL_PARAM_MULTI_LITERAL,
                                   NG_RSL_RUNTIME_PARAM,tmp_param);
  for(int i=0;tmp_param[i]!=NULL;i++) { 
    for(char* s=tmp_param[i];*s;s++) (*s)=toupper(*s);
    string tmp_s = tmp_param[i];
    if(canonical_dir(tmp_s) != 0) {
      olog<<"Bad name for runtime environemnt: "<<tmp_param[i]<<endl; goto exit;
    };
    f<<"joboption_runtime_"<<i<<"="<<value_for_shell(tmp_param[i],true)<<endl;
  };
  GLOBUS_RSL_PARAM_GET(rsl_tree,GLOBUS_RSL_PARAM_SINGLE_LITERAL,
                       GLOBUS_GRAM_PROTOCOL_STDIN_PARAM,tmp_param);
  if(tmp_param[0] == NULL) {
    GLOBUS_RSL_PARAM_GET(rsl_tree,GLOBUS_RSL_PARAM_SINGLE_LITERAL,
                                     NG_RSL_STDIN_PARAM,tmp_param);
    if(tmp_param[0] == NULL) {
      GLOBUS_RSL_PARAM_GET(rsl_tree,GLOBUS_RSL_PARAM_SINGLE_LITERAL,
                                    NG_RSL_STDIN_PARAM2,tmp_param);
    };
  };
  if(tmp_param[0] == NULL) {
    tmp_param[0]=GLOBUS_GRAM_PROTOCOL_DEFAULT_STDIN;
  };
  f<<"joboption_stdin="<<value_for_shell(tmp_param[0],true)<<endl;
  GLOBUS_RSL_PARAM_GET(rsl_tree,GLOBUS_RSL_PARAM_SINGLE_LITERAL,
                            GLOBUS_GRAM_PROTOCOL_STDOUT_PARAM,tmp_param);
  if(tmp_param[0] == NULL) { 
    string stdout=GLOBUS_GRAM_PROTOCOL_DEFAULT_STDOUT;
    f<<"joboption_stdout="<<value_for_shell(stdout,true)<<endl;
  }
  else {
    string stdout=tmp_param[0];
    if(canonical_dir(stdout) != 0) {
      olog<<"Bad name for stdout: "<<stdout<<endl; goto exit;
    };
    stdout=session_dir+stdout;
    f<<"joboption_stdout="<<value_for_shell(stdout,true)<<endl;
  };
  GLOBUS_RSL_PARAM_GET(rsl_tree,GLOBUS_RSL_PARAM_SINGLE_LITERAL,
                            GLOBUS_GRAM_PROTOCOL_STDERR_PARAM,tmp_param);
  if(tmp_param[0] == NULL) { 
    string stderr=GLOBUS_GRAM_PROTOCOL_DEFAULT_STDOUT;
    f<<"joboption_stderr="<<value_for_shell(stderr,true)<<endl;
  }
  else {
    string stderr=tmp_param[0];
    if(canonical_dir(stderr) != 0) {
      olog<<"Bad name for stderr: "<<stderr<<endl; goto exit;
    };
    stderr=session_dir+stderr;
    f<<"joboption_stderr="<<value_for_shell(stderr,true)<<endl;
  };
  GLOBUS_RSL_PARAM_GET(rsl_tree,GLOBUS_RSL_PARAM_SINGLE_LITERAL,
                      NG_RSL_CPUTIME_PARAM,tmp_param);
  if(tmp_param[0] == NULL) { f<<"joboption_cputime="<<endl; }
  else { f<<"joboption_cputime="<<numvalue_for_shell(tmp_param[0])<<endl; };
  GLOBUS_RSL_PARAM_GET(rsl_tree,GLOBUS_RSL_PARAM_SINGLE_LITERAL,
                      NG_RSL_MEMORY_PARAM,tmp_param);
  if(tmp_param[0] == NULL) { f<<"joboption_memory="<<endl; }
  else { f<<"joboption_memory="<<numvalue_for_shell(tmp_param[0])<<endl; };
  GLOBUS_RSL_PARAM_GET(rsl_tree,GLOBUS_RSL_PARAM_SINGLE_LITERAL,
                      NG_RSL_COUNT_PARAM,tmp_param);
  if(tmp_param[0] == NULL) { f<<"joboption_count=1"<<endl; }
  else { f<<"joboption_count="<<numvalue_for_shell(tmp_param[0])<<endl; };
//  f<<"joboption_jobtype=1" << endl; 
  f<<"joboption_queue="<<value_for_shell(job_desc.queue,true)<<endl;
  f<<"joboption_jobname="<<value_for_shell(job_desc.jobname,true)<<endl;
  if(job_desc.exectime.defined()) {
    f<<"joboption_starttime="<<job_desc.exectime<<endl;
  }
  else {
    f<<"joboption_starttime="<<endl;
  };
//  GLOBUS_RSL_PARAM_GET(rsl_tree,GLOBUS_RSL_PARAM_SINGLE_LITERAL,
//                        GLOBUS_GRAM_PROTOCOL_PROJECT_PARAM,tmp_param);
//  if(tmp_param[0] == NULL) { f << "grami_project=''" << endl;  }
//  else { f << "grami_project='" << tmp_param[0] << "'" << endl;  };
  f<<"joboption_gridid="<<value_for_shell(desc.get_id(),true)<<endl;
  if(opt_add) f<<opt_add<<endl;
  // put all RSL attributes to grami for super-clever submission script
  f<<"joboption_rsl="<<frsl<<endl;
  rsl_print_to_grami(f,rsl_tree);
  res=true;
exit:
  if(rsl_tree) globus_rsl_free_recursive(rsl_tree);
  f.close();
  return res;
}

/* extract joboption_jobid from grami file */
string read_grami(const JobId &job_id,const JobUser &user,const string &session_dir) {
  const char* local_id_param = "joboption_jobid=";
  int l = strlen(local_id_param);
  string id = "";
  char buf[256];
  string fgrami = user.ControlDir() + "/job." + job_id + ".grami";
  ifstream f(fgrami.c_str());
  if(!f.is_open()) return id;
  for(;!f.eof();) {
    istream_readline(f,buf,sizeof(buf));
    if(strncmp(local_id_param,buf,l)) continue;
    if(buf[0]=='\'') {
      l++; int ll = strlen(buf);
      if(buf[ll-1]=='\'') buf[ll-1]=0;
    };
    id=buf+l; break;
  };
exit:
  f.close();
  return id;
}

globus_rsl_t* read_rsl(const string &fname) {
  globus_rsl_t *rsl_tree = NULL;
  string rsl;
  char* rsl_spec;
  if(!job_description_read_file(fname,rsl)) {
    olog << "Failed reading RSL" << endl; return NULL; 
  };
  rsl_spec=strdup(rsl.c_str());
  rsl.erase();
  /* parse RSL */
  rsl_tree=globus_rsl_parse(rsl_spec);
  globus_free(rsl_spec); rsl_spec = GLOBUS_NULL;
  return rsl_tree;
}

bool write_rsl(const string &fname,globus_rsl_t *rsl) {
  char* rsl_spec = NULL;
  if((rsl_spec = globus_rsl_unparse(rsl)) == GLOBUS_NULL) return false;
  if(!job_description_write_file(fname,rsl_spec)) {
    globus_free(rsl_spec);
    olog << "Failed writing RSL" << endl; return false; 
  };
  globus_free(rsl_spec);
  return true;
}

/* parse rsl, fill job description (local) */
bool parse_rsl(const string &fname,JobLocalDescription &job_desc,string* acl) {
  char** tmp_param;
  globus_rsl_t *rsl_tree = NULL;

  bool res = false;

  /* read RSL */
  rsl_tree=read_rsl(fname);
  if (!rsl_tree) {
    olog << "Failed  parsing RSL" << endl; goto exit;
  };
  if (globus_rsl_param_get(rsl_tree,GLOBUS_RSL_PARAM_SINGLE_LITERAL,
                            NG_RSL_JOB_ID_PARAM,&tmp_param) != 0) {
    olog << "Broken RSL in jobid" << endl; goto exit;
  };
  if(tmp_param[0]) job_desc.jobid=tmp_param[0];
  if(job_desc.jobid.find('/') != string::npos) {
    olog << "slashes are not allowed in jobid" << endl; goto exit;
  };
  if (globus_rsl_param_get(rsl_tree,GLOBUS_RSL_PARAM_SINGLE_LITERAL,
                            NG_RSL_ACTION_PARAM,&tmp_param) != 0) {
    olog << "Broken RSL in action" << endl; goto exit;
  };
  if(tmp_param[0] == NULL) {
    olog << "Missing action in RSL" << endl; goto exit;
  };
  job_desc.action=tmp_param[0];
  if(strcasecmp(tmp_param[0],"request")) {
     /* not a job request, need nothing more */
    res=true; goto exit;
  };
/*
  if (globus_rsl_param_get(rsl_tree,GLOBUS_RSL_PARAM_SINGLE_LITERAL,
                            NG_RSL_LRMS_PARAM,&tmp_param) != 0) {
    olog << "Broken RSL in lrmstype" << endl; goto exit;
  };
  if(tmp_param[0] != NULL) { job_desc.lrms=tmp_param[0]; };
*/
  if (globus_rsl_param_get(rsl_tree,GLOBUS_RSL_PARAM_SINGLE_LITERAL,
                            NG_RSL_QUEUE_PARAM,&tmp_param) != 0) {
    olog << "Broken RSL in queue" << endl; goto exit;
  };
  if(tmp_param[0] != NULL) { job_desc.queue=tmp_param[0]; }
  if (globus_rsl_param_get(rsl_tree,GLOBUS_RSL_PARAM_SINGLE_LITERAL,
                            NG_RSL_REPLICA_PARAM,&tmp_param) != 0) {
    olog << "Broken RSL in replicacollection" << endl; goto exit;
  };
  if(tmp_param[0] != NULL) { job_desc.rc=tmp_param[0]; };
  if (globus_rsl_param_get(rsl_tree,GLOBUS_RSL_PARAM_SINGLE_LITERAL,
                            NG_RSL_LIFETIME_PARAM,&tmp_param) != 0) {
    olog << "Broken RSL in lifetime" << endl; goto exit;
  };
  if(tmp_param[0] != NULL) { job_desc.lifetime=tmp_param[0]; };
//  if (globus_rsl_param_get(rsl_tree,GLOBUS_RSL_PARAM_SINGLE_LITERAL,
//                            NG_RSL_STARTTIME_PARAM,&tmp_param) != 0) {
//    olog << "Broken RSL in runtime" << endl; goto exit;
//  };
//  if(tmp_param[0] != NULL) { job_desc.exectime=tmp_param[0]; };
  if (globus_rsl_param_get(rsl_tree,GLOBUS_RSL_PARAM_SINGLE_LITERAL,
                            NG_RSL_STARTTIME_PARAM,&tmp_param) != 0) {
    olog << "Broken RSL in starttime" << endl; goto exit;
  };
  if(tmp_param[0] != NULL) { job_desc.processtime=tmp_param[0]; };
  if (globus_rsl_param_get(rsl_tree,GLOBUS_RSL_PARAM_SINGLE_LITERAL,
                            NG_RSL_JOBNAME_PARAM,&tmp_param) != 0) {
    olog << "Broken RSL in jobname" << endl; goto exit;
  };
  if(tmp_param[0] != NULL) { job_desc.jobname=tmp_param[0]; };
  if (globus_rsl_param_get(rsl_tree,GLOBUS_RSL_PARAM_SINGLE_LITERAL,
                            NG_RSL_JOBREPORT_PARAM,&tmp_param) != 0) {
    olog << "Broken RSL in jobreport" << endl; goto exit;
  };
  if(tmp_param[0] != NULL) { job_desc.jobreport=tmp_param[0]; };
  if (globus_rsl_param_get(rsl_tree,GLOBUS_RSL_PARAM_SINGLE_LITERAL,
                            NG_RSL_RERUN_PARAM,&tmp_param) != 0) {
    olog << "Broken RSL in rerun" << endl; goto exit;
  };
  if(tmp_param[0] != NULL) {
    if(!stringtoint(string(tmp_param[0]),job_desc.reruns)) {
      olog << "Bad integer in rerun" << endl; goto exit;
    };
  };
  if (globus_rsl_param_get(rsl_tree,GLOBUS_RSL_PARAM_SINGLE_LITERAL,
                            NG_RSL_DISKSPACE_PARAM,&tmp_param) != 0) {
    olog << "Broken RSL in disk" << endl; goto exit;
  };
  if(tmp_param[0] != NULL) {
    double ds = 0.0;
    if((sscanf(tmp_param[0],"%lf",&ds) != 1) || (ds < 0)) {
      olog << "disk value is bad" << endl; goto exit;
    };
    job_desc.diskspace=(unsigned long long int)(ds*1024*1024*1024); /* unit - GB */
  };
  if (globus_rsl_param_get(rsl_tree,GLOBUS_RSL_PARAM_SINGLE_LITERAL,
                            NG_RSL_NOTIFY_PARAM,&tmp_param) != 0) {
    olog << "Broken RSL in notify" << endl; goto exit;
  };
  if(tmp_param[0] != NULL) { job_desc.notify=tmp_param[0]; };
  if (globus_rsl_param_get(rsl_tree,GLOBUS_RSL_PARAM_MULTI_LITERAL,
                            NG_RSL_ARGUMENTS_PARAM,&tmp_param) != 0) {
    olog << "Broken RSL in arguments" << endl; goto exit;
  };
  if(tmp_param[0] == 0) {
    olog << "Missing arguments in RSL" << endl; goto exit;
  };
  job_desc.arguments.clear();
  for(int i=0;tmp_param[i]!=NULL;i++) {
    job_desc.arguments.push_back(string(tmp_param[i]));
  };
  if (globus_rsl_param_get(rsl_tree,GLOBUS_RSL_PARAM_SEQUENCE,
                            NG_RSL_INPUT_DATA_PARAM,&tmp_param) != 0) {
    olog << "Broken RSL in inputdata" << endl; goto exit;
  };
  job_desc.inputdata.clear();
  job_desc.downloads=0;
  for(int i=0;tmp_param[i]!=NULL;i+=2) {
    FileData fdata(tmp_param[i],tmp_param[i+1]);
    job_desc.inputdata.push_back(fdata);
    if(fdata.has_lfn()) job_desc.downloads++;
    if(tmp_param[i+1]==NULL) break;
  };
  if (globus_rsl_param_get(rsl_tree,GLOBUS_RSL_PARAM_SEQUENCE,
                            NG_RSL_OUTPUT_DATA_PARAM,&tmp_param) != 0) {
    olog << "Broken RSL in outputdata" << endl; goto exit;
  };
  job_desc.outputdata.clear();
  job_desc.uploads=0;
  for(int i=0;tmp_param[i]!=NULL;i+=2) {
    FileData fdata(tmp_param[i],tmp_param[i+1]);
    job_desc.outputdata.push_back(fdata);
    if(fdata.has_lfn()) job_desc.uploads++;
    if(tmp_param[i+1]==NULL) break;
  };
  if (globus_rsl_param_get(rsl_tree,GLOBUS_RSL_PARAM_SINGLE_LITERAL,
             NG_RSL_STDLOG_PARAM,&tmp_param) != 0) {
    olog << "Broken RSL in gmlog" << endl; goto exit;
  };
  if(tmp_param[0] != NULL) { job_desc.stdlog=tmp_param[0]; };
  if (globus_rsl_param_get(rsl_tree,GLOBUS_RSL_PARAM_SINGLE_LITERAL,
             GLOBUS_GRAM_PROTOCOL_STDOUT_PARAM,&tmp_param) != 0) {
    olog << "Broken RSL in stdout" << endl; goto exit;
  };
  if(tmp_param[0] != NULL) { job_desc.stdout=tmp_param[0]; };
  if (globus_rsl_param_get(rsl_tree,GLOBUS_RSL_PARAM_SINGLE_LITERAL,
             GLOBUS_GRAM_PROTOCOL_STDERR_PARAM,&tmp_param) != 0) {
    olog << "Broken RSL in stderr" << endl; goto exit;
  };
  if(tmp_param[0] != NULL) { job_desc.stderr=tmp_param[0]; };
  if (globus_rsl_param_get(rsl_tree,GLOBUS_RSL_PARAM_SINGLE_LITERAL,
             GLOBUS_GRAM_PROTOCOL_STDIN_PARAM,&tmp_param) != 0) {
    olog << "Broken RSL in stdin" << endl; goto exit;
  };
  if(tmp_param[0] == NULL) {
    if (globus_rsl_param_get(rsl_tree,GLOBUS_RSL_PARAM_SINGLE_LITERAL,
             NG_RSL_STDIN_PARAM,&tmp_param) != 0) {
      olog << "Broken RSL" << endl; goto exit;
    };
    if(tmp_param[0] == NULL) {
      if (globus_rsl_param_get(rsl_tree,GLOBUS_RSL_PARAM_SINGLE_LITERAL,
               NG_RSL_STDIN_PARAM2,&tmp_param) != 0) {
        olog << "Broken RSL" << endl; goto exit;
      };
    };
  };
  if(tmp_param[0] != NULL) { job_desc.stdin=tmp_param[0]; };
  if (globus_rsl_param_get(rsl_tree,GLOBUS_RSL_PARAM_SINGLE_LITERAL,
             NG_RSL_FTPTHREADS_PARAM,&tmp_param) != 0) {
    olog << "Broken RSL in ftpthreads" << endl; goto exit;
  };
  if(tmp_param[0] != NULL) { 
    if(!stringtoint(string(tmp_param[0]),job_desc.gsiftpthreads)) {
      olog << "Bad integer in ftpthreads" << endl; goto exit;
    };
  };
  if (globus_rsl_param_get(rsl_tree,GLOBUS_RSL_PARAM_SINGLE_LITERAL,
             NG_RSL_CACHE_PARAM,&tmp_param) != 0) {
    olog << "Broken RSL in cache" << endl; goto exit;
  };
  if(tmp_param[0] != NULL) { 
    if( strcmp(tmp_param[0],"yes") && strcmp(tmp_param[0],"no") ) {
      olog << "Bad string in cache" << endl; goto exit;
    };
    job_desc.cache = (string)tmp_param[0];
  };
  if (globus_rsl_param_get(rsl_tree,GLOBUS_RSL_PARAM_SINGLE_LITERAL,
             NG_RSL_HOSTNAME_PARAM,&tmp_param) != 0) {
    olog << "Broken RSL in hostname" << endl; goto exit;
  };
  if(tmp_param[0] != NULL) {
    if(job_desc.clientname.length()!=0) { /* keep predefined value */
      if(job_desc.clientname.find(';') == string::npos) {
        job_desc.clientname+=";"; job_desc.clientname+=tmp_param[0];
      };
    }
    else {
      job_desc.clientname=tmp_param[0];
    };
  };
  if (globus_rsl_param_get(rsl_tree,GLOBUS_RSL_PARAM_SINGLE_LITERAL,
             NG_RSL_SOFTWARE_PARAM,&tmp_param) != 0) {
    olog << "Broken RSL in clientsoftware" << endl; goto exit;
  };
  if(tmp_param[0] != NULL) {
    job_desc.clientsoftware=tmp_param[0];
  };
  if (globus_rsl_param_get(rsl_tree,GLOBUS_RSL_PARAM_SINGLE_LITERAL,
             GLOBUS_GRAM_PROTOCOL_DRY_RUN_PARAM,&tmp_param) != 0) {
    olog << "Broken RSL in dryrun" << endl; goto exit;
  };
  if(tmp_param[0] != NULL) { 
    if(!strcasecmp(tmp_param[0],"yes")) job_desc.dryrun=true;
  };
  /* add cache=no to executable input files */
  if(job_desc.arguments.size() > 0) {
    if(( job_desc.arguments.begin()->c_str()[0] != '/' ) && 
       ( job_desc.arguments.begin()->c_str()[0] != '$' )){
      add_non_cache(job_desc.arguments.begin()->c_str(),job_desc.inputdata);
    };
  };
  if (globus_rsl_param_get(rsl_tree,GLOBUS_RSL_PARAM_MULTI_LITERAL,
                            NG_RSL_EXECUTABLES_PARAM,&tmp_param) != 0) {
    olog << "Broken RSL" << endl; goto exit;
  };
  for(int i=0;tmp_param[i]!=NULL;i++) {
    add_non_cache(tmp_param[i],job_desc.inputdata);
  };
  if(acl) {
    if (globus_rsl_param_get(rsl_tree,GLOBUS_RSL_PARAM_SINGLE_LITERAL,
               NG_RSL_ACL_PARAM,&tmp_param) != 0) {
      olog << "Broken RSL in acl" << endl; goto exit;
    };
    if(tmp_param[0] != NULL) {
      (*acl)=tmp_param[0];
    };
  };
  res=true;
exit:
  if(rsl_tree) globus_rsl_free_recursive(rsl_tree);
  return res;
}

/* do rsl substitution and nordugrid specific stuff */
/* then write file back */
bool preprocess_rsl(const string &fname,const string &session_dir,const string &jobid) {
  char** tmp_param;
  globus_rsl_t *rsl_tree = NULL;
  rsl_subst_table_t *symbol_table = NULL;

  bool res = false;

  rsl_tree=read_rsl(fname);
  if (!rsl_tree) {
    olog << "Failed parsing RSL" << endl; goto exit;
  };

  symbol_table = (rsl_subst_table_t *) globus_libc_malloc
                            (sizeof(rsl_subst_table_t));
  rsl_subst_table_init(symbol_table);
  rsl_subst_table_insert(symbol_table,
                           strdup("NORDUGRID_LOCATION"),
                           strdup(nordugrid_loc.c_str()));
  rsl_subst_table_insert(symbol_table,
                           strdup("NG_SESSION_DIR"),
                           strdup(session_dir.c_str()));
  rsl_subst_table_insert(symbol_table,
                           strdup("NG_JOB_ID"),
                           strdup(jobid.c_str()));
  rsl_subst_table_insert(symbol_table,
                           strdup("GLOBUS_LOCATION"),
                           strdup(globus_loc.c_str()));
//  if(globus_rsl_eval(rsl_tree, symbol_table) != 0) {
//    olog << "Failed evaluating RSL" << endl; goto exit;
//  };
  if(rsl_subst(rsl_tree, symbol_table) != 0) {
    olog << "Failed evaluating RSL" << endl; goto exit;
  };
  if(!write_rsl(fname,rsl_tree)) goto exit;
  res=true;
exit:
  if(rsl_tree) globus_rsl_free_recursive(rsl_tree);
  if(symbol_table) rsl_subst_table_destroy(symbol_table);
  return res;
}

/*
static char* rsl_operators [11] = {
  "?0?","=","!=",">",">=","<","<=","?7?","&","|","multi"
};
*/

static void rsl_print_to_grami(ostream &o,globus_rsl_t *cur) {
  if(globus_rsl_is_boolean(cur)) {
    globus_rsl_t* cur_;
    globus_list_t *list = cur->req.boolean.operand_list;
    for(;;) {
      if(globus_list_empty(list)) break;
      cur_=(globus_rsl_t*)globus_list_first(list);
      rsl_print_to_grami(o,cur_);
      list=globus_list_rest(list);
    };
  }
  else if(globus_rsl_is_relation(cur)) {
    if(cur->req.relation.my_operator=GLOBUS_RSL_EQ) {
      o<<"joboption_rsl_"<<cur->req.relation.attribute_name<<"='";
      rsl_value_to_grami(o,cur->req.relation.value_sequence);
    };
    o<<"'"<<endl;
  }
  else {
  };
}

static void rsl_value_to_grami(ostream &o,globus_rsl_value_t* cur) {
  if(globus_rsl_value_is_literal(cur)) {
    o<<value_for_shell(cur->value.literal.string,false); 
  }
  else if(globus_rsl_value_is_variable(cur)) {
    rsl_value_to_grami(o,cur->value.variable.sequence);
  }
  else if(globus_rsl_value_is_concatenation(cur)) {
    rsl_value_to_grami(o,cur->value.concatenation.left_value);
    rsl_value_to_grami(o,cur->value.concatenation.right_value);
  }
  else if(globus_rsl_value_is_sequence(cur)) {
    globus_rsl_value_t* cur_;
    globus_list_t *list = cur->value.sequence.value_list;
    bool first = true;
    for(;;) {
      if(globus_list_empty(list)) break;
      if(!first) { o<<" "; } else { first=false; };
      cur_=(globus_rsl_value_t*)globus_list_first(list);
      rsl_value_to_grami(o,cur_);
      list=globus_list_rest(list);
    };
  }
  else {
  };
}

