/*
  upload files from user interface to CE .
  result: 0 - ok, 1 - unrecoverable error, 2 - potentially recoverable.
*/
#include "std.h"
#include <globus_common.h>
#include <globus_ftp_client.h>
#include <globus_gram_client.h>
#include <globus_gram_protocol.h>
#include <globus_object.h>
#include "files/info_types.h"
#include "files/info_files.h"
#include "misc/canonical_dir.h"
#include "misc/globus_modules.h"
#include "misc/log_time.h"
#include "transfer/ftpsubmit.h"
#include "datamove/datamovepar.h"

#include "ui_uploader.h"

/* maximum number simultaneous uploads */
#define MAX_UPLOADS 2

static list<FileData> job_files;
static JobId job_id;
static int upload_done=0;

static bool make_lurl(string &path) {
  if(path[0] == '/') {
    path=string("file://")+path;
  }
  else {
    char buf[BUFSIZ];
    if(getcwd(buf,sizeof(buf)) == NULL) {
      odlog(0) << "Current directory path is too long." << endl;
      return false;
    };
    path=string("file://")+buf+"/"+path;
  };
  return true;
};

#ifndef __CREATE_EXEC__
int ui_uploader(const char* resource,const char* rsl,char** job_id,const char* session_url,rsl_action act,const vector<string> &filenames,int debug_level,int timeout) {
  char* jobid = NULL;
  if(job_id) jobid=(*job_id);
  LogTime::level=debug_level;
#else
int main(int argc,char** argv) {
  const char* resource = NULL;
  const char* rsl = NULL;
  char* jobid = NULL;
  const char* session_url = NULL;
  rsl_action act = RSL_ACTION_REQUEST;
  LogTime::level=2;
  int timeout = -1;
#endif
  int res=0;
  int i;
  int retries=0;
  LogTime::active=false;

#ifdef __CREATE_EXEC__
  for(;;) {
    opterr=0;
    char optc=getopt(argc,argv,"+hcCpi:r:R:");
    if(optc == -1) break;
    switch(optc) {
      case 'h': {
        cerr<<"ui-uploader [-r resource] [-R rsl] [-c|C|p] [-i job_id] session_dir_url|session_root_url [list]"<<endl;
        exit(1);
      }; break;
      case 'c': {
        if(act != RSL_ACTION_REQUEST) {
          cerr<<"Only one non-default action can be specified"<<endl;
          exit(1);
        };
        act=RSL_ACTION_CANCEL;
      }; break;
      case 'C': {
        if(act != RSL_ACTION_REQUEST) {
          cerr<<"Only one non-default action can be specified"<<endl;
          exit(1);
        };
        act=RSL_ACTION_CLEAN;
      }; break;
      case 'p': {
        if(act != RSL_ACTION_REQUEST) {
          cerr<<"Only one non-default action can be specified"<<endl;
          exit(1);
        };
        act=RSL_ACTION_RENEW;
      }; break;
      case 'i': {
        if(jobid != NULL) {
          cerr<<"Only one id can be given"<<endl;
          exit(1);
        };
        jobid=strdup(optarg);
      }; break;
      case 'r': {
        if(resource != NULL) {
          cerr<<"Only one resource can be given"<<endl;
          exit(1);
        };
        resource=optarg;
      }; break;
      case 'R': {
        if(rsl != NULL) {
          cerr<<"Only one rsl can be specified"<<endl;
          exit(1);
        };
        rsl=optarg;
      }; break;
      case '?': {
        cerr<<"Unsupported option '"<<(char)optopt<<"'"<<endl;
        exit(1);
      }; break;
      case ':': {
        cerr<<"Missing parameter for option '"<<(char)optopt<<"'"<<endl;
        exit(1);
      }; break;
      default: {
        cerr<<"Undefined processing error"<<endl;
        exit(1);
      };
    };
  };
#endif
  char* rsl_act = NULL;
  char* rsl_act_templ="&(executable=/bin/echo)(savestate=yes)(action=%s)(jobid=%s)";
  if((act == RSL_ACTION_CANCEL) || (act == RSL_ACTION_CLEAN) || (act == RSL_ACTION_RENEW)) {
    if(!resource) { odlog(0)<<"Missing resource contact string"<<endl; return 1; };
    if(!jobid) { odlog(0)<<"Missing job identifier"<<endl; return 1; };
    rsl_act=(char*)malloc(strlen(rsl_act_templ)+strlen(jobid)+6+1);
    if(!rsl_act) { odlog(0)<<"Memory allocation error"<<endl; return 1; };
    if(act == RSL_ACTION_CLEAN) {
      sprintf(rsl_act,rsl_act_templ,"clean",jobid);
    }
    else {
      sprintf(rsl_act,rsl_act_templ,"cancel",jobid);
    };
    rsl=rsl_act; 
  }
  else if(act == RSL_ACTION_REQUEST) {
#ifdef __CREATE_EXEC__
    session_url = argv[optind];
#endif
    if(!session_url) {
      odlog(0)<< "Missing session directory url" << endl; return 1;
    };
  }
  else {
    odlog(0)<<"This action is not defined"<<endl;
    return 1;
  };
  GlobusModuleCommon mod_common;
  if(!mod_common.active()) {
    odlog(0)<<"COMMON module activation failed\n"; return 1;
  };
  if(resource) {
    if(rsl == NULL) {
      odlog(0)<<"Missing RSL"<<endl;
      return 1;
    };
    if(strncmp(resource,"gsiftp://",9)) {
      if(globus_module_activate(GLOBUS_GRAM_CLIENT_MODULE) != GLOBUS_SUCCESS) {
        odlog(0)<<"GRAM_CLIENT module activation failed\n"; 
        return 1;
      };
      char* job_contact;
      int error=globus_gram_client_job_request((char*)resource,rsl,0,NULL,&job_contact);
      if(error != GLOBUS_SUCCESS) {
        odlog(0)<<"RSL submission failed: "<<globus_gram_protocol_error_string(error)<<endl;
        globus_module_deactivate(GLOBUS_GRAM_CLIENT_MODULE);
        return 1;
      };
      globus_module_deactivate(GLOBUS_GRAM_CLIENT_MODULE);
      if(job_contact == NULL) {
        odlog(0)<<"Globus gatekeeper did not return job contact"<<endl;
        return 1;
      };
      if(strncmp("https://",job_contact,8)) {
        odlog(0)<<"Bad contact returned: "<<job_contact<<endl;
        return 1;
      };
      char *s=strchr(job_contact+8,'/');
      if(s == NULL) {
        odlog(0)<<"Bad contact returned: "<<job_contact<<endl;
        return 1;
      };
      s++; /* start of numbers */
      char* s_ = s+(strlen(s)-1);
      if((s == s_) || ((*s_) != '/')) {
        odlog(0)<<"Bad contact returned: "<<job_contact<<endl;
        return 1;
      };
/*  no more / to . - just remove */
      for(s_=s;s_;) { 
        s_=strchr(s,'/');
        if(s_) memmove(s_,s_+1,strlen(s_));
      }; 
      if(jobid == NULL) jobid=strdup(s);
      odlog(1)<<"Assigned job id: "<<jobid<<endl;
#ifndef __CREATE_EXEC__
      if(job_id) if((*job_id) == NULL) (*job_id)=jobid;
#endif
      globus_free(job_contact);
    }
    else { /* submission through gsiftp */
      GlobusModuleFTPControl mod_ftp_control;
      if(!mod_ftp_control.active()) {
        odlog(0)<<"FTP_CONTROL module activation failed\n"; 
        return 1;
      };
      globus_url_t url_;
      if(globus_url_parse(resource,&url_) != GLOBUS_URL_SUCCESS) {
        odlog(0)<<"Bad resource url"<<endl;
        return 1;
      };
      char* host = url_.host;
      if(!host) {
        odlog(0)<<"Missing host in resource url"<<endl;
        return 1;
      };
      unsigned short port = url_.port;
      if(port == 0) {
        odlog(0)<<"Missing or wrong port in resource url"<<endl;
        return 1;
      };
      char* path = url_.url_path;
      if(!path) {
        odlog(0)<<"Missing path in resource url"<<endl;
        return 1;
      };

      if(act == RSL_ACTION_CANCEL) {
        if(!ftpsubmit(host,port,path,ftpsubmit_cancel_req,&jobid,timeout)) {
          odlog(0)<<"Cancel request failed"<<endl;
          return 1;
        };
      }
      else if(act == RSL_ACTION_CLEAN) {
        if(!ftpsubmit(host,port,path,ftpsubmit_clean_req,&jobid)) {
          odlog(0)<<"Clean request failed"<<endl;
          return 1;
        };
      }
      else if(act == RSL_ACTION_RENEW) {
        if(!ftpsubmit(host,port,path,ftpsubmit_renew_req,&jobid)) {
          odlog(0)<<"Proxy renew request failed"<<endl;
          return 1;
        };
      }
      else {
        if(!ftpsubmit(host,port,path,rsl,&jobid)) {
          odlog(0)<<"RSL submission failed"<<endl;
          return 1;
        };
      };
      odlog(1)<<"Assigned job id: "<<jobid<<endl;
#ifndef __CREATE_EXEC__
      if(job_id) if((*job_id) == NULL) (*job_id)=jobid;
#endif
    };
  };
  if(act != RSL_ACTION_REQUEST) {
    if(rsl_act) free(rsl_act);
    return 0;
  };
  /* 2 arguments required: url of session dir and file containing
     list of files to upload (pairs: remote file name + local file name;
     if local file name not given it is the same as remote; if
     list is not given - read from stdin */
#ifdef __CREATE_EXEC__
  char* file_list = argv[optind+1];
#endif
  string base_url=session_url;
  if(base_url.length() == 0) {
    odlog(0)<< "Session directory url is empty" << endl; return 1;
  };
  if(base_url[base_url.length()-1] == '/') {
    base_url.erase(base_url.length()-1,1);
  };
  if(jobid != NULL) base_url=base_url+"/"+jobid;
  DataMovePar mover;
  if(timeout > 0) {
    mover.set_default_min_speed(0,timeout);
    mover.set_default_max_inactivity_time(timeout);
  };
#ifdef __CREATE_EXEC__
  if(file_list) {
    if(!job_Xput_read_file(string(file_list),job_files)) {
      odlog(0)<< "Can't read list of files from " << file_list << endl;
      res=1; goto exit;
    };
  }
  else {
    if(!job_Xput_read_file(job_files)) {
      odlog(0)<< "Can't parse list of files" << endl;
      res=1; goto exit;
    };
  };
  /* push names to mover */
  for(FileData::iterator i=job_files.begin();i!=job_files.end();++i) {
    string source;
    if(i->lfn.length() != 0) { source=i->lfn; }
    /* pfn always starts with '/' */
    else { source=i->pfn.substr(1); };
    string destination = base_url+i->pfn;
    if(!make_lurl(source)) {
      odlog(1)<<"Failed converting local name to local url"<<endl;
      res=1; goto exit;
    };
    if(!mover.Add(source.c_str(),destination.c_str())) {
      odlog(0)<<"Can't add data pair to list: "<<endl; res=1; goto exit;
    };
  };
#else
  /* push names to mover */
  for(vector<string>::const_iterator iv=filenames.begin();iv!=filenames.end();++iv) {
    string destination = *iv;
    string source; 
    canonical_dir(destination);
    ++iv; if(iv==filenames.end()) { source=""; }
    else { source = *iv; };
    if(source.length() == 0) { source=destination.substr(1); };
    destination = base_url+destination;
    if(!make_lurl(source)) {
      odlog(1)<< "Failed converting local name to local url" << endl;
      res=1; goto exit;
    };
    if(!mover.Add(source.c_str(),destination.c_str())) {
      odlog(0)<<"Can't add data pair to list: "<<endl; res=1; goto exit;
    };
    if(iv==filenames.end()) break;
  };
#endif
  mover.secure(false);
  mover.passive(true);  /* make it work even from behind firewall */
  if(!mover.Transfer(DataCache(),UrlMap(),MAX_UPLOADS)) {
    odlog(0)<<"FAILED to transfer files"<<endl; res=2; goto exit;
  };
  // Check if all files have been properly uploaded
  for(;;) {
    string source;
    string destination;
    DataMovePar::result dres;
    if(!mover.Get(source,destination,dres)) { // no more data
      break;
    };
    if(dres == DataMovePar::success) {  // remove this file from list
      odlog(1)<<"Uploaded "<<source<<" to "<<destination<<endl;
    }
    else {
      odlog(1)<<"Failed "<<source<<" to "<<destination<<endl;
      res=1;
    };
  };
  /* still can be uploads running - must cancel here */
  odlog(1) << "Leaving uploader" << endl;
exit:
#ifndef __CREATE_EXEC__
  if(jobid) {
    if(job_id) { if((*job_id) != jobid) free(jobid); }
    else { free(jobid); };
  };
#else
  if(jobid) free(jobid);
#endif
  return res;
}
