/*
  
   Copyright 2004 Jonas Bardino

   This file is part of NorduGRID GridSSH.

    NorduGRID GridSSH is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    NorduGRID GridSSH is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with NorduGRID GridSSH; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

/* 

Expiration of keys can be handled by this expire queue implementation.
The queue is a simple linked list with each element in the list containing 
the information required to expire the authorization validity of a certain key.
The threaded environment adds a need for locking the queue to avoid race 
conditons .

*/

#include "runtime_conf.h"
#include "expire_queue.h"
#include "file_io.h"
#include "log.h"

expire_queue *expire_q;


int initialize_expire_q(void)
{

  if(conf->expire_mode != NEVER_EXPIRE) {
    expire_q = malloc(sizeof(expire_queue));
    assert(expire_q != NULL);
    globus_mutex_init(&(expire_q->lock), NULL);
    expire_q->next = 0;
  } else {
    expire_q = NULL;
  }

  return 0;
}

int finalize_expire_q(void)
{
  /* All children are reaped at this point - no locking */
  int expired = 0;
  expire_elem *e = NULL;

  if(conf->expire_mode != NEVER_EXPIRE) {

    if(expire_q->next != 0) {
      e = expire_q->queue;
      /* next != 0 implicitly assures that e != NULL */
      assert(e != NULL);
      while(e != NULL) {
	debugmsg(5,"%s expiring key %s", __FUNCTION__, e->pubkey);
	remove_line(e->pubkey, e->auth_filename);
	expire_q->queue = e->next;
	e->next = NULL;
	free(e);
	e = expire_q->queue;
	expired++;
      }
    }

    globus_mutex_destroy(&(expire_q->lock));
    free(expire_q);
  }

  debugmsg(2,"%s expired %d keys", __FUNCTION__, expired);
  return expired;
}

int check_expire(void)
{
  expire_elem *e = NULL;
  int expired = 0;
  time_t now = 0;

  globus_mutex_lock(&(expire_q->lock));

  if(expire_q->next != 0) {
    now = time(NULL);
    debugmsg(4,"%s - now is %d - next removal is %d",__FUNCTION__,now,expire_q->next);

    e = expire_q->queue;
    /* next != 0 implicitly assures that e != NULL */
    assert(e != NULL);
    while(e != NULL && e->expire <= now) {
      remove_line(e->pubkey, e->auth_filename);
      debugmsg(2,"%s: removed expired key %d",__FUNCTION__, expired);
      expire_q->queue = e->next;
      e->next = NULL;
      debugmsg(8,"%s: freeing expired key element %d strings",__FUNCTION__,expired);
      free(e->pubkey);
      free(e->auth_filename);
      debugmsg(8,"%s: freeing expired key element %d",__FUNCTION__,expired);
      free(e);
      e = expire_q->queue;
      expired++;
    }
    debugmsg(4,"%s: setting next removal",__FUNCTION__);
    if(expire_q->queue == NULL) {
      expire_q->next = 0;
    } else {
      expire_q->next = (expire_q->queue)->expire;
    }
    debugmsg(8,"%s: leaving",__FUNCTION__);
  }

  globus_mutex_unlock(&(expire_q->lock));

  return expired;
}

int sched_expire(char *key, char *filename, time_t time)
{
  /* expire queue is a linked list sorted with growing expiration time */
  expire_elem *e = NULL, *p = NULL, *n = NULL;

  e = malloc(sizeof(expire_elem));
  assert(e != NULL);

  debugmsg(6,"%s: creating queue element",__FUNCTION__);
  /* args will be freed from caller - must copy strings */
  e->pubkey = malloc((strlen(key)+1)*sizeof(char));
  e->auth_filename = malloc((strlen(filename)+1)*sizeof(char));
  assert(e->pubkey != NULL);
  assert(e->auth_filename != NULL);
  strcpy(e->pubkey, key);
  strcpy(e->auth_filename, filename);
  e->expire = time;
  e->next = NULL;
  
  globus_mutex_lock(&(expire_q->lock));

  if(expire_q->next == 0 || expire_q->next >= time) {
    /* no elements or only elements with later expiry - prepend element */
    debugmsg(5,"%s: inserting in front of chain",__FUNCTION__);
    e->next = expire_q->queue;
    expire_q->queue = e;
  } else {
    /* now e must be somewhere inside chain: find exact position */
    debugmsg(6,"%s: searching chain",__FUNCTION__);
    p = expire_q->queue;
    n = p->next;
    while(n != NULL && n->expire <= time) {
      p = n;
      n = n->next;
    }
    /* Now n points at the first element with expire > time */
    /* p points at the element before it - insert in between */
    debugmsg(5,"%s: inserting inside chain",__FUNCTION__);
    e->next = n;
    p->next = e;
  }

  debugmsg(4,"%s: setting next removal time",__FUNCTION__);
  /* first element always has the lowest expiration time */
  expire_q->next = (expire_q->queue)->expire;
  globus_mutex_unlock(&(expire_q->lock));

  return 0;   
}
