/*
  
   Copyright 2004 Jonas Bardino

   This file is part of NorduGRID GridSSH.

    NorduGRID GridSSH is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    NorduGRID GridSSH is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with NorduGRID GridSSH; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

/*

Gridssh client using globus I/O with authentication and encryption included.

Example usage: gridssh -p 1234 tmp.imada.sdu.dk

Built upon the echoserver client example by 
Arto Ters <arto.teras@hip.fi> 18.6.2002
from para2002 (http://www.csc.fi/para2002/tutorial.phtml).

The client initially uses a grid proxy for authentication and 
authorization at a server. The result is that the server inserts a public
key matching the private key available in the local globus userkey.pem in 
the list of authorized keys. Afterwards the client can ssh/scp/sftp into the 
server by using the userkey.
The actual gridssh commands are therefore just passed on as normal ssh commands 
to the server.
*/

#include "runtime_conf.h"
#include "gridssh.h"
#include "log.h"
#include "file_io.h"
#include "home_dir.h"
#include "auth_client.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>

/* basename */
#include <libgen.h>

/* getopt */
#include <unistd.h>

/* fork/waitpid*/
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>

/* getopt prototypes/globals may not be included in unistd */
int getopt(int argc, char * const argv[], const char *optstring);
extern char *optarg;
extern int optind, opterr, optopt;

/*  Global constants  */
extern runtime_conf *conf;

int initialize_conf(int argc, char **argv)
{
  conf = malloc(sizeof(runtime_conf));
  assert(conf != NULL);
  
  /* Initialize runtime configuration structure, conf */
  conf->argv = argv;
  conf->argc = argc;
  conf->max_procs = UNUSED;
  conf->log_stats = 0;
  conf->use_cache = 0;
  conf->debug_level = 0;
  conf->sshd_port = DEFAULT_SSHD_PORT;
  conf->gridsshd_port = DEFAULT_GRIDSSHD_PORT;
  conf->address = UNUSED;
  conf->uid = getuid();
  conf->gid = getgid();
  conf->pid = getpid();
  conf->prog = basename(argv[0]);
  conf->ssh_client = "ssh";
  conf->ssh_user = NULL;
  conf->out_file = stdout;
  conf->daemon_mode = UNUSED;
  conf->do_auth = 1;
  conf->cache_filename = DEFAULT_CACHE_FILE;

  return 0;
}

int finalize_conf(void)
{
  conf->argv = NULL;
  conf->prog = NULL;
  conf->ssh_client = NULL;
  if(conf->ssh_user) free(conf->ssh_user);
  conf->ssh_user = NULL;
  free(conf);
  return 0;
}


void print_usage(void)
{
  showstr("### %s %s ###\n", conf->prog, GRIDSSH_VERSION);
  showstr("Usage: %s [options] host\n", conf->prog);
  showstr("Options:\n");
  showstr("\t-c client_path: use non-standard %s client\n", conf->ssh_client);
  showstr("\t-C: enable auth cache i.e. don't do auth if cache indicates it can be avoided\n");
  showstr("\t-d debug_level: select debug level (default 0 i.e. debug off)\n");
  showstr("\t-p gridsshd_port: force %s to connect to the specified gridsshd port (default %s)\n", conf->prog, DEFAULT_GRIDSSHD_PORT);
  showstr("\t-P sshd_port: force %s to connect to the specified sshd port (default %s)\n", conf->prog, DEFAULT_SSHD_PORT);
  showstr("\t-v: Print version details\n");
  showstr("Example: %s -d 3 -p 1234 tmp.imada.sdu.dk\n", conf->prog);
}


int main(int argc, char *argv[]) {
    
  int error = 0;

  /* optarg vars */
  int ch = -1;
  int itmp = -1;
  /* Valid options ('+' stops arg processing as soon as a non option is found) */
  const char *optstr = "+c:Cd:Dhp:P:v";

  /* ssh call variables */
  char **ssh_argv = NULL;
  int ssh_argc = -1;
  /* char *src_user= NULL, *dst_user = NULL;*/
  int i = -1;
  int verbose_ssh = 0;

  /* fork / waitfor */
  pid_t pid;
  int child_status = -1;

  char *home_dir = NULL, *user_at_host_str = NULL, *server = NULL;
  auth_status *status = NULL;
  

  initialize_conf(argc, argv);
  /* output to stdout and stderr */
  initialize_log(NULL, NULL);
  initialize_file_io();
  initialize_home_dir();

  /*** Don't do anything before this point as we need conf for most actions ***/
  
  /*  Get command line arguments  */
  while (-1 != (int) (ch = getopt(argc, argv, optstr))) {
    switch (ch) {
    case 'c':
      conf->ssh_client = optarg;
      break;
    case 'C':
      conf->use_cache = 1;
      break;
    case 'd':
      itmp = atoi(optarg);
      if (itmp >= 0)
	conf->debug_level = itmp;
      else {
	showstr("Illegal argument to -d option: debug_lvl must be a nonnegative integer!\n");
	print_usage();
	exit(EXIT_FAILURE);
      }
      break;
    case 'D':
      verbose_ssh = 1;
      break;
    case 'h':
      print_usage();
      exit(EXIT_SUCCESS);
    case 'p':
      itmp = atoi(optarg);
      if (itmp > 0 && itmp < MAX_PORT) {
	conf->gridsshd_port = optarg;
      } else {
	showstr("Illegal argument to -p option: gridsshd port must be a positive integer (< %d)!\n", MAX_PORT);
	print_usage();
	exit(EXIT_FAILURE);
      }
      break;
    case 'P':
      itmp = atoi(optarg);
      if (itmp > 0 && itmp < MAX_PORT) {
       	conf->sshd_port = optarg;
      } else { 
       	showstr("Illegal argument to -P option: sshd port must be a positive integer(< %d)!\n", MAX_PORT);
       	print_usage();
       	exit(EXIT_FAILURE);
      }
      break;
    case 'v':
      showstr("### %s %s ###\n", conf->prog, GRIDSSH_VERSION);
      break;
      /*
	getopt return '?' if it finds an option not listed in the options str
	ing
      */
    case '?':
      showstr("Unrecognized option: %c\n", optopt);
      print_usage();
      exit(EXIT_FAILURE);
    default:
      showstr("getopt error? condition unaccounted for?\n");
      exit(EXIT_FAILURE);
    }
  }

  if(optind >= argc) {
    print_usage();
    exit(EXIT_FAILURE);
  } 

  server = argv[optind];

  /* do authentication and authorization with each of the involved servers */

  status = authorize(server, conf->gridsshd_port);
  if(status->expire == NOT_AUTHORIZED) {
    errmsg("authorization with %s failed - can't continue", server);
    exit(EXIT_FAILURE);
  }

  /* authentication and authorization done - Now forward actual ssh call */

  /* exec arg list must end with a NULL */
  ssh_argc = (argc-optind) + INTERNAL_SSH_ARGS + 1 + verbose_ssh;
  debugmsg(6,"building argv of length %d", ssh_argc);
  ssh_argv = (char **)malloc(ssh_argc * sizeof(char *));
  assert(ssh_argv != NULL);
  ssh_argv[0] = conf->ssh_client;
  if((home_dir = get_home()) == NULL) {
    errmsg("failed to retrieve home_dir - can't continue");
    exit(EXIT_FAILURE);
  }
  ssh_argv[1] = malloc((strlen(KEY_FLAG)+strlen(home_dir)+1+strlen(DEFAULT_KEY_FILE)+1)*sizeof(char));
  assert(ssh_argv[1] != NULL);
  sprintf(ssh_argv[1], "%s%s/%s", KEY_FLAG, home_dir,DEFAULT_KEY_FILE);
  free(home_dir);
  ssh_argv[2] = malloc((strlen(PORT_FLAG)+strlen(conf->sshd_port)+1)*sizeof(char));
  assert(ssh_argv[2] != NULL);
  sprintf(ssh_argv[2], "%s%s", PORT_FLAG, conf->sshd_port);

  if(verbose_ssh) ssh_argv[INTERNAL_SSH_ARGS] = "-v";

  debugmsg(6,"appending args (optind %d, argc %d)", optind, argc);
  for(i = optind; i < argc; i++) {
    if(i != optind) {
      debugmsg(6,"appending raw arg %s in ssh_argv[%d]", argv[i],(i-1)+INTERNAL_SSH_ARGS+verbose_ssh);
      ssh_argv[(i-optind)+INTERNAL_SSH_ARGS+verbose_ssh] = argv[i];
    } else {
      debugmsg(5,"appending user prefixed host arg %s@%s in ssh_argv[%d]", status->ssh_user, argv[i],(i-1)+INTERNAL_SSH_ARGS+verbose_ssh);
      user_at_host_str = malloc((strlen(status->ssh_user)+1+strlen(argv[i])+1)*sizeof(char));
      assert(user_at_host_str != NULL);
      sprintf(user_at_host_str, "%s@%s", status->ssh_user, argv[i]);
      ssh_argv[(i-optind)+INTERNAL_SSH_ARGS+verbose_ssh] = user_at_host_str;
    }
  }
  
  /* Update ssh_argc to not extend actually used entries */
  ssh_argc = (argc-optind)+INTERNAL_SSH_ARGS+verbose_ssh + 1;
  
  debugmsg(8,"NULL terminating argv (%d / %d)", (argc-optind)+INTERNAL_SSH_ARGS+verbose_ssh, ssh_argc);
  ssh_argv[(argc-optind)+INTERNAL_SSH_ARGS+verbose_ssh] = (char *)NULL;
  
  if(conf->debug_level >= 4) {
    showstr("execv(");
    for(i = 0; i < ssh_argc-1; i++) {
      showstr("%s ", ssh_argv[i]);
    }
    showstr(")\n");
  }

  /* exec ssh command in child to avoid truncating stack etc */
  pid = fork();

  if (pid < 0) {
    errmsg("fork to execute %s failed", conf->ssh_client);
  } else if (pid == 0) {
    debugmsg(3,"Executing actual %s command", conf->ssh_client);
    error = execvp(ssh_argv[0], ssh_argv);
    /* 
       execvp/execve does not return on success, and the text, data, bss, 
       and stack of the calling process are overwritten by that of the 
       program loaded. So in short it doesn't make sense to do anything after
       this point.
    */
    exit(EXIT_FAILURE);
  } else {
    debugmsg(6,"parent waiting for child to exit");
    waitpid(pid, &child_status, 0);
    debugmsg(3,"child returned with status %d (%d, %d)", child_status, WIFEXITED(child_status), WEXITSTATUS(child_status));
    /* status contains both exit code and signal - extract the former */    
    error = WEXITSTATUS(child_status);
    /* TODO: login, C-c, C-d gives error unlike same sequence in pure ssh */
    if(error != 0) {
      debugmsg(3,"Execution of %s failed with error %d", conf->ssh_client, error);

      if(conf->use_cache && error == SSH_ERROR) {
	showstr("---\nThis command failure may be due to inconsistent cache contents possibly caused by a gridssh server restart. If you think that it is the case try running the command again without the '-C' option or after removing at least the related server entries from the cache file (~/%s)\n---\n", conf->cache_filename);
      /* TODO: remove related cache entries upon failure? */
      }
    }
  }
  
  debugmsg(3,"Cleaning up");
  ssh_argv[0] = NULL;
  free(ssh_argv[1]);
  free(ssh_argv[2]);
  free(ssh_argv[3]);
  ssh_argv[4] = NULL;
  for(i = optind; i < argc; i++) {
    ssh_argv[(i-optind)+INTERNAL_SSH_ARGS+verbose_ssh] = NULL;
  }
  free(ssh_argv);

  free(status->server);
  status->server = NULL;
  free(status->ssh_user);
  status->ssh_user = NULL;
  free(status);
  finalize_home_dir();
  finalize_file_io();
  finalize_log();
  finalize_conf();

  if(error) exit(EXIT_FAILURE);

  return EXIT_SUCCESS;
}
