/*
  
   Copyright 2004 Jonas Bardino

   This file is part of NorduGRID GridSSH.

    NorduGRID GridSSH is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    NorduGRID GridSSH is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    
    You should have received a copy of the GNU General Public License
    along with NorduGRID GridSSH; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    
*/

  
/*
  
General logging functions
  
debugmsg() only actually does something if __DEBUG_MODE__ is defined.
Otherwise the defines in debug.h will replace those calls with empty lines.
Thus we don't need to clutter up programs with ifdefs to avoid slowdown. 

Warnings that should be available when debugmode is off should use warnmsg().
Similarly error messages should use errmsg() while remaining messages should 
use showmsg().

2004, Jonas Bardino

*/
  

#include <stdio.h>
#include <stdarg.h>
#include <unistd.h>             /* for: getopt(), sleep() */
#include <assert.h>
#include <time.h>
#include "log.h"
#include "runtime_conf.h"

/* prototype not included in time.h */
struct tm *localtime_r(const time_t *timep, struct tm *result);

extern runtime_conf *conf;

#ifdef COLORDEBUG
enum colors { C_BLACK, C_RED, C_GREEN, C_YELLOW, C_BLUE, C_PURPLE, C_CYAN, C_WHITE }; /* used for color-output */

void setfgcolor(int colornum)
{
  /* Change the output color */
  if (colornum >= 0 && colornum < 8)
    printf("%c[3%dm", 0x1B, colornum);
  else
    printf("%c[3%dm", 0x1B, 7);
}
#endif

int initialize_log(char *out, char *err)
{
  if(out != NULL) {
    conf->out_file = fopen(out,"a");
    if(conf->out_file == NULL) {
      conf->out_file = fopen("/dev/stdout","a");
      errmsg("Failed to open %s for out logging - falling back to stdout", out);
      return 0;
    }
  } else {
    conf->out_file = stdout;
  }
  /* Do not buffer out - this should make fflush's redundant */
  setbuf(conf->out_file, (char *) 0);

  if(err != NULL) {
    conf->err_file = fopen(err,"a");
    if(conf->err_file == NULL) {
      conf->err_file = fopen("/dev/stderr","a");
      errmsg("Failed to open %s for err logging - falling back to stderr", err);
      return 0;
    }
  } else {
    conf->err_file = stderr;
  }
  /* Do not buffer err - this should make fflush's redundant */
  setbuf(conf->err_file, (char *) 0);

  if(out != NULL && conf->debug_level > 0) {
    fprintf(stdout,"Logging output to %s\n", out);
  }
  return 1;
}

int finalize_log(void)
{

  debugmsg(5,"Finalizing log using fd %x and %x", conf->out_file, conf->err_file);
  /* No buffering so fclose should never fail */
  fclose(conf->out_file);
  fclose(conf->err_file);
  return 1;
}

/* Next two functions are only defined like this if compiled with -D__DEBUG_MODE__ */
#ifdef __DEBUG_MODE__
void debugmsg(int lvl, char *fmt, ...)
{
  time_t now = 0;
  struct tm now_buf;
  struct tm *t;
  char now_str[TIME_STR_LEN];
  va_list args;
  
  assert(conf != NULL);
  
  if (conf->debug_level >= lvl) {
    va_start(args, fmt);
    
#ifdef COLORDEBUG
    setfgcolor(lvl % 8);
#endif
    
    time(&now);
    t=localtime_r(&now,&now_buf);
    strftime(now_str,100,"%b %d %H:%M:%S ",t);
    fprintf(conf->out_file, "%s%s[%d]: ", now_str, conf->prog, conf->pid);
    vfprintf(conf->out_file, fmt, args);
    
#ifdef COLORDEBUG
    /* reset color */
    setfgcolor(C_WHITE);
#endif

    fprintf(conf->out_file, "\n");
    va_end(args);
  }
}

void debug_sleep(int lvl, unsigned int secs)
{
  /* Used to indicate possible dead-/livelocks */

  assert(conf != NULL);

  if (conf->debug_level >= lvl) {
    showmsg("%s[%d]: sleeping for %d seconds", conf->prog, conf->pid, secs);
    sleep(secs);
  }
}
#endif


void errmsg(char *fmt, ...)
{
  time_t now = 0;
  struct tm now_buf;
  struct tm *t;
  char now_str[TIME_STR_LEN];
  va_list args;

  assert(conf != NULL);

  va_start(args, fmt);

#ifdef COLORDEBUG
  /* errors in red */
  setfgcolor(1);
#endif

  time(&now);
  t=localtime_r(&now,&now_buf);
  strftime(now_str,100,"%b %d %H:%M:%S ",t);

  fprintf(conf->err_file, "%s%s[%d]: (Error) ", now_str, conf->prog, conf->pid);
  vfprintf(conf->err_file, fmt, args);

#ifdef COLORDEBUG
  /* reset color */
  setfgcolor(C_WHITE);
#endif

  fprintf(conf->err_file, "\n");
  va_end(args);
}


void warnmsg(int lvl, char *fmt, ...)
{
  time_t now = 0;
  struct tm now_buf;
  struct tm *t;
  char now_str[TIME_STR_LEN];
  va_list args;

  assert(conf != NULL);

  if (conf->debug_level >= lvl) {
    va_start(args, fmt);

#ifdef COLORDEBUG
    setfgcolor(3);
#endif

    time(&now);
    t=localtime_r(&now,&now_buf);
    strftime(now_str,100,"%b %d %H:%M:%S ",t);

    fprintf(conf->out_file, "%s%s[%d]: (Warning) ", now_str, conf->prog, conf->pid);
    vfprintf(conf->out_file, fmt, args);

#ifdef COLORDEBUG
    /* reset color */
    setfgcolor(C_WHITE);
#endif

    fprintf(conf->out_file, "\n");
    va_end(args);
  }
}

void showmsg(char *fmt, ...)
{
  time_t now = 0;
  struct tm now_buf;
  struct tm *t;
  char now_str[TIME_STR_LEN];
  va_list args;

  va_start(args, fmt);
    
  time(&now);
  t=localtime_r(&now,&now_buf);
  strftime(now_str,100,"%b %d %H:%M:%S ",t);

  fprintf(conf->out_file, "%s%s[%d]: ", now_str, conf->prog, conf->pid);
  vfprintf(conf->out_file, fmt, args);
    
  fprintf(conf->out_file, "\n");
  va_end(args);
}

void showstr(char *fmt, ...)
{
  /* show string without adding newline */
  va_list args;

  va_start(args, fmt);
  
  vfprintf(conf->out_file, fmt, args);
  
  va_end(args);
}
