"""
The ARC logging module. It is build upon the logging module in Python.
See: http://docs.python.org/lib/module-logging.html

The idea with this module is to unify logging to file in daemons and
command line tools. Basically this should mean that there should be NO
print statements anywhere, since they are quite annoying in libraries.

To use:
 import logger
 log = logger.Logger()

The interface for log is the same as for python logging object;
acually it is a standard logging boject. The above example as good
for CLIs, since it emits INFO and above to stderr and nothing more.
A more advanced example:

 import logger
 logger.setup_logger(stream_debug = True,
                     file_handler = True, file_url = '/tmp/log')
 log = logger.Logger()

This will create a logger that will output everything (including debug) to
stderr and INFO and above to /tmp/log.
"""

import os
import logging

import util
import errors


_DEFAULT_LOGGER_NAME = 'arc'
_DEFAULT_FILE_DIR = '.jm'
_DEFAULT_FILE_NAME = 'log'

"""The logger."""
_log = None


class LoggingError(errors.ARCError):
	"""
	Raised when an error happens within the logger module.
	"""


def _get_log_limit(debug_output=False):
	"""
	Given whether debug should be emitted, this function returns the logging
	level suitable.
	"""
	if debug_output:
		return logging.DEBUG
	else:
		return logging.INFO


def _get_stream_handler(debug_output=False):
	"""
	Creates a stream handler which will emit mesages to stderr.
	If the debug_output is True, debug output will be emitted as well.
	"""
	stream_handler = logging.StreamHandler()
	stream_formatter = logging.Formatter('%(levelname)s %(message)s')
	stream_handler.setFormatter(stream_formatter)

	logging_level = _get_log_limit(debug_output)
	stream_handler.setLevel(logging_level)

	return stream_handler


def _get_file_handler(file_url, debug_output=False):
	"""
	Creates a file handler, which will emit messages to a file.
	The file_url arguments specifies file path.
	If debug_output is True, debug output will be emitted as well.
	"""
	file_handler = logging.FileHandler(file_url)
	file_formatter = logging.Formatter('%(asctime)s %(levelname)s %(message)s')
	file_handler.setFormatter(file_formatter)

	logging_level = _get_log_limit(debug_output)	
	file_handler.setLevel(logging_level)

	return file_handler


def _get_file_url(file_url=None):
	"""
	If the file_url arguments is None, this function returns the default
	url path for the log file. If file_url is not None, file_url is returned
	as is.
	"""
	if file_url is None:
		home = os.getenv('HOME')
		if home is None:
			raise LoggingError('Could not get home path')
		file_url = os.path.join(home, _DEFAULT_FILE_DIR, _DEFAULT_FILE_NAME)

	return file_url


def setup_logger(stream_handler=True, stream_debug=False,
                 file_handler=False, file_debug=False, file_url=None):
	"""
	Sets up the logger. Arguments
	stream_handler - should logging be emitted to stderr
	stream_debug   - should debug messages be emitted to stream
	file_handler   - should loggin happen to a file
	file_debug     - should debug meesages be emitted to file
	file_url       - which file to log to
	"""
	logger = logging.getLogger(_DEFAULT_LOGGER_NAME)

	if stream_handler:
		handler = _get_stream_handler(stream_debug)
		logger.addHandler(handler)

	if file_handler:
		file_url = _get_file_url(file_url)
		handler = _get_file_handler(file_url, file_debug)
		logger.addHandler(handler)

	logger.setLevel(logging.DEBUG)
	global _log
	_log = logger


class Logger(object):
	"""
	Class to represent the logger. Not really a class since it returns
	a loggin object (and not an instance of itself), but people like OO,
	and this way is more future safe.
	"""
	__metaclass__ = util.MetaSingleton

	def __new__(class_):
		"""
		Return the internal logger object to the caller.
		"""
		if _log is None:
			setup_logger()

		return _log

