#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <iostream>
#include <string>
#include <vector>
#include <unistd.h>
#include <sys/stat.h>

#include "Preferences.h"
#include "CertInfo.h"
#include "DateTime.h"
#include "MdsQuery.h"

#define __UI_LIBRARY__
#include "ui_downloader.h"
#include "ui_uploader.h"

#include "ngui.h"


extern "C"
int nggetxx (const vector <string> & jobs,
	     const vector <string> & joblists,
	     const vector <string> & clusterselect_,
	     const vector <string> & clusterreject_,
	     const vector <string> & status,
	     const bool all,
	     string & directory,
	     const bool usejobname,
	     const bool keep,
	     int timeout,
	     int debug,
	     const bool anonymous) {

  vector <string> clusterselect = ResolveAliases (clusterselect_);
  vector <string> clusterreject = ResolveAliases (clusterreject_);

  vector <string> jobids;
  vector <Cluster> clusterlist;

  int error = 0;
  ActivateGlobus();

  // setting defaults

  if (timeout == UNDEFINED) timeout = iGetEnv ("NGTIMEOUT");
  if (timeout == UNDEFINED) timeout = iGetDef ("NGTIMEOUT");
  if (timeout == UNDEFINED) timeout = DEFAULT_TIMEOUT;

  if (debug == UNDEFINED) debug = iGetEnv ("NGDEBUG");
  if (debug == UNDEFINED) debug = iGetDef ("NGDEBUG");
  if (debug == UNDEFINED) debug = 0;

  if (directory.empty()) directory = GetEnv ("NGDOWNLOAD");
  if (directory.empty()) directory = GetDef ("NGDOWNLOAD");
  if (directory.empty()) {
    char buffer[PATH_MAX];
    getcwd (buffer, PATH_MAX);
    directory = buffer;
  }

  // get user info

  CertInfo user;
  if (!user) {
    error = 1;
    goto errorexit;
  }

  if (debug) {
    cout << "User subject name: " << user.GetSN() << endl;
    cout << "Remaining proxy lifetime: "
	 << Period (user.TimeLeft()) << endl;
  }

  // finding the information

  jobids = GetJobIDs (jobs, joblists, clusterselect, clusterreject, status,
		      all);

  if (all && jobids.empty()) {
    cout << "ngget: No jobs" << endl;
    goto errorexit;
  }

  if (jobids.empty()) {
    cerr << "ngget: no valid jobids given" << endl;
    cerr << "Use \"ngget -help\" for help" << endl;
    error = 1;
    goto errorexit;
  }

  for (vector <string>::iterator vsi = jobids.begin();
       vsi != jobids.end(); vsi++) {
    int pos0 = 0;
    if (vsi->substr (0, 9) == "gsiftp://") pos0 = 9;
    int pos = vsi->find_first_of (":/", pos0);
    if (pos == string::npos) {
      cerr << "ngget: invalid jobid: " << *vsi << endl;
      error = 1;
      continue;
    }
    string clustername = vsi->substr (pos0, pos - pos0);
    bool found = false;
    for (vector <Cluster>::iterator cli = clusterlist.begin();
	 !found && cli != clusterlist.end(); cli++)
      if (clustername == cli->GetName()) found = true;
    if (!found) clusterlist.push_back (clustername);
  }

  if (clusterlist.empty()) {
    cerr << "ngget: no valid jobids given" << endl;
    error = 1;
    goto errorexit;
  }

  FindClusterInfo (clusterlist, Mds::JobManipulation, user.GetSNx(), anonymous,
		   timeout, debug);

  for (vector <string>::iterator vsi = jobids.begin();
       vsi != jobids.end(); vsi++) {

    Cluster * c = NULL;
    string jobstatus;
    string jobname;

    bool found = false;
    bool selected = false;

    for (vector <Cluster>::iterator cli = clusterlist.begin();
	 !found && cli != clusterlist.end(); cli++)
      for (vector <Queue>::iterator qli = cli->queues.begin();
	   !found && qli != cli->queues.end(); qli++)
	for (vector <Job>::iterator jli = qli->jobs.begin();
	     !found && jli != qli->jobs.end(); jli++)
	  if (*vsi == jli->GetId()) {
	    if (jli->IsStatus (status)) {
	      selected = true;
	      c = &*cli;
	      jobstatus = jli->GetStatus();
	      jobname = jli->GetName();
	    }
	    found = true;
	  }

    if (!found) {
      cerr << "ngget: no job with jobid " << *vsi << " found" << endl;
      error = 1;
      continue;
    }
    if (!selected)
      continue;
    if (jobstatus == "DELETED") {
      cerr << "ngget: job " << *vsi << " has already been deleted" << endl;
      error = 1;
      continue;
    }
    if (jobstatus != "FINISHED" && jobstatus != "FAILED" &&
	jobstatus != "CANCELLED") {
      cerr << "ngget: job " << *vsi << " has not finished yet." << endl;
      error = 1;
      continue;
    }
    int pos = vsi->rfind ('/');
    if (pos == string::npos) {
      cerr << "ngget: invalid jobid: " << *vsi << endl;
      error = 1;
      continue;
    }
    string shortid = vsi->substr (pos + 1);

    string downloaddir;
    if (usejobname && !jobname.empty())
      downloaddir = directory + '/' + jobname;
    else
      downloaddir = directory + '/' + shortid;

    struct stat stx;
    if (stat (downloaddir.c_str(), &stx) == 0) {
      cerr << "Error: download directory " << downloaddir
	   << " already exists." << endl;
      error = 1;
      continue;
    }

    const vector <string> filenames;

    cout << "ngget: downloading files to " << downloaddir << endl;

    int err = ui_downloader ((c->GetContact() + '/' + shortid).c_str(), true,
			     downloaddir.c_str(), filenames, true, false,
			     debug, timeout);
    if (err) {
      cerr << "ngget: download incomplete - job kept on gatekeeper." << endl;
      error = 1;
      continue;
    }

    if (!keep) {
      cout << "ngget: download successful - deleting job from gatekeeper."
	   << endl;

      char * shortjobid = strdup (shortid.c_str());
      int err = ui_uploader (c->GetContact().c_str(), NULL, &shortjobid, NULL,
			     RSL_ACTION_CLEAN, filenames, debug, timeout);
      free (shortjobid);

      if (err) {
	cerr << "ngget: job " << *vsi << " could not be deleted." << endl;
	error = 1;
	continue;
      }

      RemoveJobID (*vsi);
    }
  }

 errorexit:
  DeactivateGlobus();
  return error;
}
